/* @(#)splint.c	8.3 (ESO-IPG) 01/03/95 14:09:20 */
/* @(#)splint.c	1.2 (ESO-IPG) 2/9/94 18:16:12 */
/* @(#)splint.c	5.1.1.1  (ESO)  4/5/93  17:52:13 */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* .COPYRIGHT   (C) 1993 European Southern Observatory     */
/* .IDENT       .prg                                       */
/* .AUTHORS     Pascal Ballester (ESO/Garching)            */
/*              Cristian Levin   (ESO/La Silla)            */
/* .KEYWORDS    Spectroscopy, Long-Slit                    */
/* .PURPOSE                                                */
/* .VERSION     1.0  Package Creation  17-MAR-1993         */
/* ------------------------------------------------------- */

/* @(#)splint.c    1.0.0.0 (ESO-La Silla) 10/08/91 12:00:00 */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.IDENT        splint.c
.MODULE       subroutines -- sprebin.exe
.LANGUAGE     C
.AUTHOR       Cristian Levin - ESO La Silla
.PURPOSE      This file contains two routines for making spline
              interpolation:
                  - using Hermite polynomials.
                  - using natural-cubic spline.
              The first one is currently used for rebinning.
.KEYWORDS     rebin, spline interpolation.
.VERSION 1.0  1-Apr-1991   Implementation
.ENVIRONMENT  UNIX
------------------------------------------------------------*/

/*
  functions of dsplint:


  double dsplint_lewin( double xp, double *x, float *y, int n, int *istart )
  


  double dsplint2( double xp, double *x, float *y, float *y2, int n, 
                   int *kstart )



  void spline(float x[], float y[], int n, float yp1, float ypn, float y2[])



  void splint(float xa[], float ya[], float y2a[], int n, float x, float *y)



  void splin2(float x1a[], float x2a[], float **ya, float **y2a, int m, int n,
	      float x1, float x2, float *y)



  int splie2(float x1a[], float x2a[], float **ya, int m, int n, float **y2a)



  int splin3(float x1a[], float x2a[], float **ya, float **y2a, int m, int n,
             int num, float x1[], float x2, float y[])

 */


#include <stdio.h>
#include <math.h>
#include <proto_fors.h>

/***************************************************************
 *
 * splint(): spline interpolation based on Hermite polynomials.
 *
 ***************************************************************/
#ifdef __STDC__
   double dsplint_lewin( double xp, double *x, float *y, int n, int *istart )
#else
   double dsplint_lewin( xp, x, y, n, istart )
   double xp,*x;
   float *y;
   int n,*istart;
#endif

/* double xp;	 x-value to interpolate */
/* double *x;	 x-array [1..n] */
/* float  *y;	 y-array [1..n] */
/* int *istart; initial index   */

{
    double yp1, yp2, yp;
    double xpi, xpi1, l1, l2, lp1, lp2;
    int i;

    if ( xp < x[1] || xp > x[n] )
	return(0.0);

    for ( i = *istart; i <= n && xp >= x[i]; i++ )
	;

    *istart = i;
    i--;
    
    lp1 = 1.0 / (x[i] - x[i+1]);
    lp2 = -lp1;

    if ( i == 1 )
	yp1 = (y[2] - y[1]) / (x[2] - x[1]);
    else
	yp1 = (y[i+1] - y[i-1]) / (x[i+1] - x[i-1]);

    if ( i >= n - 1 )
	yp2 = (y[n] - y[n-1]) / (x[n] - x[n-1]);
    else
	yp2 = (y[i+2] - y[i]) / (x[i+2] - x[i]);

    xpi1 = xp - x[i+1];
    xpi  = xp - x[i];
    l1   = xpi1*lp1;
    l2   = xpi*lp2;

    yp = y[i]*(1 - 2.0*lp1*xpi)*l1*l1 + 
         y[i+1]*(1 - 2.0*lp2*xpi1)*l2*l2 + 
         yp1*xpi*l1*l1 + yp2*xpi1*l2*l2;

    return(yp);
}

/***************************************************************
 *
 * dsplint2(): natural cubic-spline interpolation. (double)
 *
 ***************************************************************/

#ifdef __STDC__
   double dsplint2( double xp, double *x, float *y, float *y2, int n, 
                  int *kstart )
#else
   double dsplint2( xp, x, y, y2,  n, kstart )
   double xp,*x;
   float  *y,*y2;
   int    n,*kstart;
#endif

/* double xp;	 x-value to interpolate */

/* double xp;	 x-value to interpolate */
/* double *x;	 x-array [1..n] */
/* float  *y;	 y-array [1..n] */
/* float  *y2;	 y2-array [1..n] (2-nd derivatives) */
/* int n;	 */
/* int *kstart;	 initial index */

{
#ifdef __STDC__
  void nrerror(char error_text[]);
#else
  void nrerror();
#endif

  int klo, khi, k;
  double a, b, h, yp;

  klo = *kstart;
  khi = n;
  
  if ( xp < x[1] || xp > x[n] )
    return(0.0);
  else if ( xp == x[1] )
    return(y[1]);
  
  for ( k = klo; k < n && xp > x[k]; k++ )
    ;
  
  klo = *kstart = k-1;
  khi = k;
  
  h = x[khi] - x[klo];
  if ( h == 0.0 )
    nrerror( "Error in spline interpolation" );
  
  a = (x[khi] - xp) / h;
  b = (xp - x[klo]) / h;
  
  yp = a*y[klo] + b*y[khi] + ((a*a*a - a)*y2[klo] + (b*b*b - b)*y2[khi])*
    (h*h) / 6.0;
  
  return(yp);
}


/* original numerical recipes floating point 4 byte */
#ifdef __STDC__
void spline(float x[], float y[], int n, float yp1, float ypn, float y2[])
#else
void spline(x, y, n, yp1, ypn, y2)
float x[], y[], yp1, ypn, y2[];
int n;
#endif
{
	int i,k;
	float p,qn,sig,un,*u;

	u=vector(1,n-1);
	if (yp1 > 0.99e30)
		y2[1]=u[1]=0.0;
	else {
		y2[1] = -0.5;
		u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
	}
	for (i=2;i<=n-1;i++) {
		sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p=sig*y2[i-1]+2.0;
		y2[i]=(sig-1.0)/p;
		u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
		u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
	}
	if (ypn > 0.99e30)
		qn=un=0.0;
	else {
		qn=0.5;
		un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
	}
	y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0);
	for (k=n-1;k>=1;k--)
		y2[k]=y2[k]*y2[k+1]+u[k];
	free_vector(u,1,n-1);
}

#ifdef __STDC__
void splint(float xa[], float ya[], float y2a[], int n, float x, float *y)
#else
  void splint(xa, ya, y2a, n, x, y)
  float xa[], ya[], y2a[], x, *y;
int n;
#endif
{
#ifdef __STDC__
  void nrerror(char error_text[]);
#else
  void nrerror();
#endif
	int klo,khi,k;
	float h,b,a;

	klo=1;
	khi=n;
	while (khi-klo > 1) {
		k=(khi+klo) >> 1;
		if (xa[k] > x) khi=k;
		else klo=k;
	}
	h=xa[khi]-xa[klo];
	if (h == 0.0) nrerror("Bad xa input to routine splint");
	a=(xa[khi]-x)/h;
	b=(x-xa[klo])/h;
	*y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
}


 
#ifdef __STDC__
int splie2(float x1a[], float x2a[], float **ya, int m, int n, float **y2a)
#else
int splie2(x1a, x2a, ya, m, n, y2a)
float x1a[], x2a[], **ya, **y2a;
int m, n; 
#endif
{
  int j;
  for (j=1; j<m; j++)
    {
      spline(x2a, ya[j], n, 1.0e30, 1.0e30, y2a[j]); 
    }
  return 0;
}


/* but this is the real splin2 of the NR page 128 */
#ifdef __STDC__
void splin2(float x1a[], float x2a[], float **ya, float **y2a, int m, int n,
	    float x1, float x2, float *y)
#else
void splin2(x1a,  x2a,  ya, y2a, m,  n, x1,  x2, y)
     float  x1a[], x2a[], **ya, **y2a, x1, x2, *y;
     int m, n;
#endif
{
  /*
  void spline(float x[], float y[], int n, float yp1, float ypn, float y2[]);
  void splint(float xa[], float ya[], float y2a[], int n, float x, float *y);
  */
  int j;
  float *ytmp,*yytmp;
  
  ytmp=vector(1,n);
  yytmp=vector(1,n);
  for (j=1;j<=m;j++)
    splint(x2a,ya[j],y2a[j],n,x2,&yytmp[j]);
  spline(x1a,yytmp,m,1.0e30,1.0e30,ytmp);
  splint(x1a,yytmp,ytmp,m,x1,y);
  free_vector(yytmp,1,n);
  free_vector(ytmp,1,n);
}


#ifdef __STDC__
int splin3(float x1a[], float x2a[], float **ya, float **y2a, int m, int n,
	   int num, float x1[], float x2, float y[])
#else
int splin3(x1a, x2a, ya, y2a, m, n, num, x1, x2, y)
float x1a[], x2a[], **ya, **y2a, x1[], x2, *y;
int m, n, num;
#endif
{
#ifdef __STDC__
  void nrerror(char error_text[]);
#else
  void nrerror();
#endif

  int j;
  float *ytmp,*yytmp;

  int klo,khi,k;
  float h,b,a;

  klo=1;
  khi=m;

  ytmp=vector(1,m);
  yytmp=vector(1,m);
  for (j=1;j<=m;j++)
    splint(x2a,ya[j],y2a[j],n,x2,&yytmp[j]);

  spline(x1a,yytmp,m,1.0e30,1.0e30,ytmp); 

  while (khi-klo > 1) 
    {
      k=(khi+klo) >> 1;
      if (x1a[k] > x1[0]) khi=k;
      else klo=k;
    }
  h=x1a[khi]-x1a[klo];
  if (h == 0.0) nrerror("Bad xa input to routine splint");

  for (j=0;j<num;j++)
    {
      if ( (x1[j] > x1a[1]) && (x1[j] < x1a[m]) )
	{
	  if (x1[j] > x1a[khi])
	    {
	      while (x1[j] > x1a[khi])
		{
		  klo++;
		  khi++;
		}
	      h=x1a[khi]-x1a[klo];
	      if (h == 0.0) nrerror("Bad xa input to routine splint");
	    }
	}
      
      a=(x1a[khi]-x1[j])/h;
      b=(x1[j]-x1a[klo])/h;
      y[j] = a*yytmp[klo] + b*yytmp[khi] + 
	((a*a*a-a)*ytmp[klo] + (b*b*b-b)*ytmp[khi])*(h*h)/6.0;
    }
  free_vector(yytmp,1,m);
  free_vector(ytmp,1,m);
  return 0;
}




/* original numerical recipes floating point 4 byte */
#ifdef __STDC__
void dspline(double x[], double y[], int n, double yp1, double ypn, 
	     double y2[])
#else
     void dspline(x, y, n, yp1, ypn, y2)
     double x[], y[], yp1, ypn, y2[];
     int n;
#endif
{
  int i,k;
  double p,qn,sig,un,*u;
  
  u = dvector(1,n-1);
  if (yp1 > 0.99e30)
    y2[1]=u[1]=0.0;
  else {
    y2[1] = -0.5;
    u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
  }
  for (i=2;i<=n-1;i++) {
    sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
    p=sig*y2[i-1]+2.0;
    y2[i]=(sig-1.0)/p;
    u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
    u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
  }
  if (ypn > 0.99e30)
    qn=un=0.0;
  else {
    qn=0.5;
    un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
  }
  y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0);
  for (k=n-1;k>=1;k--)
    y2[k]=y2[k]*y2[k+1]+u[k];
  free_dvector(u,1,n-1);
}


#ifdef __STDC__
void 
dsplint(double xa[], double ya[], double y2a[], int n, double x, 
	double *y)
#else
     void dsplint(xa, ya, y2a, n, x, y)
     double xa[], ya[], y2a[], x, *y;
     int n;
#endif
{
#ifdef __STDC__
  void nrerror(char error_text[]);
#else
  void nrerror();
#endif
  int klo,khi,k;
  double h,b,a;
  
  klo=1;
  khi=n;
  while (khi-klo > 1) {
    k=(khi+klo) >> 1;
    if (xa[k] > x) khi=k;
    else klo=k;
  }
  h=xa[khi]-xa[klo];
  if (h == 0.0) nrerror("Bad xa input to routine splint");
  a=(xa[khi]-x)/h;
  b=(x-xa[klo])/h;
  *y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
}








