/* midified_horne.c       modified optimum extraction 
  
  
   by Thomas Szeifert 
      Otmar Stahl

      FORS project
      data reduction software group
      Landessternwarte Heidelberg

      17 September 1997

 */



#include <math.h>
#include <midas_def.h>
#include <proto_fors.h>


#define mymax(A,B) ((A)  > (B) ? (A) : (B))
#define mymin(A,B) ((A)  < (B) ? (A) : (B))
#define myabs(A) ((A)  >  (float) 0.0 ? (A) : (-A))




#ifdef __STDC__
int 
modified_horne(float *frame, float *sky, float *outframe, float *varframe, 
	       float mask[],float profile[], float variance[],
	       float y[], float z[], int slitlen, float *raw, 
	       int Npix[], int iparm[], float rparm[])
#else
int 
modified_horne( frame, sky, outframe, varframe, mask, profile, variance, 
		y, z, slitlen, raw, Npix, iparm, rparm)
float *frame,*sky,*outframe,*varframe,mask[],profile[],variance[],
  y[],z[],*raw,rparm[];
int slitlen,Npix[],iparm[];
#endif
{

  int i, j, k, index, index1, index2, kk;
  int medwin=7,ncos=0,ocos=-1,lcos=1,l=0;
  int Order, N_iter,debug;

  float ratio, diff, diff2, rmax, sum1, sum2, sum3, mp;
  float stst[2];
  float Gain, V0, Thres;

  char string[80];

  /* new - dlfit not lfit:*/
  double *dcoef;
  float *xin,*xout,*sigin;


  /* previously in the header as automatic variables: */
  Gain = rparm[1];
  V0 = (SQR(rparm[0])) / (SQR(Gain));
  Thres = SQR(rparm[2]);

  Order  = iparm[0];
  N_iter = iparm[1];
  debug  = iparm[2];

  if (debug > 2){
    printf("horne V0 = %7.1f Gain = %7.1f Thres = %7.1f\n",V0, Gain, Thres);
    printf("      Order = %4d N_iter = %4d debug = %4d \n",Order, N_iter,debug);
    printf("      Npix = %4d %4d slitlength = %4d \n",Npix[0],Npix[1],slitlen);
  }
  if (debug > 0)
    if (Order < 0)
      printf ("... extract with equal weights  - no cosmic filter\n");
    else if (N_iter <= 0)
      printf ("... extract with equal weights  - filter cosmic\n");
    else
      printf ("... modified optimum extraction - filter cosmic\n");

  if (Order > 0)
    dcoef = (double *) dvector(1,Order);
  xin = (float *) fvector(0,Npix[0]-1);
  xout = (float *) fvector(0,Npix[0]-1);
  sigin = (float *) fvector(0,Npix[0]-1);


  stst[0] = (float) 1.0;
  stst[1] = (float) 1.0;

  for (i = 0; i < Npix[0]; i++)
    {
      xin[i] = xout[i] = stst[0] + stst[1] * (float) (i - Npix[0]);
      outframe[i] = (float) 0.0; 
      varframe[i] = (float) 0.0; 
    }




/*                 init arrays                            */

  for (j = 0; j < slitlen; j++)
    {
      for (i = 0; i < Npix[0]; i++)
	{
	  index = j * Npix[0] + i;
	  mask[index] = (float) 1.0;
	  profile[index] = (float) 0.0;
	  variance[index] = (float) 0.0;
	  outframe[i] += (frame[index] - sky[index]);
	}
    }

  for (i = 0; i < Npix[0]; i++)
    {
      raw[i] = outframe[i];
      if (outframe[i] <= (float) 0.0){
	for (j = 0; j < slitlen; j++)
	  {
	    index = j * Npix[0] + i;
	    mask[index] = (float) 0.0;
	  }
      }
    }

/*              return with standard extraction (equal weights)     */

  if (Order <= 0)
    {

      for (i = 0; i < Npix[0]; i++)
	  varframe[i] = (float) sqrt( (double) (V0 + outframe[i]/Gain) );

      free_fvector(xin,0,Npix[0]-1);
      free_fvector(xout,0,Npix[0]-1);
      free_fvector(sigin,0,Npix[0]-1);

      return (0);
    }

  /*              construct spatial profile
     lcos = 1     first compute median and filter cosmic
                  set (lcos=0) if the no cosmics were found (ocos = ncos) 
		  or if the number of cosmic iterations l > slitlen
     lcos = 0     fit the spatial profile, continue filtering    */


  for (k = 0; k <= N_iter; k++)
    {
      ocos = ncos;

      for (i = 0; i < Npix[0]; i++)
	outframe[i] = mymax (outframe[i], 1.0);

      for (j = 0; j < slitlen; j++)
	{
	  for (i = 0; i < Npix[0]; i++)
	    {
	      index = j * Npix[0] + i;
	      y[i] = (frame[index] - sky[index]) / outframe[i];
	      sigin[i] = (float) sqrt(mymax((V0 + frame[index]/Gain),V0)) / 
		outframe[i];
	    }

	  /*   filter spatial profile                              */
	  if (lcos == 1){ 
	    for (i = medwin; i < Npix[0] - medwin; i++)
	      {
		index = j * Npix[0] + i;

		profile[index] = heap_median ((2*medwin+1), &y[i-medwin]); 

	      }
	  
	    /*      treat edges                                   */

	    index1 = j * Npix[0] + medwin;
	    index2 = (j + 1) * Npix[0] - (medwin+1);
	    for (i = 0; i < medwin; i++)
	      {
		index = j * Npix[0] + i;
		profile[index] = profile[index1];
	      }
	    for (i = Npix[0] - medwin; i < Npix[0]; i++)
	      {
		index = j * Npix[0] + i;
		profile[index] = profile[index2];
	      }
	  }
	  else{
	    for (i = 0; i < Npix[0]; i++)
	      {
		index = j * Npix[0] + i;
		profile[index] = y[i];
	      }
	  }
	  /*        fit a polynomial to spatial profile          */

	  if (lcos != 1)
	    {

	      index = j * Npix[0];
	      select_fit_poly(&profile[index], xin, sigin, Npix[0], 
			      &mask[index], z, xout, Npix[0], Order, dcoef); 

	      for (i = 0; i < Npix[0]; i++)
		{
		  index = j * Npix[0] + i;
		  profile[index] = z[i];
		}
	    }
	}
      

      /*      normalize spatial profile, enforce positivity         */
      
      for (i = 0; i < Npix[0]; i++)
	z[i] = 0.0;

      for (j = 0; j < slitlen; j++)
	{
	  for (i = 0; i < Npix[0]; i++)
	    {
	      index = j * Npix[0] + i;
	      
	      /*               THIS HAS TO BE INVESTIGATED !!       */

	      /*
		profile[index] = mymax (profile[index], 0.0);
	      */

	      z[i] += profile[index];
	    }
	}
      
      for (j = 0; j < slitlen; j++)
	{
	  for (i = 0; i < Npix[0]; i++)
	    {
	      index = j * Npix[0] + i;
	      if (z[i] != (float) 0.0)
		profile[index] = profile[index] / z[i];
	      
	    }
	}
      
      /*             compute variance                     */
      

      /* note:  V0 = SQR(RON/GAIN) !!! Thres = SQR(Thres) !!! */
      for (j = 0; j < slitlen; j++)
	{
	  for (i = 0; i < Npix[0]; i++)
	    {
	      index = j * Npix[0] + i;
	      variance[index] = V0 + 
		(outframe[i] * profile[index] + sky[index]) / Gain;
	    }
	}
      for (i = 0; i < Npix[0]; i++)
	{
	  rmax = 1.0;
	  kk = -1;
	  for (j = 0; j < slitlen; j++)
	    {
	      index = j * Npix[0] + i;
	      diff = (frame[index] - sky[index]) - outframe[i] *
		profile[index];
	      diff2 = diff * diff;
	      ratio = (diff2 / (Thres * variance[index]));

	      if (mask[index] < (float) 1.0 && ratio < (float) 1.0)
		mask[index] = (float) 1.0;    /* heal the good pixels */

	      if (ratio * mask[index] > rmax) 
		{
		  rmax = ratio;
		  kk = index;                 /* find the worst pixel */
		}
	      
	    }
	  
	  if (kk >= 0 && rmax > (float) 1.0){
	    mask[kk] = (float) 0.0;           /* reject the worst pixel */
	    ncos++;
	  }


	  /*      now mask the cosmics and extract the weighted spectrum */

	  sum1 = (float) 0.0;
	  sum2 = (float) 0.0;
	  sum3 = (float) 0.0;
	  
	  for (j = 0; j < slitlen; j++)
	    {
	      index = j * Npix[0] + i;
	      mp = mask[index] * profile[index];
	      sum1 += mp * (frame[index] - sky[index]) / variance[index];
	      sum2 += mp * (profile[index] / variance[index]);
	      sum3 += mp;
	    }
	    

	  if (sum2 != (float) 0.0)
	    outframe[i] = sum1 / sum2;
	  
	  if (k == N_iter && sum2 > 0)
	    varframe[i] = (float) sqrt(myabs(sum3/sum2));
	  else if (k == N_iter)
	    varframe[i] = (float) 0.0;

	}


      if (debug > 0)
	{
	  if (lcos == 1)
	    sprintf(string,"cosmic  iteration %4d -- %4d detections",k,ncos);
	  else
	    sprintf(string,"profile iteration %4d -- %4d detections",k,ncos);
	  if (debug > 1)
	    SCTPUT(string);
	  else
	    {
	      SCTDIS(string,80);
	      if (k == N_iter)
		printf("\n");
	    }
	}


      if ((ncos == ocos && lcos == 1) || ++l >= 3) 
	lcos = 0;
      if (lcos == 1)
	k--;

    }
  
  free_dvector(dcoef,1,Order);
  free_fvector(xin,0,Npix[0]-1);
  free_fvector(xout,0,Npix[0]-1);
  free_fvector(sigin,0,Npix[0]-1);

  return (1);
}


/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
select_fit_poly (float *inimage, float *ximage, float *sigin, int npixi, 
		 float *mask, float *outimage, float *ximageo, int npixo, 
		 int fit_deg, double ad[])
#else
void
select_fit_poly ( inimage, ximage, sigin, npixi, mask, outimage, ximageo, 
		  npixo, fit_deg, ad)
  float *inimage, *ximage, *outimage, *ximageo,*mask,*sigin;
  int npixi, npixo, fit_deg;
  double ad[];
#endif

/*      fit a polynomial (Numerical Recipes), modified for echelle!  */

{

  int i,isel=0;
  float *xsel,*ysel,*sigsel;

  xsel = (float *) fvector(0,npixi-1);
  ysel = (float *) fvector(0,npixi-1);
  sigsel = (float *) fvector(0,npixi-1);

  for (i = 0 ; i < npixi ; i++)
    if (mask[i] > 0.5 && inimage[i] != (float) 0.0 && sigin[i] > (float) 0.0){
      xsel[isel] = ximage[i];
      ysel[isel] = inimage[i];
      sigsel[isel] = sigin[i];
      isel++;
    }

  if (isel > fit_deg)
    fit_poly_weight (ysel, xsel, sigsel, isel, outimage, ximageo, npixo, 
		     fit_deg, ad); 
  else
    for (i = 0 ; i < npixi ; i++)
      outimage[i] = inimage[i]; 

  free_fvector(xsel,0,npixi-1);
  free_fvector(ysel,0,npixi-1);
  free_fvector(sigsel,0,npixi-1);

}

#undef mymax
#undef mymin
#undef myabs
