/*

  mutil.c 

  on the FORS library we have all the codes which are jointly used by the 
  packages mos-lsw ipol spol quick and the FORS echelle codes

  In this file one can append more mathematical function written by the team,
  but the file contains some sorting and median routines from the 
  numerical recipies. If entire modules of the numerical recipies are 
  required on should better not copy them to this file but to a new file,
  like in the case of fitnol or dsplint.



  functions in mutil.c



  fit_poly (float inimage[], float ximage[], int npixi, 
    float *outimage, float *ximageo, int npixo, 
    int fit_deg, double *ad)


  dfit_poly (double inimage[], double ximage[], int npixi, 
    double*outimage, double *ximageo, int npixo, 
    int fit_deg, double *ad, int poltyp)


  eval_fpoly (float x, float *a, int n)


  eval_dpoly (double x, double *a, int n)


  dpoly (double x, double p[], int np)


  eval_dcheb (double x, double *a, int n)


  dcheb (double x, double pl[], int nl)


  eval_dleg (double x, double *a, int n)


  dleg (double x, double pl[], int nl)


  median (float x[], int n)


  piksrt (int n, float arr[])


  median2 (float x[], int n)


  float select(unsigned long k, unsigned long n, float arr[])


  select_pos (int k, int n, float arr[])


  pik_median (int n, float arr[])


  heap_median (int n, float arr[])


  the following five are in an preliminary "state of the art":
    dechfit2() echfit() evaldech2() evaldech() dfitech()


  polint (float xa[], float ya[], int n, float x, float *y, float *dy)


  dsvdcmp (double **a, int m, int n, double w[], double **v)


  dsvbksb (double **u, double w[], double **v, int m, int n, double b[], 
    double x[])


  dpythag (double a, double b)
*/


#include <stdio.h>
#include <math.h>
#include <proto_fors.h>


static int degy,degxy;


/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
fit_poly (float *inimage, float *ximage, int npixi, 
	  float *outimage, float *ximageo, int npixo, 
	  int fit_deg, double *ad)
#else
void
fit_poly ( inimage, ximage, npixi, outimage, ximageo, npixo, fit_deg, ad )
  float *inimage, *ximage, *outimage, *ximageo;
  int npixi, npixo, fit_deg;
  double *ad;
#endif

/*      fit a polynomial (Numerical Recipes), modified for echelle!  */

{
  /*  float xout;*/
  double xout;
  double *xin, *yin;
  int i;

  /* new - dlfit not lfit*/
  double *sig, chisq, **covar;
  int    *ia;
  covar = (double **) dmatrix (1, fit_deg, 1, fit_deg);
  sig = (double *) dvector (1, npixi);
  ia = (int *) ivector (1, fit_deg);

  xin = dvector (1, npixi);
  yin = dvector (1, npixi);


  for (i = 0; i < npixi; i++)
    {
      xin[i+1] = (double) ximage[i];
      yin[i+1] = (double) inimage[i];
      sig[i+1] = (double) 1.0;
    }
  for (i = 1 ; i <= fit_deg ; i++)
    ia[i] = 1;


  dlfit (xin, yin, sig, npixi, ad, ia, fit_deg, covar, &chisq, dpoly);

  /*
    for (i = 0; i < npixi; i++)
    {
    xin[i] = (double) ximage[i];
    yin[i] = (double) inimage[i];
    }

    lfit (xin, yin, npixi, ad, fit_deg, dpoly);
    */
  


  for (i = 0; i < npixo; i++)
    {
      xout = (double) ximageo[i];
      outimage[i] = (float) eval_dpoly (xout, ad, fit_deg);
    }


  free_dvector(xin, 1, npixi); 
  free_dvector(yin, 1, npixi);
  free_dmatrix(covar, 1, fit_deg, 1, fit_deg);
  free_dvector(sig,1, npixi);
  free_ivector(ia, 1, fit_deg);
}

/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
dfit_poly (double inimage[], double ximage[], int npixi, 
	  double*outimage, double *ximageo, int npixo, 
	  int fit_deg, double *ad, int poltyp)
#else
void
dfit_poly ( inimage, ximage, npixi, outimage, ximageo, npixo, fit_deg, ad, 
	    poltyp )
  double inimage[], ximage[], *outimage, *ximageo;
  int npixi, npixo, fit_deg, poltyp;
double *ad;
#endif

/*      fit a polynomial (Numerical Recipes), modified for echelle!  */

{
  double xout;
  double *xin, *yin;
  int i;

  /* new - dlfit not lfit*/
  double *sig, chisq, **covar;
  int    *ia;
  covar = (double **) dmatrix (1, fit_deg, 1, fit_deg);
  sig = (double *) dvector (1, npixi);
  ia = (int *) ivector (1, fit_deg);

  xin = dvector (1, npixi);
  yin = dvector (1, npixi);


  for (i = 0; i < npixi; i++)
    {
      xin[i+1] = (double) ximage[i];
      yin[i+1] = (double) inimage[i];
      sig[i+1] = 1;
    }
  for (i = 1 ; i <= fit_deg ; i++)
    ia[i] = 1;


  if (poltyp == 0)
    dlfit (xin, yin, sig, npixi, ad, ia, fit_deg, covar, &chisq, dpoly);
  else if (poltyp == 1)
    dlfit (xin, yin, sig, npixi, ad, ia, fit_deg, covar, &chisq, dcheb);
  else
    dlfit (xin, yin, sig, npixi, ad, ia, fit_deg, covar, &chisq, dleg);


  for (i = 0; i < npixo; i++)
    {
      xout = ximageo[i];
      if (poltyp == 0)
	outimage[i] = (float) eval_dpoly (xout, ad, fit_deg);
      else if (poltyp == 1)
	outimage[i] = (float) eval_dcheb (xout, ad, fit_deg);
      else
	outimage[i] = (float) eval_dleg  (xout, ad, fit_deg);
    }

  free_dvector(xin, 1, npixi); 
  free_dvector(yin, 1, npixi);
  free_dmatrix(covar, 1, fit_deg, 1, fit_deg);
  free_dvector(sig,1, npixi);
  free_ivector(ia, 1, fit_deg);


}
/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
dfit_cali_2D (double inimage[], double ximage[], double yimage[], int npixi, 
	  double outimage[], double ximageo[], double yimageo[], int npixo, 
	  int fit_deg, double ad[], int poltyp, int cali_degy, int cali_degxy)
#else
void
dfit_cali_2D ( inimage, ximage, yimage, npixi, outimage, ximageo, yimageo, 
	    npixo, fit_deg, ad, poltyp ,cali_degy,cali_degxy)
  double inimage[], ximage[], yimage[], outimage[], ximageo[], yimageo[];
  int npixi, npixo, fit_deg, poltyp, cali_degy, cali_degxy;
  double ad[];
#endif

/*      fit a polynomial (Numerical Recipes), modified for echelle!  */

{
  double xout;
  double *x1in,*x2in, *yin;
  double *sig, chisq, **covar;

  int *ia;
  int i,nl;


  x1in = dvector (1, npixi);
  x2in = dvector (1, npixi);
  yin = dvector (1, npixi);
  sig = dvector (1, npixi);

  degy = cali_degy, degxy = cali_degxy;

  nl = fit_deg + degy + degxy + 1; /* nuber of coefs */
  covar = dmatrix (1, nl, 1, nl);
  ia = ivector (1, nl);



  for (i = 0; i < npixi; i++)
    {
      x1in[i+1] = ximage[i];
      x2in[i+1] = yimage[i];
      yin[i+1] = inimage[i];
      sig[i+1] = (double) 1.0;
    }

  for (i = 1 ; i <= nl ; i++)
    ia[i] = 1;



  if (poltyp == 0)
    dlfit_2D(x1in, x2in, yin, sig, npixi, ad, ia, nl, covar, &chisq, dpoly_2D);
  else if (poltyp == 1)
    dlfit_2D(x1in, x2in, yin, sig, npixi, ad, ia, nl, covar, &chisq, dcheb_2D);
  else
    dlfit_2D(x1in, x2in, yin, sig, npixi, ad, ia, nl, covar, &chisq, dleg_2D);


  /*  ximageo[i];*/
  if (poltyp == 0)
    eval_disp_2D (ximageo,  yimageo, outimage, npixo, ad, nl, "POLY");
  else if (poltyp == 1)
    eval_disp_2D (ximageo,  yimageo, outimage, npixo, ad, nl, "CHEB");
  else
    eval_disp_2D (ximageo,  yimageo, outimage, npixo, ad, nl, "LEGE");


  free_dvector(x1in, 1, npixi); 
  free_dvector(x2in, 1, npixi); 
  free_dvector(yin, 1, npixi);
  free_dvector(sig,1, npixi);

  free_dmatrix(covar, 1, nl, 1, nl);
  free_ivector(ia, 1, nl);

  return;
}



/* ------------------------------------------------------------------------- */
#ifdef __STDC__
void eval_disp_2D (double x[], double y[], double l[], int n, double coef[],
		   int ncoef, char poltyp[])
#else
void eval_disp_2D (x, y, l, n, coef,ncoef, poltyp)
     double x[], y[], l[], coef[];
     int n,ncoef;
     char poltyp[];
#endif

{

  int actvals, i, icoef;
  double *xp;
  xp = dvector(1,ncoef);

  for (i = 0; i < n; i++)
    {
      l[i] = 0.;

      if (toupper(poltyp[0]) == 'L')
        dleg_2D (x[i], y[i], xp, ncoef);
      else if (toupper(poltyp[0]) == 'C')
        dcheb_2D (x[i], y[i], xp, ncoef);
      else if (toupper(poltyp[0]) == 'P')
        dpoly_2D (x[i], y[i], xp, ncoef);
      else
	printf("ERROR - You have tried an invalid polynom type \n"); 

      for (icoef = 1; icoef <= ncoef; icoef++)
	l[i] += coef[icoef] * xp[icoef];
    }

  free_dvector(xp,1,ncoef);

  return;
}


/* ------------------------------------------------------------------------- */
#ifdef __STDC__
void eval_disp_2D_rebi (double x[], double y[], double l[], int n, 
			double coef[], int ncoef, char poltyp[],int cali_degy,
			int cali_degxy)
#else
void eval_disp_2D_rebi (x, y, l, n, coef,ncoef, poltyp,degy,degxy)
     double x[], y[], l[], coef[];
     int n,ncoef,degy,degxy;
     char poltyp[];
#endif

{

  int actvals, i, icoef;
  double *xp;

  degy = cali_degy;
  degxy = cali_degxy;
  xp = dvector(1,ncoef);


  for (i = 0; i < n; i++)
    {
      l[i] = 0.;

      if (toupper(poltyp[0]) == 'L')
        dleg_2D (x[i], y[i], xp, ncoef);
      else if (toupper(poltyp[0]) == 'C')
        dcheb_2D (x[i], y[i], xp, ncoef);
      else if (toupper(poltyp[0]) == 'P')
        dpoly_2D (x[i], y[i], xp, ncoef);
      else
	printf("ERROR - You have tried an invalid polynom type \n"); 

      for (icoef = 1; icoef <= ncoef; icoef++)
	l[i] += coef[icoef] * xp[icoef];
    }

  free_dvector(xp,1,ncoef);

  return;
}






/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
fit_poly_weight (float *inimage, float *ximage, float *sigin, int npixi, 
	  float *outimage, float *ximageo, int npixo, 
	  int fit_deg, double *ad)
#else
void
fit_poly_weight ( inimage, ximage, sigin, npixi, outimage, ximageo, npixo, 
		  fit_deg, ad )
  float *inimage, *ximage, *outimage, *ximageo,*sigin;
  int npixi, npixo, fit_deg;
  double *ad;
#endif

/*      fit a polynomial (Numerical Recipes), modified for echelle!  */

{
  /*  float xout;*/
  double xout;
  double *xin, *yin;
  int i;

  /* new - dlfit not lfit*/
  double *sig, chisq, **covar;
  int    *ia;

  covar = (double **) dmatrix (1, fit_deg, 1, fit_deg);
  sig = (double *) dvector (1, npixi);
  ia = (int *) ivector (1, fit_deg);

  xin = (double *) dvector (1, npixi);
  yin = (double *) dvector (1, npixi);



  for (i = 0; i < npixi; i++)
    {
      xin[i+1] = (double) ximage[i];
      yin[i+1] = (double) inimage[i];
      sig[i+1] = (double) sigin[i];;
    }
  for (i = 1 ; i <= fit_deg ; i++)
    ia[i] = 1;


  dlfit (xin, yin, sig, npixi, ad, ia, fit_deg, covar, &chisq, dpoly);



  for (i = 0; i < npixo; i++)
    {
      xout = (double) ximageo[i];
      outimage[i] = (float) eval_dpoly (xout, ad, fit_deg);
    }


  free_dvector(xin, 1, npixi); 
  free_dvector(yin, 1, npixi);
  free_dvector(sig, 1, npixi);
  free_dmatrix(covar, 1, fit_deg, 1, fit_deg);
  free_ivector(ia, 1, fit_deg);
}



/************************************************************

 * eval_dpoly(): fast polynomial evaluation.
 *
 ************************************************************/
#ifdef __STDC__
float 
eval_fpoly (float x, float *a, int n)
#else
float 
eval_fpoly (x, a, n)
     float x, *a;
     int n;
#endif

/*  *a: coefficients (range: 1..n) */

{
  int i;
  float y = 0.0;

  for (i = n; i >= 1; i--)
    y = x * y + a[i];

  return (y);
}


#ifdef __STDC__
double 
eval_dpoly (double x, double *a, int n)
#else
double 
eval_dpoly (x, a, n)
     double x, *a;
     int n;
#endif

/* *a: coefficients (range: 1..n) */

{
  int i;
  double y = 0.0;

  for (i = n; i >= 1; i--)
    y = x * y + a[i];

  return (y);
}

#ifdef __STDC__
void 
dpoly (double x, double p[], int np)
#else
void 
dpoly (x, p, np)
     double x, p[];
     int np;
#endif
{
  int j;

  p[1] = 1.0;
  for (j = 2; j <= np; j++)
    p[j] = p[j - 1] * x;
}


/* ------------------------------------------------------------------------- */
#ifdef __STDC__
double 
eval_dcheb (double x, double *a, int n)
#else
double
eval_dcheb (x, a, n)
     double x, *a;
     int n;
#endif
{

  int icoef;
  double *xp,y;

  xp = dvector(1,n);

  y = 0.;
  dcheb (x, xp, n);

  for (icoef = 1; icoef <= n; icoef++)
    y += a[icoef] * xp[icoef];

  free_dvector(xp,1,n);
  return (y);
}
/* ------------------------------------------------------------------------- */
#ifdef __STDC__
void
dcheb (double x, double pl[], int nl)
#else
void 
dcheb (x, pl, nl)
     double x, pl[];
     int nl;
#endif
{
  int j;
  double twox;

  pl[1] = 1.0;
  pl[2] = x;
  if (nl > 2)
    {
      twox = 2.0 * x;
      for (j = 3; j <= nl; j++)
          pl[j] = twox * pl[j - 1] -  pl[j - 2];
    }
}

/* ------------------------------------------------------------------------- */
#ifdef __STDC__
double 
eval_dleg (double x, double *a, int n)
#else
double
eval_dleg (x, a, n)
     double x, *a;
     int n;
#endif
{

  int icoef;
  double *xp,y;

  xp = dvector(1,n);

  y = 0.;
  dleg (x, xp, n);

  for (icoef = 1; icoef <= n; icoef++)
    y += a[icoef] * xp[icoef];

  free_dvector(xp,1,n);
  return (y);
}

/* ------------------------------------------------------------------------- */
#ifdef __STDC__
void
dleg (double x, double pl[], int nl)
#else
void 
dleg (x, pl, nl)
     double x, pl[];
     int nl;
#endif

{
  int j;
  double twox, f2, f1, d;

  pl[1] = 1.0;
  pl[2] = x;
  if (nl > 2)
    {
      twox = 2.0 * x;
      f2 = x;
      d = 1.0;
      for (j = 3; j <= nl; j++)
        {
          f1 = d;
          d += 1.0;
          f2 += twox;
          pl[j] = (f2 * pl[j - 1] - f1 * pl[j - 2]) / d;
        }
    }
}



/* ---------------------------------------------------------------------------
       Non linear fitting utilities in two dimension   by Thomas Szeifert. 

       The functions:
       
       fleg_2D       (Legendre  polynoms)
       fcheb_2D      (Chebichev polynoms)
       fpoly_2D      ("normal"  polynoms)

       are specially designed for fits of the dispersion relation in 
       spectroscopy - we presume a small slope of the central wavelength and
       the first order dispersion coeficient and fit these in y-direction
       with "normal" polynoms of the order degy and degxy respectively.

-----------------------------------------------------------------------------*/


#ifdef __STDC__
  void dleg_2D( double x1, double x2, double pl[], int nl)
#else
  void dleg_2D(x1,x2,pl,nl)
  int nl;
  double x1,x2,pl[];
#endif  
{
  /* Legendre polynom in x1 - polynom in y and one mixed polynom x1*x2*/
  int j;
  float twox,f2,f1,d;
  
  pl[1]=1.0;
  pl[2]=x1;
  if (nl > 2) 
    {
      twox=2.0*x1;  /*Legendre in x*/
      f2=x1;
      d=1.0;
      for (j=3;j<=nl-(degy+degxy);j++) 
	{
	  f1=d;
	  d += 1.0;
	  f2 += twox;
	  pl[j]=(f2*pl[j-1]-f1*pl[j-2])/d;
	}

      pl[nl-(degy+degxy)+1]=x2;  /*polynom in y*/
      for (j=nl-(degy+degxy)+2;j<=(nl-degxy);j++)
	pl[j] = x2*pl[j-1];
      
      pl[nl-degxy+1] = x1*x2;    /*mixed polynom*/  
      for (j=nl-degxy+2;j<=nl;j++)
	pl[j] = x2*pl[j-1];
    }
  return;
}


#ifdef __STDC__
  void dcheb_2D( double x1, double x2, double pl[], int nl)
#else
  void dcheb_2D(x1,x2,pl,nl)
  int nl;
  double x1,x2,pl[];
#endif  
{
  /* Chebichev polynom in x1 - polynom in y and one mixed polynom x1*x2*/
  int j;
  double twox;


  pl[1]=1.0;
  pl[2]=x1;
  if (nl > 2) 
    {
      twox=2.0*x1;  /*Chebichev in x*/
      for (j=3;j<=nl-(degy+degxy);j++) 
	pl[j]=twox*pl[j-1]-pl[j-2];
  
      pl[nl-(degy+degxy)+1]=x2;  /*polynom in y*/
      for (j=nl-(degy+degxy)+2;j<=(nl-degxy);j++)
	pl[j] = x2*pl[j-1];
      
      pl[nl-degxy+1] = x1*x2;    /*mixed polynom*/  
      for (j=nl-degxy+2;j<=nl;j++)
	pl[j] = x2*pl[j-1];
    }

  /*
  printf("dcheb_2D... fit_deg= %d degy = %d degxy = %d \n", nl, degy, degxy);
  */
  return;
}

#ifdef __STDC__
  void dpoly_2D( double x1, double x2, double pl[], int nl)
#else
  void dpoly_2D(x1,x2,pl,nl)
  int nl;
  double x1,x2,pl[];
#endif  
{
  /* standard polynom in x1 (nl) && in x2 (degy) and 
     mixed polynom x1*x2 (degxy) */
  int j;

  
  pl[1]=1.0;
  if (nl > 1) 
    {
      for (j=2;j<=nl-(degy+degxy);j++)  /*polynom in x*/
	pl[j]=x1*pl[j-1];

      pl[nl-(degy+degxy)+1] = x2;       /*polynom in y*/
      for (j=nl-(degy+degxy)+2;j<=(nl-degxy);j++)
	pl[j] = x2*pl[j-1];

      pl[nl-degxy+1] = x1*x2;           /*mixed polynom in x & y*/  
      for (j=nl-degxy+2;j<=nl;j++)
	pl[j] = x2*pl[j-1];
    }
  return;
}





/************************************************************

 * median(): efficient search of the median.
 *
 ************************************************************/

#ifdef __STDC__
float 
median (float x[], int n)
#else
float 
median (x, n)
     float x[];
     int n;
#endif
{
  int n2, n2p;
#ifdef __STDC__
  void piksrt (int, float[]);
#else
  void piksrt ();
#endif

  piksrt (n, x);
  n2p = (n2 = n / 2) + 1;
  return ((n % 2) ? x[n2p] : 0.5 * (x[n2] + x[n2p]));
}


/************************************************************

 * piksrt(): sorting by straight insertion.
 *
 ************************************************************/

#ifdef __STDC__
void 
piksrt (int n, float arr[])
#else
void 
piksrt (n, arr)
     float arr[];
     int n;
#endif

{
  int i, j;
  float a;

  for (j = 2; j <= n; j++)
    {
      a = arr[j];
      i = j - 1;
      while (i > 0 && arr[i] > a)
	{
	  arr[i + 1] = arr[i];
	  i--;
	}
      arr[i + 1] = a;
    }
}

/************************************************************

median2(): efficient search of the median. (Thomas)

select (NR p 342) is much faster then piksrt! 
 
*************************************************************/

#ifdef __STDC__
float 
median2 (float x[], int n)
#else
float 
median2 (x, n)
     float x[];
     int n;
#endif
{
  int n2, n2p;
  float tmp;

  n2p = (n2 = n / 2) + 1;
  tmp = nr_select(n2p,n,x);
  return ((n % 2) ? tmp : 0.5 * (tmp + nr_select((n2p-1),(n2p-1),x)));
}


/************************************************************

nr_select(): select the k largest element of n elements in arr[]

NR p342 
************************************************************/
#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;


#ifdef __STDC__
float nr_select(unsigned long k, unsigned long n, float arr[])
#else
float nr_select( k, n, arr )
     unsigned long k,n;
     float arr[];
#endif
{
  unsigned long i,ir,j,l,mid;
  float a,temp;

  l=1;
  ir=n;
  for (;;) {
    if (ir <= l+1) {
      if (ir == l+1 && arr[ir] < arr[l]) {
	SWAP(arr[l],arr[ir])
	  }
      return arr[k];
    } else {
      mid=(l+ir) >> 1;
      SWAP(arr[mid],arr[l+1])
	if (arr[l+1] > arr[ir]) {
	  SWAP(arr[l+1],arr[ir])
	    }
      if (arr[l] > arr[ir]) {
	SWAP(arr[l],arr[ir])
	  }
      if (arr[l+1] > arr[l]) {
	SWAP(arr[l+1],arr[l])
	  }
      i=l+1;
      j=ir;
      a=arr[l];
      for (;;) {
	do i++; while (arr[i] < a);
	do j--; while (arr[j] > a);
	if (j < i) break;
	SWAP(arr[i],arr[j])
	  }
      arr[l]=arr[j];
      arr[j]=a;
      if (j >= k) ir=j-1;
      if (j <= k) l=i;
    }
  }
}

#undef SWAP

/************************************************************

median2(): efficient search of the median. (Thomas)

select (NR p 342) is much faster then piksrt! 
 
*************************************************************/

#ifdef __STDC__
double
dmedian2 (double x[], int n)
#else
double
dmedian2 (x, n)
     double x[];
     int n;
#endif
{
  int n2, n2p;
  double tmp;

  n2p = (n2 = n / 2) + 1;
  tmp = dnr_select(n2p,n,x);
  return ((n % 2) ? tmp : 0.5 * (tmp + dnr_select((n2p-1),(n2p-1),x)));
}


/************************************************************

nr_select(): select the k largest element of n elements in arr[]

NR p342 
************************************************************/
#define SWAP(a,b) temp=(a);(a)=(b);(b)=temp;


#ifdef __STDC__
double dnr_select(unsigned long k, unsigned long n, double arr[])
#else
double dnr_select( k, n, arr )
     unsigned long k,n;
     double arr[];
#endif
{
  unsigned long i,ir,j,l,mid;
  double a,temp;

  l=1;
  ir=n;
  for (;;) {
    if (ir <= l+1) {
      if (ir == l+1 && arr[ir] < arr[l]) {
	SWAP(arr[l],arr[ir])
	  }
      return arr[k];
    } else {
      mid=(l+ir) >> 1;
      SWAP(arr[mid],arr[l+1])
	if (arr[l+1] > arr[ir]) {
	  SWAP(arr[l+1],arr[ir])
	    }
      if (arr[l] > arr[ir]) {
	SWAP(arr[l],arr[ir])
	  }
      if (arr[l+1] > arr[l]) {
	SWAP(arr[l+1],arr[l])
	  }
      i=l+1;
      j=ir;
      a=arr[l];
      for (;;) {
	do i++; while (arr[i] < a);
	do j--; while (arr[j] > a);
	if (j < i) break;
	SWAP(arr[i],arr[j])
	  }
      arr[l]=arr[j];
      arr[j]=a;
      if (j >= k) ir=j-1;
      if (j <= k) l=i;
    }
  }
}

#undef SWAP



/*

  almost the same as select of the NR but some minor modifications by
  someone else

 */

#ifdef __STDC__
float
select_pos (int k, int n, float arr[])
#else
float
select_pos (k, n, arr)
     int k, n;
     float arr[];
#endif

/*      select the kth smallest value in arr[1...n], NR, pg. 342 */

#define SWAPD(a, b) 	{float temp = (a); (a) = (b); (b) = temp; }

{

  int i, ir, j, l, mid;
/*  double a, temp; */
  float a; 

  l = 1;
  ir = n;
  for (;;)
    {
      if (ir <= l + 1)
	{
	  if (ir == l + 1 && arr[ir] < arr[l])
	    {
	      SWAPD (arr[l], arr[ir])
	    }
	  return arr[k];
	}
      else
	{
	  mid = (l + ir) >> 1;
	  SWAPD (arr[mid], arr[l + 1])
	    if (arr[l + 1] > arr[ir])
	    {
	      SWAPD (arr[l + 1], arr[ir])
	    }
	  if (arr[l] > arr[ir])
	    {
	      SWAPD (arr[l], arr[ir])
	    }
	  if (arr[l + 1] > arr[l])
	    {
	      SWAPD (arr[l + 1], arr[l])
	    }
	  i = l + 1;
	  j = ir;
	  a = arr[l];
	  for (;;)
	    {
	      do
		i++;
	      while (arr[i] < a);
	      do
		j--;
	      while (arr[j] > a);
	      if (j < i)
		break;
	      SWAPD (arr[i], arr[j])
	    }
	  arr[l] = arr[j];
	  arr[j] = a;
	  if (j >= k)
	    ir = j - 1;
	  if (j <= k)
	    l = i;
	}
    }
}
#undef SWAPD
#undef SWAP
/*---------------------------------------------------------------------------*/
#ifdef __STDC__               /*same as heap_median*/
  float
  pik_median (int n, float arr[])
#else
  float
  pik_median ( n, arr )
  float arr[];
  int   n;
#endif

/*              find median of distribution (== heap_median)       */

{
  int i, j, index;
  /*  float a, b[101];*/
  float a, *b, buff;
  b = vector(0,n+1);

  for (j = 0; j < n; j++)
    b[j] = arr[j];

  for (j = 1; j < n; j++)
    {
      a = b[j];
      i = j - 1;
      while (i >= 0 && b[i] > a)
	{
	  b[i + 1] = b[i];
	  i--;
	}
      b[i + 1] = a;
    }
  index = (n - 1) / 2;
  buff = b[index];
  free_vector (b,0,n+1);
  return(buff);
}
/*---------------------------------------------------------------------------*/
#ifdef __STDC__               /*same as heap_median*/
  double
  dpik_median (int n, double arr[])
#else
  double
  dpik_median ( n, arr )
  double arr[];
  int   n;
#endif

/*              find median of distribution (== heap_median)       */

{
  int i, j, index;
  /*  float a, b[101];*/
  double a, *b, buff;
  b = dvector(0,n+1);

  for (j = 0; j < n; j++)
    b[j] = arr[j];

  for (j = 1; j < n; j++)
    {
      a = b[j];
      i = j - 1;
      while (i >= 0 && b[i] > a)
	{
	  b[i + 1] = b[i];
	  i--;
	}
      b[i + 1] = a;
    }
  index = (n - 1) / 2;
  buff = b[index];
  free_dvector (b,0,n+1);
  return(buff);
}

#ifdef __STDC__                /* same as pik_median */
float
heap_median (int n, float arr[])
#else
float 
heap_median (n, arr)
  float arr[];
  int   n;
#endif
/*              find median of distribution (heapsort)           */
{
  int i, j, index;
  /*float a, b[101];*/ 
  float a, *b, buff;
  b = vector(0,n+1);

  for (j = 0; j < n; j++)
    b[j] = arr[j];

  for (j = 1; j < n; j++)
    {
      a = b[j];
      i = j - 1;
      while (i >= 0 && b[i] > a)
	{
	  b[i + 1] = b[i];
	  i--;
	}
      b[i + 1] = a;
    }
  index = (n - 1) / 2;
  buff = b[index];
  free_vector (b,0,n+1);
  return(buff);
}



#ifdef __STDC__
  int f2i(float x)
#else
  int f2i(x)
    float x;
#endif
{
  int ix;
  ix = (int) (x +0.5);

  return ix;
}




#define EPS  0.00003
/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
dfit_ech (double inimage[], double ximage[], double yimage[], int npixi, 
	  double outimage[], double ximageo[],double yimageo[], int npixo, 
	  int fit_deg, double a[], int *reset,int debug,int ifunc)
#else
void
dfit_ech ( inimage, ximage, yimage, npixi, outimage, ximageo, yimageo, npixo, 
	  fit_deg, a,reset,debug,ifunc)
  double inimage[], ximage[], yimage[], outimage[], ximageo[], yimageo[];
  int npixi, npixo, fit_deg, *reset,debug,ifunc;
  double a[];
#endif
{
  /*
    fit a global dispersion relation through all echelle orders like
    Hensberge and Verschuren wrote in the ESO Messenger
    
    Thomas Szeifert    -    LSW Heidelberg    -    6 Juli 1997
 
    */

  double x1out,x2out;
  double *x1in,*x2in, *yin;
  int i,nfit=fit_deg,ncoefs=fit_deg,iter=0;

  double *sig,**covar,**alpha,alamda=(double)-1.0,chisq=0,ochisq=0;
  int *ia;

  x1in = dvector (1, npixi);
  x2in = dvector (1, npixi);
  yin = dvector (1, npixi);

  for (i = 1; i <= npixi; i++)
    {
      x1in[i] = (double) ximage[i-1];
      x2in[i] = (double) yimage[i-1];
      yin[i] = (double) inimage[i-1];
    }




  /* initialize first guess parameters of the echfit: */
  if (*reset == 1){

    if (ifunc != 1){
      a[1] = (double) 3625.0; 
      a[2] = (double) -36.5;
      a[3] = 0.; 
      a[4] = (double) -0.057;
      a[5] = (double) .25;
      for (i = 6; i <= fit_deg ; i++)
	a[i] = (double) 0.0;
    }
    else{
      a[1] = (double) 3625.0; 
      a[2] = (double) -36.5;
      a[3] = 0.; 
      a[4] = (double) -0.057;
      a[5] = 0.;
      a[6] = (double) .25;
      for (i = 7; i <= fit_deg ; i++)
	a[i] = (double) 0.0;
      
    }
    *reset = 0;
  }
      

  sig = dvector( 1, npixi );
  ia = ivector( 1, nfit );
  covar = dmatrix( 1, nfit, 1, nfit );
  alpha = dmatrix( 1, nfit, 1, nfit );
  

  for ( i = 1; i <= npixi; i++ )
    sig[i] = (double) 1.0;
  for ( i = 1; i <= nfit; i++ )
    ia[i] = 1;



  if (debug >= 2){
    printf("   ...dfit_ech: guess values\n");
    for (i = 1; i <= nfit; i++)
      printf("  %15.7f",a[i]);
    printf("\n");
  }

  if (ifunc != 1)
    mrqmin2D( x1in, x2in, yin, sig, npixi, a, ia, ncoefs, covar, alpha,
	      &chisq, dechfit, &alamda );
  else
    mrqmin2D( x1in, x2in, yin, sig, npixi, a, ia, ncoefs, covar, alpha,
	      &chisq, dechfit2, &alamda );

  if (debug >= 1)
    printf("   ...dfit_ech: %2d iterations %7d lines ==> chi = %9.7f\n", iter, npixi, sqrt(chisq));
  
  if (debug >= 2){
    for (i = 1; i <= nfit; i++)
      printf("  %15.7f",a[i]);
    printf("\n");
  }



  do {
    iter++;
    ochisq = chisq;
    
    if (ifunc != 1)
      mrqmin2D( x1in, x2in, yin, sig, npixi, a, ia, ncoefs, covar, alpha,
		&chisq, dechfit, &alamda );
    else
      mrqmin2D( x1in, x2in, yin, sig, npixi, a, ia, ncoefs, covar, alpha,
		&chisq, dechfit2, &alamda );

    if (debug >= 1)
      printf("   ...dfit_ech: %2d iterations %7d lines ==> chi = %9.7f\n", iter, npixi, sqrt(chisq));
    
    if (debug >= 2){
      for (i = 1; i <= nfit; i++)
	printf("  %15.7f",a[i]);
      printf("\n");
    }
    
  } while ( fabs(ochisq - chisq) / chisq > EPS || iter <= 3) ; 



  alamda = 0.;  /* To de-allocate memory */

  if (ifunc != 1)
    mrqmin2D( x1in, x2in, yin, sig, npixi, a, ia, ncoefs, covar, alpha,
	    &chisq, dechfit, &alamda );
  else
    mrqmin2D( x1in, x2in, yin, sig, npixi, a, ia, ncoefs, covar, alpha,
	    &chisq, dechfit2, &alamda );


  if (debug == 0)
    printf("   ...dfit_ech: %2d iterations %7d lines ==> chi = %9.7f\n",
	   iter, npixi, sqrt(chisq));

  if (debug >= 2){
    for (i = 1; i <= fit_deg; i++)
      printf("  %15.7f",a[i]);
    printf("\n");
  }



  free_dvector( sig, 1, npixi );
  free_ivector( ia, 1, nfit );
  free_dmatrix( covar, 1, nfit, 1, nfit );
  free_dmatrix( alpha, 1, nfit, 1, nfit );



  for (i = 0; i < npixo; i++)
    {
      x1out = (double) ximageo[i];
      x2out = (double) yimageo[i];
      if (ifunc != 1)
	outimage[i] = eval_dech (x1out, x2out, a, fit_deg);
      else
	outimage[i] = eval_dech2 (x1out, x2out, a, fit_deg);
    }

  free_dvector(x1in, 1, npixi);
  free_dvector(x2in, 1, npixi);
  free_dvector(yin, 1, npixi);


}


#ifdef __STDC__
double 
eval_dech (double x1, double x2, double a[], int n)
#else
double 
eval_dech (x1,x2, a, n)
     double x1,x2, a[];
     int n;
#endif
{
  double y,deins=1.0;
  int i;

  y = a[1] + a[2]*x2 + a[3]*x2*x1; 

  for (i = 5 ; i <= n ; i++)
    y = y + a[i]*pow(x1,(double)(i-4));
    
  y = y / (deins + a[4]*x2);

  return (y);
}


#ifdef __STDC__
double 
eval_dech2 (double x1, double x2, double a[], int n)
#else
double 
eval_dech2 (x1,x2, a, n)
     double x1,x2, a[];
     int n;
#endif
{
  double y,deins=1.0;
  int i;

  y = a[1] + a[2]*x2 + a[3]*x2*x1 + a[5]*x2*x2*x2; 

  for (i = 6 ; i <= n ; i++)
    y = y + a[i]*pow(x1,(double)(i-5));
    
  y = y / (deins + a[4]*x2);

  return (y);
}

#ifdef __STDC__
void 
dechfit (double x1, double x2, double a[], double *y , double dyda[],int n)
#else
void 
dechfit (x1, x2, a, y, dyda,n)
     double x1, x2 , *y , a[] , dyda[];
     int n;
#endif
{
  int i;
  double fac,fact,deins=1.0;

  fac = deins/(deins+a[4]*x2);

  fact = a[1] + a[2]*x2 + a[3]*x2*x1;

  for (i = 5; i <= n; i++)
    fact = fact + a[i]*pow(x1,(double)(i-4));

  dyda[1] = (double) 1.0;
  dyda[2] = x2;
  dyda[3] = x1*x2;

  for (i = 1 ; i <= 3 ; i++)
    dyda[i] = dyda[i]*fac;

  for (i = 5 ; i <= n; i++)
    dyda[i] = pow(x1,(double)(i-4))*fac;
  
  dyda[4] = x2 * fac*fac * fact * (-1.0);

  *y = fac*fact;
}


#ifdef __STDC__
void 
dechfit2 (double x1, double x2, double a[], double *y , double dyda[],int n)
#else
void 
dechfit2 (x1, x2, a, y, dyda,n)
     double x1, x2 , *y , a[] , dyda[];
     int n;
#endif
{
  int i;
  double fac,fact,deins=1.0;

  fac = deins/(deins+a[4]*x2);

  fact = a[1] + a[2]*x2 + a[3]*x2*x1 + a[5]*x2*x2*x2;

  for (i = 6; i <= n; i++)
    fact = fact + a[i]*pow(x1,(double)(i-5));

  dyda[1] = (double) 1.0;
  dyda[2] = x2;
  dyda[3] = x1*x2;
  dyda[5] = x2*x2*x2;
  for (i = 1 ; i <= 3 ; i++)
    dyda[i] = dyda[i]*fac;

  for (i = 6 ; i <= n; i++)
    dyda[i] = pow(x1,(double)(i-5))*fac;
  
  dyda[4] = x2 * fac*fac * fact * (-1.0);

  *y = fac*fact;
}

#undef EPS








/********************************************************************

 * polint(): polynomial interpolation or extrapolation from N input
 *           points, using the Neville's algorithm.
 *
 ********************************************************************/
#ifdef __STDC__
void 
polint (float xa[], float ya[], int n, float x, float *y, float *dy)
#else
void 
polint (xa, ya, n, x, y, dy)
     float xa[], ya[], x, *y, *dy;
     int n;
#endif

{
  int i, m, ns = 1;
  float den, dif, dift, ho, hp, w;
  float *c, *d;

  dif = fabs (x - xa[1]);
  c = fvector (1, n);
  d = fvector (1, n);
  for (i = 1; i <= n; i++)
    {				/* find the index ns of the closest */
      if ((dift = fabs (x - xa[i])) < dif)
	{			/* table entry. */
	  ns = i;
	  dif = dift;
	}
      c[i] = ya[i];		/* initialize the tableau of c's and d's */
      d[i] = ya[i];
    }
  *y = ya[ns--];		/* initial aprox. of y */
  for (m = 1; m < n; m++)
    {
      for (i = 1; i <= n - m; i++)
	{
	  ho = xa[i] - x;
	  hp = xa[i + m] - x;
	  w = c[i + 1] - d[i];
	  if ((den = ho - hp) == 0.0)
	    {
	      nrerror ("Error in polynomial interpolation");
	      return;
	    }
	  den = w / den;
	  d[i] = hp * den;
	  c[i] = ho * den;
	}
      *y += (*dy = (2 * ns < (n - m) ? c[ns + 1] : d[ns--]));
    }
  free_fvector (c, 1, n);
  free_fvector (d, 1, n);
}

/********************************************************************

 * dpolint(): polynomial interpolation or extrapolation from N input
 *           points, using the Neville's algorithm.
 *
 ********************************************************************/
#ifdef __STDC__
void 
dpolint (double xa[], double ya[], int n, double x, double *y, double *dy)
#else
void 
dpolint (xa, ya, n, x, y, dy)
     double xa[], ya[], x, *y, *dy;
     int n;
#endif

{
  int i, m, ns = 1;
  double den, dif, dift, ho, hp, w;
  double *c, *d;

  dif = fabs (x - xa[1]);
  c = dvector (1, n);
  d = dvector (1, n);
  for (i = 1; i <= n; i++)
    {				/* find the index ns of the closest */
      if ((dift = fabs (x - xa[i])) < dif)
	{			/* table entry. */
	  ns = i;
	  dif = dift;
	}
      c[i] = ya[i];		/* initialize the tableau of c's and d's */
      d[i] = ya[i];
    }
  *y = ya[ns--];		/* initial aprox. of y */
  for (m = 1; m < n; m++)
    {
      for (i = 1; i <= n - m; i++)
	{
	  ho = xa[i] - x;
	  hp = xa[i + m] - x;
	  w = c[i + 1] - d[i];
	  if ((den = ho - hp) == 0.0)
	    {
	      nrerror ("Error in polynomial interpolation");
	      return;
	    }
	  den = w / den;
	  d[i] = hp * den;
	  c[i] = ho * den;
	}
      *y += (*dy = (2 * ns < (n - m) ? c[ns + 1] : d[ns--]));
    }
  free_dvector (c, 1, n);
  free_dvector (d, 1, n);

  return;
}


#ifdef __STDC__
void 
dsvdcmp (double **a, int m, int n, double w[], double **v)
#else
void 
dsvdcmp (a, m, n, w, v)
     double **a, w[], **v;
     int m, n;
#endif

{
#ifdef __STDC__
  double dpythag (double, double);
#else
  double dpythag ();
#endif

  int flag, i, its, j, jj, k, l, nm;
  double anorm, c, f, g, h, s, scale, x, y, z, *rv1;

  rv1 = dvector (1, n);
  g = scale = anorm = 0.0;
  for (i = 1; i <= n; i++)
    {
      l = i + 1;
      rv1[i] = scale * g;
      g = s = scale = 0.0;
      if (i <= m)
	{
	  for (k = i; k <= m; k++)
	    scale += fabs (a[k][i]);
/*	  if (scale)   */
	  if ((int) scale)   
	    {
	      for (k = i; k <= m; k++)
		{
		  a[k][i] /= scale;
		  s += a[k][i] * a[k][i];
		}
	      f = a[i][i];
	      g = -SIGN (sqrt (s), f);
	      h = f * g - s;
	      a[i][i] = f - g;
	      for (j = l; j <= n; j++)
		{
		  for (s = 0.0, k = i; k <= m; k++)
		    s += a[k][i] * a[k][j];
		  f = s / h;
		  for (k = i; k <= m; k++)
		    a[k][j] += f * a[k][i];
		}
	      for (k = i; k <= m; k++)
		a[k][i] *= scale;
	    }
	}
      w[i] = scale * g;
      g = s = scale = 0.0;
      if (i <= m && i != n)
	{
	  for (k = l; k <= n; k++)
	    scale += fabs (a[i][k]);
/*	  if (scale) */
	  if ((int) scale)   
	    {
	      for (k = l; k <= n; k++)
		{
		  a[i][k] /= scale;
		  s += a[i][k] * a[i][k];
		}
	      f = a[i][l];
	      g = -SIGN (sqrt (s), f);
	      h = f * g - s;
	      a[i][l] = f - g;
	      for (k = l; k <= n; k++)
		rv1[k] = a[i][k] / h;
	      for (j = l; j <= m; j++)
		{
		  for (s = 0.0, k = l; k <= n; k++)
		    s += a[j][k] * a[i][k];
		  for (k = l; k <= n; k++)
		    a[j][k] += s * rv1[k];
		}
	      for (k = l; k <= n; k++)
		a[i][k] *= scale;
	    }
	}
      anorm = DMAX (anorm, (fabs (w[i]) + fabs (rv1[i])));
    }
  for (i = n; i >= 1; i--)
    {
      if (i < n)
	{
/*	  if (g) */
	  if ((int) g)
	    {
	      for (j = l; j <= n; j++)
		v[j][i] = (a[i][j] / a[i][l]) / g;
	      for (j = l; j <= n; j++)
		{
		  for (s = 0.0, k = l; k <= n; k++)
		    s += a[i][k] * v[k][j];
		  for (k = l; k <= n; k++)
		    v[k][j] += s * v[k][i];
		}
	    }
	  for (j = l; j <= n; j++)
	    v[i][j] = v[j][i] = 0.0;
	}
      v[i][i] = 1.0;
      g = rv1[i];
      l = i;
    }
  for (i = IMIN (m, n); i >= 1; i--)
    {
      l = i + 1;
      g = w[i];
      for (j = l; j <= n; j++)
	a[i][j] = 0.0;
      if ((int) g)
	{
	  g = 1.0 / g;
	  for (j = l; j <= n; j++)
	    {
	      for (s = 0.0, k = l; k <= m; k++)
		s += a[k][i] * a[k][j];
	      f = (s / a[i][i]) * g;
	      for (k = i; k <= m; k++)
		a[k][j] += f * a[k][i];
	    }
	  for (j = i; j <= m; j++)
	    a[j][i] *= g;
	}
      else
	for (j = i; j <= m; j++)
	  a[j][i] = 0.0;
/*      ++a[i][i];  */
	a[i][i] = a[i][i]+1;
    }
  for (k = n; k >= 1; k--)
    {
      for (its = 1; its <= 30; its++)
	{
	  flag = 1;
	  for (l = k; l >= 1; l--)
	    {
	      nm = l - 1;
	      if ((double) (fabs (rv1[l]) + anorm) == anorm)
		{
		  flag = 0;
		  break;
		}
	      if ((double) (fabs (w[nm]) + anorm) == anorm)
		break;
	    }
	  if (flag)
	    {
	      c = 0.0;
	      s = 1.0;
	      for (i = l; i <= k; i++)
		{
		  f = s * rv1[i];
		  rv1[i] = c * rv1[i];
		  if ((double) (fabs (f) + anorm) == anorm)
		    break;
		  g = w[i];
		  h = dpythag (f, g);
		  w[i] = h;
		  h = 1.0 / h;
		  c = g * h;
		  s = -f * h;
		  for (j = 1; j <= m; j++)
		    {
		      y = a[j][nm];
		      z = a[j][i];
		      a[j][nm] = y * c + z * s;
		      a[j][i] = z * c - y * s;
		    }
		}
	    }
	  z = w[k];
	  if (l == k)
	    {
	      if (z < 0.0)
		{
		  w[k] = -z;
		  for (j = 1; j <= n; j++)
		    v[j][k] = -v[j][k];
		}
	      break;
	    }
	  if (its == 30)
	    nrerror ("no convergence in 30 dsvdcmp iterations");
	  x = w[l];
	  nm = k - 1;
	  y = w[nm];
	  g = rv1[nm];
	  h = rv1[k];
	  f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0 * h * y);
	  g = dpythag (f, 1.0);
	  f = ((x - z) * (x + z) + h * ((y / (f + SIGN (g, f))) - h)) / x;
	  c = s = 1.0;
	  for (j = l; j <= nm; j++)
	    {
	      i = j + 1;
	      g = rv1[i];
	      y = w[i];
	      h = s * g;
	      g = c * g;
	      z = dpythag (f, h);
	      rv1[j] = z;
	      c = f / z;
	      s = h / z;
	      f = x * c + g * s;
	      g = g * c - x * s;
	      h = y * s;
	      y *= c;
	      for (jj = 1; jj <= n; jj++)
		{
		  x = v[jj][j];
		  z = v[jj][i];
		  v[jj][j] = x * c + z * s;
		  v[jj][i] = z * c - x * s;
		}
	      z = dpythag (f, h);
	      w[j] = z;
	      if ((int) z)
		{
		  z = 1.0 / z;
		  c = f * z;
		  s = h * z;
		}
	      f = c * g + s * y;
	      x = c * y - s * g;
	      for (jj = 1; jj <= m; jj++)
		{
		  y = a[jj][j];
		  z = a[jj][i];
		  a[jj][j] = y * c + z * s;
		  a[jj][i] = z * c - y * s;
		}
	    }
	  rv1[l] = 0.0;
	  rv1[k] = f;
	  w[k] = x;
	}
    }
  free_dvector (rv1, 1, n);
}


#ifdef __STDC__
void 
dsvbksb (double **u, double w[], double **v, int m,
	 int n, double b[], double x[])
#else
void 
dsvbksb (u, w, v, m, n, b, x)
     double **u, w[], **v, b[], x[];
     int m, n;
#endif

{

  int jj, j, i;
  double s, *tmp;

  tmp = dvector (1, n);
  for (j = 1; j <= n; j++)
    {
      s = 0.0;
      if ((int) w[j])
	{
	  for (i = 1; i <= m; i++)
	    s += u[i][j] * b[i];
	  s /= w[j];
	}
      tmp[j] = s;
    }
  for (j = 1; j <= n; j++)
    {
      s = 0.0;
      for (jj = 1; jj <= n; jj++)
	s += v[j][jj] * tmp[jj];
      x[j] = s;
    }
  free_dvector (tmp, 1, n);
}

#ifdef __STDC__
double 
dpythag (double a, double b)
#else
double 
dpythag (a, b)
     double a, b;
#endif
{
  double absa, absb;
  absa = fabs (a);
  absb = fabs (b);
  if (absa > absb)
    return absa * sqrt (1.0 + DSQR (absb / absa));
  else
    return (absb == 0.0 ? 0.0 : absb * sqrt (1.0 + DSQR (absa / absb)));
}
