/*
  .AUTHOR       Cristian Levin - ESO La Silla


  rebin.c              Thomas Szeifert  -   LSW Heidelberg 080797
  

  functions in rebin.c:

  void rebin( double *win, double *wout, float *yin, float *yout, 
    int nin, int nout, double step, int Method)

  void closest_index( double rx, double *win, double *wout, int nin, 
    int nout, double *x, int *ix , double step)

------------------------------------------------------------*/

#include <math.h>
#include <proto_fors.h>

#ifdef __STDC__
int dfinterpol(double x1[], float y1[], int n1, 
              double x2[], float y2[], int n2, int meth);
#else
int dfinterpol();
#endif


#ifdef __STDC__
void rebin( double *win, double *wout, float *yin, float *yout, 
           int nin, int nout, double step, int Method)
#else
void rebin( win, wout, yin, yout, nin, nout, step,Method)
  double *win,*wout,step;
  float  *yin,*yout;
  int    nin,nout,Method;
#endif
{
    int i, j, k, idx, ix = 0;
    int j1, j2;
    double x1, x2;
    /* double dx; */
    double a, b, c, y;
    double d, dd, ddd;
    double rx = -0.5;
    double yval[3];

    int istart = 2;
    int Imin = MAXINTENS;
    int Imax = MININTENS;

    if ( Method == SPLINE ) {	/* No flux conservation ! */
      /*
	for ( i = 0; i < nout; i++ ) {
	yout[i] = dsplint_lewin( wout[i], win-1, yin-1, nin, &istart );
	if ( yout[i] < Imin ) Imin = yout[i];
	if ( yout[i] > Imax ) Imax = yout[i];
	}
	*/
      dfinterpol(win,yin,nin,wout,yout,nout,SPLINE);
      return;
    }

    for ( i = 0; i < nout; i++ ) yout[i] = (float) 0.0;

    closest_index( rx, win, wout, nin, nout, &x1, &ix, step );
    j1 = NINT(x1);
    
    for ( i = 0; i < nout; i++ ) {
      rx += 1.0;
      closest_index( rx, win, wout, nin, nout, &x2, &ix, step);
      j2 = NINT(x2);
      
      /* dx = x2 - x1; */
      d = 0.0;
      if ( Method == QUADRATIC ) {
	if ( i == 0 ) {
	  for ( k = 0; k < 3; k++ ) {
	    idx = j1 + k - 1;
	    if ( idx < 0 || idx > nin - 1 )
	      yval[k] = 0.0;
	    else
	      yval[k] = yin[idx];
	  }
	  a = 0.5*(yval[0] + yval[2]);
	  b = 0.5*(a - yval[0]);
	  c = (13.0*yval[1] - a) / 12.0;
	  a = (a - yval[1]) / 3.0;
	  y = x1 - j1;
	  dd = ((a*y + b)*y + c)*y - 0.25*b + 0.125*a + 0.5*c;
	}
	
	for ( k = 0; k < 3; k++ ) {
	  idx = j2 + k - 1;
	  if ( idx < 0 || idx > nin - 1 )
	    yval[k] = 0.0;
	  else
	    yval[k] = yin[idx];
	}
	a = 0.5*(yval[0] + yval[2]);
	b = 0.5*(a - yval[0]);
	c = 1.083333333333*yval[1] - 0.083333333333*a;
	a = 0.333333333333*(a - yval[1]);
	y = x2 - j2;
	d -= dd;
	dd = ((a*y + b)*y + c)*y - 0.25*b;
	ddd = 0.125*a + 0.5*c;
	d += dd - ddd;
	dd += ddd;
      }
      else if ( Method == LINEAR ) {
	if ( i == 0 ) {
	  if ( j1 < 0 || j1 > nin - 1 )
	    dd = 0.0;
	  else
	    dd = ((((double) j1) - x1) - 0.5) * (double) yin[j1];
	}
	
	if ( j2 < 0 || j2 > nin - 1 )
	  ddd = 0.0;
	else
	  ddd = (double) yin[j2];
	
	d += dd;
	dd = ( ((double) j2) - x2 - 0.5 ) * ddd;

	d = (d - dd) - ddd;
	
      }

      
      for ( j = j1; j <= j2; j++ )
	if ( j >= 0 && j < nin )
	  d += (double) yin[j];

      yout[i] = (float) d;
      x1 = x2;
      j1 = j2;
      
      if ( yout[i] < Imin ) Imin = yout[i];
      if ( yout[i] > Imax ) Imax = yout[i];
      
    } /* end of loop */
}


#ifdef __STDC__
void closest_index( double rx, double *win, double *wout, int nin, 
                    int nout, double *x, int *ix , double step)
#else
void closest_index( rx, win, wout, nin, nout, x, ix, step )
  double rx,*win,*wout,*x,step;
  int    nin,nout,*ix;
#endif

{
  double wave, waux;
  int irx, ix2;


  irx = (int) rx; /* integer part of rx */

  if ( irx < 0 )
    irx = 0;
  else if ( irx >= nout - 1 )
    irx = nout - 2;
  
  wave = wout[irx] + (rx - (double) irx) * step;
  
  waux = win[*ix];
  if ( waux > wave ) { /* initial guess is too high */
    while ( *ix > 0 && waux > wave ) {
      (*ix)--;
      waux = win[*ix];
    }
    ix2 = *ix + 1;
  }
  else {		/* initial guess is too low */
    while ( *ix < nin - 1 && waux < wave ) {
      (*ix)++;
      waux = win[*ix];
    }
    ix2 = *ix - 1;
  }
  
  /* interpolate to find X */
  *x = ((double) *ix) + ((wave - waux) / (win[ix2]-waux)) * 
    (double) (ix2 - *ix);
  
  return;
}






#ifdef __STDC__
int dfinterpol(double dx1[], float y1[], int n1, 
	       double dx2[], float y2[], int n2, int meth)
#else
int dfinterpol(dx1, y1, n1, dx2, y2, n2, meth)
     int n1,n2, meth;
     float y1[],y2[];
     double dx1[],dx2[];
#endif
{
  int i,istart=0,j;
  double *dbuff,*dbuffx1,*dbuffy1,dy2;
  double stdev;


  dbuffx1 = dvector(1,n1);
  dbuffy1 = dvector(1,n1);


  for (i = 1 ; i <= n1 ; i++)
    {
      dbuffy1[i] = (double) y1[i-1];
      dbuffx1[i] = dx1[i-1];
    }


  if (meth == SPLINE)
    {
      dbuff = dvector(1,n1);

      dspline(dbuffx1, dbuffy1, n1, 1.0e30, 1.0e30, dbuff);
      for (i = 0 ; i < n2 ; i++){
	if (dx2[i] > dx1[0] && dx2[i] < dx1[n1-1]){
	  dsplint(dbuffx1, dbuffy1, dbuff, n1, dx2[i], &dy2);
	  y2[i] = (float) dy2;
	}
	else {
	  y2[i] = (float) 0.0;
	}
      }

      free_dvector(dbuff,1,n1);
    }

  free_dvector(dbuffx1,1,n1);
  free_dvector(dbuffy1,1,n1);

  return 0;
}






