/*proto_ipolmagni.h                  by Thomas Szeifert LSW Heidelberg   10/96 

This program maps data for a given part of the frame (a slitlet of a 
polarimetric mask) and computes:
   -> sky
   -> point spread function
   -> flux
for all stars which were selected in the input table. We search for stars that
are in a given slitlet defined by a second input table and loop through this
table. 

We add the pixel value in the innermost ring    r < inputi[0] 
and add the fit value within the second ring    r < inputi[1] 
and add the fit value for all pixels outside the mask and for multiple stars

From my point of view the single Moffat profile fits well enough for 
polarimetry, were we don't need absolute fluxes. If you are going to use
the program for absolute photometry you will have to change the program 
in the following way:

|  take a sum of two Moffat functions with different exponents beta1 & beta2
|  ->  funcs = f * ( (1 - epsilon)/ arg**beta1 + epsilon / arg**beta2)
|  do some analysis to get the first deviates analytically and
|  see function moffat for more explanations. You have to change the fitparam
|  array and ncoef (fcoef = 6 but not 4) because of the two more coeficients!

see R. Racine 1996 - PASP 108, 699 for PSF fitting and more advanced models
*/



/* WARNING:  these arrays,pointers and variables come from ipolmagni.c*/

extern double Start[], Step[], dbeta[], skybuff;
extern float rparam[];
extern int Npix[];
extern int debug, reset, guess; 
/* debug can be 0,1 or 2; higher == more informations*/
extern int ncoef, psf, fcoef; 
/* Moffat needs fcoef = 4 fit parameters for the PSF if beta is free*/
/* fitting routines need this pointers*/
extern float *xsel,*ysel,*fsel;
extern char text[];

/* WARNING:  these were the arrays cited above */




/*is the pixel or another star in the ring?*/
extern int reject(
#ifdef __STDC__
float , float , float[] , float[] , int , int , int
#endif
); 


/*fraction of the pixel in the ring*/
extern float fraction(
#ifdef __STDC__
int,int,float,float,int
#endif
); 


/*calculate function and deviates -- beta free*/
extern void moffat(
#ifdef __STDC__
double ,double , double [], double *, double [] , int
#endif
);


/*calculate function and deviates -- fix beta = dparam = e.g. 4*/
extern void moffatbf(
#ifdef __STDC__
double ,double , double [], double *, double [] , int
#endif
);


/*calculate function and deviates -- (1-eps**2) F(beta=2) + eps**2F(beta=7) */
extern void moffatbf2(
#ifdef __STDC__
double ,double , double [], double *, double [] , int
#endif
);


/*calculate function and deviates -- call the specified function*/
extern void evalpsf(
#ifdef __STDC__
double ,double , double [], double * , double [], int
#endif
);


/*prepare IO for PSF-fit*/
extern void fit2D(
#ifdef __STDC__
double *, double *, double *, int, double *
#endif
);







