! +++++++++++++++++++++++++++++++++++++++++++++++++
!.COPYRIGHT    (C) 1995 Landessternwarte Heidelberg
!.IDENT        ipoloffset.prg
!.AUTHORS      Sabine Moehler (LSW)
!.KEYWORDS     Polarimetry
!  
!.PURPOSE      execute the command OFFSET/IPOL
!
!.VERSION      1.0 Creation 11/05/1995
! 
! OFFSET/IPOL ipol_table sc slit_tab off try
!
! -------------------------------------------------
!
! CROSSREF IPOL_TAB SC SLIT_TAB OFF TRY
!
DEFINE/PARAM P1 {IPOL_TAB} TABLE   "Enter input table:"
DEFINE/PARAM P2 {SC}       CHAR    "Column with input images?"
DEFINE/PARAM P3 {SLIT_TAB} CHAR    "Column with slitlet tables?"
DEFINE/PARAM P4 {OFF(1)},{OFF(2)} NUM "Crude offset between slitlets? (x,y)"
DEFINE/PARAM P5 F CHAR    "First try or All frames?"
!------------------------------------------------------------------------------
! Local keys ...
!-------------------------------------------------
! angle of retarder plate
DEFINE/LOCAL RET_ANG/I/1/1 0
! crude offsets between ordinary and extra-ordinary rays
DEFINE/LOCAL OFFSET/R/1/2 0.,0.
! limits of slitlets for extraction of subimages
DEFINE/LOCAL YSTART/D/1/1 0.
DEFINE/LOCAL YEND/D/1/1 0.
! ray type (O/E)
DEFINE/LOCAL TYP/C/1/1 F
! First try or All frames
DEFINE/LOCAL TRY/C/1/1 F
! For seeing
DEFINE/LOCAL SEE/R/1/2 0.,0.
!------------------------------------------------------------------------------
! ... for sorting ...
!------------------------------------------------------------------------------
DEFINE/LOCAL I/I/1/1 0
DEFINE/LOCAL II/I/1/1 0.
DEFINE/LOCAL J/I/1/1 0
DEFINE/LOCAL JJ/I/1/1 0.
DEFINE/LOCAL K/I/1/1 0.
DEFINE/LOCAL KK/I/1/1 0.
DEFINE/LOCAL L/I/1/1 1.
DEFINE/LOCAL LL/I/1/1 0.
DEFINE/LOCAL NUMROW/I/1/1 0.
!------------------------------------------------------------------------------
! ... for input frames/tables/angles
!------------------------------------------------------------------------------
DEFINE/LOCAL INIMA/C/1/60 " " ALL
DEFINE/LOCAL INSC/C/1/60 " " ALL
DEFINE/LOCAL INTAB/C/1/60 " " ALL
DEFINE/LOCAL TMP/C/1/60 " " ALL
DEFINE/LOCAL INANG/R/1/1 -9999.9
DEFINE/LOCAL ANG/I/1/1 -9999
WRITE/KEY TRY/C/1/1 {P5}
!------------------------------------------------------------------------------
! Check whether same flat or individual flats have been used for slitlet search
!------------------------------------------------------------------------------
COPY/DK {P1}.tbl TBLGETFF/C/1/1 GETFF
!------------------------------------------------------------------------------
! store crude offset estimate
!------------------------------------------------------------------------------
WRITE/KEY OFFSET/R/1/2 {P4}
!------------------------------------------------------------------------------
! Get number of rows in ipol_tab
!------------------------------------------------------------------------------
COPY/DK {P1}.tbl TBLCONTR/I/4/4 NUMROW
!------------------------------------------------------------------------------


! Get low and high row numbers for science frames in ipol_tab

COUNT/IPOL {P1}

! Create dummy table to determine slitlet offsets

CREA/TAB &y ? ? NULL

! Create new table to hold names of extracted frames, raw frames, input frames
! tables with dispersion coefficients:
!
CREATE/TAB {P1}EX ? ? NULL
CREATE/COL {P1}EX {SC} ? A40 C*40
CREATE/COL {P1}EX {P2} ? A40 C*40
CREATE/COL {P1}EX {SC_EX} ? A40 C*40
CREATE/COL {P1}EX :RET_ANG ? R*4 F8.2
CREATE/COL {P1}EX :SEE_X "pixel" R*4 F8.2
CREATE/COL {P1}EX :SEE_Y "pixel" R*4 F8.2
COPY/DD  {P1}.tbl TBLGETFF/C/1/1 {P1}EX.tbl TBLGETFF/C/1/1
WRITE/OUT "Table {P1}EX.tbl created to hold names of output frames"
SET/FORMAT I2



! Same flat for all frames == one slitlet table for all frames:

IF GETFF(1:1) .eq. "S" THEN
! Get slitlet tables:
  DO II = 1 {NUMROW}
     WRITE/KEY TMP {{P1},{P3},@{II}}
     IF TMP(1:1) .ne. " " THEN
       WRITE/KEY INTAB {TMP}
     ENDIF
  ENDDO
ENDIF


! LOOP:

DO J = 1 {N_SC}
! 	Individual flat for each frame == one slitlet table 
! 	for each frame
   IF GETFF(1:1) .eq. "I" THEN

! 		Get slitlet tables:
      DO II = SC_STRT({J}) SC_END({J})
    	 WRITE/KEY TMP {{P1},{P3},@{II}}
     	 IF TMP(1:1) .ne. " " THEN
       	    WRITE/KEY INTAB {TMP}
     	 ENDIF
      ENDDO
   ENDIF
!---------------------------------------------------
! Get science frames
!---------------------------------------------------
   DO II = SC_STRT({J}) SC_END({J})
      WRITE/KEY INIMA {{P1},{P2},@{II}}
	 WRITE/KEY INSC {{P1},{SC},@{II}}
    	 WRITE/KEY INANG {{P1},:RET_ANG,@{II}}
!   COPY/TK {P1} {P2} @{II} INIMA/C/1/60
    	 IF INIMA(1:1) .ne. " " THEN
      	    JJ = JJ+1

! mark objects in first frame
      	    IF JJ .LE. 1 THEN
               WRITE/OUT "************************************************ "
               WRITE/OUT "Mark only objects in slitlets with ordinary rays "
               WRITE/OUT "         (lower one of the pair)"
               WRITE/OUT "************************************************ "
               CLEAR/CHAN OVER
               LOAD {INIMA}
               DO LL = 1 {NSLIT}
          	  WRITE/KEY TYP {{INTAB},:RAY_TYP,@{LL}}

! mark only objects in ordinary ray slitlets 

          	  IF TYP(1:1) .eq. "O" THEN
            	     K = {{INTAB},:SLIT,@{LL}}
                     WRITE/OUT "********************************************* "
                     WRITE/OUT " Mark AT LEAST 3 reference objects in slit {K}"
                     WRITE/OUT "********************************************* "
            	     CENTER/GAUSS cursor &o
                     !SET/MIDAS OUTPUT=LOGONLY
                     STAT/TABLE &o :XFWHM
                     COPY/KD OUTPUTR/R/3/1 &o.tbl SEE/R/1/1
                     STAT/TABLE &o :YFWHM
                     COPY/KD OUTPUTR/R/3/1 &o.tbl SEE/R/2/1
                     !SET/MIDAS OUTPUT=YES

! store positions in reference table

                     COMP/KEY OUTPUTI = M$INDEXB(INIMA,".")-1
                     WRITE/KEY REFIMA/C/1/60 {INIMA(1:{OUTPUTI(1)})}

! store retarder plate angle of reference frame

                     WRITE/KEY REF_RET/R/1/1 {INANG}

! search objects in corresponding extra-ordinary ray 
! slitlets 

                     COPY/TAB &o REF{REFIMA}O{K}.tbl
                     COPY/TAB &o &e
                     COMP/TAB &e :XSTART = :XSTART+{OFFSET(1)}
                     COMP/TAB &e :XEND = :XEND+{OFFSET(1)}
                     COMP/TAB &e :YSTART = :YSTART+{OFFSET(2)}
                     COMP/TAB &e :YEND = :YEND+{OFFSET(2)}
                     !SET/MIDAS OUTPUT=LOGONLY
                     CENTER/GAUSS {INIMA},&e &e
                     STAT/TABLE &e :XFWHM
	             SEE(1) = outputr(3)
                     STAT/TABLE &e :YFWHM
	             SEE(2) = outputr(3)
	             COPY/KD SEE/R/1/2 &e.tbl SEE/R/1/2 	
                     !SET/MIDAS OUTPUT=YES

! extract subimages 		(ordinary rays)

                     YSTART = {{INTAB},:YSTART,@{LL}}
                     YEND = {{INTAB},:YEND,@{LL}}
                     EXTRACT/IMA O{K}{INIMA} = {INIMA}[<,{YSTART}:>,{YEND}]
                     COPY/DD &o.tbl SEE/R/1/2 O{K}{INIMA} SEE/R/1/2
                     WRITE/DESCR O{K}{INIMA} REFIMA/C/1/40 O{K}{REFIMA}

! 		     store names of extracted subimages

                     WRITE/TAB {P1}EX {SC} @{L} {INSC}
                     WRITE/TAB {P1}EX {P2} @{L} {INIMA}
                     WRITE/TAB {P1}EX :RET_ANG @{L} {INANG}
                     WRITE/TAB {P1}EX {SC_EX} @{L} O{K}{INIMA}
                     WRITE/TAB {P1}EX :SEE_X @{L} {SEE(1)}
                     WRITE/TAB {P1}EX :SEE_Y @{L} {SEE(2)}
                     L = {L}+1

! extract and align subimages 	(extra-ordinary rays)

                     LL = {LL}+1
            	     YSTART = {{INTAB},:YSTART,@{LL}}
                     YEND = {{INTAB},:YEND,@{LL}}
                     EXTRACT/IMA &e = {INIMA}[<,{YSTART}:>,{YEND}]
!                     SET/MIDAS OUTPUT=LOGONLY
                     ALIGN/IMAGE &e &o SHIFT ? yes
!                     SET/MIDAS OUTPUT=YES
write/out            "REBIN/ROTATE &e E{K}{INIMA} keyword O{K}{REFIMA} "
                     REBIN/ROTATE &e E{K}{INIMA} keyword  O{K}{REFIMA}.bdf YES
                     COPY/DD &e.tbl SEE/R/1/2 E{K}{INIMA} SEE/R/1/2
                     WRITE/DESCR E{K}{INIMA} REFIMA/C/1/40 O{K}{REFIMA}
!---------------------------------------------------
! store align coefficients
!---------------------------------------------------
                     WRITE/KEY ANG {INANG}
                     WRITE/KEY AL{LL}R{ANG}/D/1/9 0 all
                     COPY/KEY TRANSFRM/D/1/9 AL{LL}R{ANG}/D/1/9 
                     COPY/KD AL{LL}R{ANG}/D/1/9 {P1}.tbl AL{LL}R{ANG}/D/1/9
                     LL = {LL}-1
!---------------------------------------------------
! store names of extracted and aligned subimages
!---------------------------------------------------
                     WRITE/TAB {P1}EX {SC} @{L} {INSC}
                     WRITE/TAB {P1}EX {P2} @{L} {INIMA}
                     WRITE/TAB {P1}EX :RET_ANG @{L} {INANG}
                     WRITE/TAB {P1}EX {SC_EX} @{L} E{K}{INIMA}
                     COPY/DK E{K}{INIMA} SEE/R/1/2 SEE/R/1/2
                     WRITE/TAB {P1}EX :SEE_X @{L} {SEE(1)}
                     WRITE/TAB {P1}EX :SEE_Y @{L} {SEE(2)}
                     L = {L}+1
                  ENDIF
               ENDDO
            ELSE
!---------------------------------------------------
! all other frames
!---------------------------------------------------
               IF TRY(1:1) .eq. "F" GOTO end
               DO LL = 1 {NSLIT}
                  WRITE/KEY TYP {{INTAB},:RAY_TYP,@{LL}}
                  IF TYP(1:1) .eq. "O" THEN
!---------------------------------------------------
! "ordinary" slitlets: search objects
!---------------------------------------------------
                     K = {{INTAB},:SLIT,@{LL}}
                     !SET/MIDAS OUTPUT=LOGONLY
                     CENTER/GAUSS {INIMA},REF{REFIMA}O{K}.tbl &o
                     STAT/TABLE &o :XFWHM
		     SEE(1) = outputr(3)	
                     STAT/TABLE &o :YFWHM
		     SEE(2) = outputr(3)	
	             !SET/MIDAS OUTPUT=YES
		     COPY/KD SEE/R/1/2 &o.tbl SEE/R/1/2	

! extract and align subimages

            	     YSTART = {{INTAB},:YSTART,@{LL}}
                     YEND = {{INTAB},:YEND,@{LL}}
                     EXTRACT/IMA &o = {INIMA}[<,{YSTART}:>,{YEND}]
!                     SET/MIDAS OUTPUT=LOGONLY
                     ALIGN/IMAGE &o REF{REFIMA}O{K}.tbl SHIFT ? yes
!                     SET/MIDAS OUTPUT=YES
 write/out           "REBIN/ROTATE &o O{K}{INIMA} keyword O{K}{REFIMA} "
                     REBIN/ROTATE &o O{K}{INIMA} keyword O{K}{REFIMA}.bdf YES
                     COPY/DD &o.tbl SEE/R/1/2 O{K}{INIMA} SEE/R/1/2
                     WRITE/DESCR O{K}{INIMA} REFIMA/C/1/40 O{K}{REFIMA}

! store align coefficients

                     WRITE/KEY ANG {INANG}
                     WRITE/KEY  AL{K}R{ANG}/D/1/9 0 all
                     COPY/KEY TRANSFRM/D/1/9 AL{K}R{ANG}/D/1/9 
                     COPY/KD AL{K}R{ANG}/D/1/9 {P1}.tbl AL{K}R{ANG}/D/1/9

! store names of extracted and aligned subimages

                     WRITE/TAB {P1}EX {SC} @{L} {INSC}
                     WRITE/TAB {P1}EX {P2} @{L} {INIMA}
                     WRITE/TAB {P1}EX :RET_ANG @{L} {INANG}
                     WRITE/TAB {P1}EX {SC_EX} @{L} O{K}{INIMA}
                     COPY/DK O{K}{INIMA} SEE/R/1/2 SEE/R/1/2
                     WRITE/TAB {P1}EX :SEE_X @{L} {SEE(1)}
                     WRITE/TAB {P1}EX :SEE_Y @{L} {SEE(2)}
                     L = {L}+1
                  ELSE 

! "extra-ordinary" slitlets: search objects

                     KK = {{INTAB},:SLIT,@{LL}}
!                     K = {KK}-1
                     K = {KK}
                     COPY/TAB &o &e

! extract and align subimages

                     COMP/TAB &e :XSTART = :XSTART+{OFFSET(1)}
                     COMP/TAB &e :XEND = :XEND+{OFFSET(1)}
                     COMP/TAB &e :YSTART = :YSTART+{OFFSET(2)}
                     COMP/TAB &e :YEND = :YEND+{OFFSET(2)}
                     !SET/MIDAS OUTPUT=LOGONLY
                     CENTER/GAUSS {INIMA},&e &e
                     STAT/TABLE &e :XFWHM
                     COPY/KD OUTPUTR/R/3/1 &e.tbl SEE/R/1/1
                     STAT/TABLE &e :YFWHM
                     COPY/KD OUTPUTR/R/3/1 &e.tbl SEE/R/2/1
                     !SET/MIDAS OUTPUT=YES
                     YSTART = {{INTAB},:YSTART,@{LL}}
                     YEND = {{INTAB},:YEND,@{LL}}
                     EXTRACT/IMA &e = {INIMA}[<,{YSTART}:>,{YEND}]
!                     SET/MIDAS OUTPUT=LOGONLY
                     ALIGN/IMAGE &e REF{REFIMA}O{K}.tbl SHIFT ? yes
!                     SET/MIDAS OUTPUT=YES
 write/out          "REBIN/ROTATE &e E{K}{INIMA} keyword O{K}{REFIMA} "
                     REBIN/ROTATE &e E{K}{INIMA} keyword O{K}{REFIMA}.bdf YES
                     COPY/DD &e.tbl SEE/R/1/2 E{K}{INIMA} SEE/R/1/2
                     WRITE/DESCR E{K}{INIMA} REFIMA/C/1/40 O{K}{REFIMA}

! store align coefficients

                     WRITE/KEY ANG {INANG}
                     WRITE/KEY  AL{KK}R{ANG}/D/1/9 0 all
                     COPY/KEY TRANSFRM/D/1/9 AL{KK}R{ANG}/D/1/9 
                     COPY/KD AL{KK}R{ANG}/D/1/9 {P1}.tbl AL{KK}R{ANG}/D/1/9

! store names of extracted and aligned subimages

                     WRITE/TAB {P1}EX {SC} @{L} {INSC}
                     WRITE/TAB {P1}EX {P2} @{L} {INIMA}
                     WRITE/TAB {P1}EX :RET_ANG @{L} {INANG}
                     WRITE/TAB {P1}EX {SC_EX} @{L} E{K}{INIMA}
                     COPY/DK E{K}{INIMA} SEE/R/1/2 SEE/R/1/2
                     WRITE/TAB {P1}EX :SEE_X @{L} {SEE(1)}
                     WRITE/TAB {P1}EX :SEE_Y @{L} {SEE(2)}
                     L = {L}+1
                  ENDIF
               ENDDO
            ENDIF
         GOTO next
      ENDIF
   ENDDO
next:
ENDDO
end:








