/*ipolover.c        read frames with degree and angle of polarisation and
                    plot the data in given coordinate axes


		    by Thomas Szeifert & Otmar Stahl; LSW Heidelberg; 25.Feb.97
*/

#define MAXROW  9999
#define STRGLEN 80

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <midas_def.h>
#include <plot_def.h>
#include <mutil.h>
/*
  #include <proto_mutil.h>
*/

/* es geht anscheinend ohne die folgenden protos; falls man sie mitnimmt ist 
   einiges doppelt und anderes wird ueberschrieben - auch Systemparameter!

#include <agl.h>
#include <stdlib.h>
#include <osparms.h>
   */


int debug=0;           /* switch this on if there are problems */


#ifdef __STDC__
int  ipolover (float [], float[], int [], double[], 
	       double [], float [], float [], float, int);
void mapslit (int,float,float,int *,int *,float [],double[],double[],int[]);
int f2i(float);
#else
int  ipolover ();      /*Otmars overplot but with PCDATA() call*/ 
void mapslit ();       /*read the data from the disk*/
int f2i();
#endif

main ()
{
  char   ina[STRGLEN],inb[STRGLEN],plcmeta[STRGLEN],devnam[STRGLEN];
  int    actval,plmode,aid,bid,inpix[2],ibuff[2],istart,ipxl,iunit,inull,
    crash=1;
  float  wcframx[4], wcframy[4],*pol,*theta,scale;
  double dstart[2],dstep[2],dbuff[2];



  SCSPRO ("ipolplot");




  /*read the keywords and open the files on the disk*/

  SCKRDR("INPUTR", 1, 1, &actval, &scale, &iunit, &inull);

  SCKGETC ("IN_A", 1, STRGLEN, &actval, ina);
  SCKGETC ("IN_B", 1, STRGLEN, &actval, inb);
  SCFOPN(ina, D_R4_FORMAT, 0, F_IMA_TYPE, &aid);
  SCFOPN(inb, D_R4_FORMAT, 0, F_IMA_TYPE, &bid);

  SCDRDD(aid,"START",1,2,&actval,dstart,&iunit,&inull);
  SCDRDD(aid,"STEP",1,2,&actval,dstep,&iunit,&inull);
  SCDRDI(aid,"NPIX",1,2,&actval,inpix,&iunit,&inull);




  /*check if frames match in Start,Step and Npix*/ 

  SCDRDD(bid,"STEP",1,2,&actval,dbuff,&iunit,&inull);  
  if (abs((dstep[0]-dbuff[0])/dstep[0])<1.0e-6 && 
      abs((dstep[1]-dbuff[1])/dstep[1])<1.0e-6){ 
    SCDRDD(bid,"START",1,2,&actval,dbuff,&iunit,&inull); 
    if (abs((dstart[0]-dbuff[0])/dstep[0])<0.1 && 
	abs((dstart[1]-dbuff[1])/dstep[1])<0.1){ 
      SCDRDI(bid,"NPIX",1,2,&actval,ibuff,&iunit,&inull);
      if ( inpix[0] == ibuff[0] && inpix[1] == ibuff[1]) {
	printf("...degree: %s   angle: %s \n",ina,inb);
	crash = 0;
      }
    }
  }

  if (crash == 1){
    fprintf(stderr,"...ERROR: frames have different START,STEP or NPIX\n"); 
    SCSEPI ();
    exit(2);
  }




  /*read plot parameters and open device*/
  
  PCKRDR( "XAXIS", 4, &actval, wcframx );
  PCKRDR( "YAXIS", 4, &actval, wcframy );
  PCOPEN (" "," ", 1, &plmode);
  PCKRDC ("PLNAME", STRGLEN, &actval, plcmeta);
  PCKRDC ("DNAME", 60, &actval, devnam);




  /*do the work*/

  ipxl = inpix[0]*( f2i( (wcframy[1]-wcframy[0]) / ((float) dstep[1]) )  + 1);
  if (debug > 2)
    printf("ipxl = %d inpix[0] = %d wcframy[1] = %6.1f wcframy[0] = %6.1f  dstep[1] = %6.1f \n", ipxl, inpix[0], wcframy[1], wcframy[0], dstep[1]);
  

  pol   = (float *) osmmget(2 * (ipxl+1) * sizeof (float));
  theta = (float *) osmmget(2 * (ipxl+1) * sizeof (float));

  mapslit (aid,wcframy[0],wcframy[1],&istart,&ipxl,pol,dstart,dstep,inpix);
  mapslit (bid,wcframy[0],wcframy[1],&istart,&ipxl,theta,dstart,dstep,inpix);

  if (debug >= 1)
    printf ("start with ipolover  istart = %6d\n",istart);
  ipolover (pol,theta,inpix,dstart,dstep,wcframx,wcframy,scale,istart);

  osmmfree((char *) pol);
  osmmfree((char *) theta);




  /*close all and finish it*/

  SCFCLO(aid);
  SCFCLO(bid);

  PCCLOS ();

  SCSEPI ();
  return(0);
}










#ifdef __STDC__ 
int ipolover(float degree[], float angle[], int npix[], 
	     double start[], double step[], float wcframx[], float wcframy[], 
	     float scale, int istart)
#else
int ipolover(degree, angle, npix, start, step, wcframx, wcframy,scale,istart)
     float degree[], angle[], wcframx [], wcframy[], scale;
     int npix[],istart;
     double start[], step[];
#endif
{
  int i, j, ii;
  int ifirst, ilast, jfirst, jlast;
  float xdata, ydata, xoff, yoff, factor, fac, xx[2], yy[2];
  float pol, ang;


  fac = (float) (180.0/PI);

  factor = scale;  

  istart = npix[0] * f2i( (float) ((wcframy[0]-start[1])/step[1]) );

  ifirst = (int) ((wcframx[0] - start[0])/step[0]);
  ilast  = (int) ((wcframx[1] - start[0])/step[0]);
  jfirst = (int) ((wcframy[0] - start[1])/step[1]);
  jlast  = (int) ((wcframy[1] - start[1])/step[1]);

  ifirst = IMAX(ifirst, 0);
  ilast = IMIN(ilast, npix[0]);
  jfirst = IMAX(jfirst, 0);
  jlast = IMIN(jlast, npix[1]);
  
  if (debug >=2)
    printf("   ...ifirst = %6d  ilast = %6d jfirst = %6d jlast = %6d \n", 
	  ifirst, ilast,jfirst, jlast);

  for (j = jfirst ; j < jlast; j++) 
    {
      for (i = ifirst ; i < ilast; i++) 
	{ 	  
	  xdata = (float) ((((float) i) * step[0]) + start[0]);
	  ydata = (float) ((((float) j) * step[1]) + start[1]);

	  ii = j * npix[0] + i - istart;


	  /* 
	     the polarisation angle counts from the north: therefore ++90
	     0 degree means that the E vectors vibration is to the north/south
	     a la Serkovski
	  */
	  pol = degree[ii];
	  ang = (angle[ii] + 90.0) / fac;
      	  

	  xoff =  factor * pol * (float) cos(ang);
	  yoff =  factor * pol * (float) sin(ang);


	  xx[0] = (float) (xdata - xoff);
	  xx[1] = (float) (xdata + xoff);
	  yy[0] = (float) (ydata - yoff);
	  yy[1] = (float) (ydata + yoff);


	  if (debug > 2)
	    printf("   ...j = %6d  i = %6d istart = %6d ii = %6d \n xx[0] = %5.1f, xx[1] = %5.1f, yy[0] = %5.1f, yy[1] = %5.1f \n", j,i,istart,ii, xx[0], xx[1], yy[0], yy[1]);

	  /*	  AG_GPLL (xx, yy, 2);*/

	  PCDATA ( 0, 1, 0,xx,yy, 0.0, 2);

	}
    }   
  

  return 0;  
}



#ifdef __STDC__
void mapslit(int fid, float yst, float yen, int *istart, int *ipxl, 
               float image[],double dstart[],double dstep[],int inpix[]) 
#else
void mapslit(fid,yst,yen,istart,ipxl,image,dstart,dstep,inpix)
    int fid, *istart, *ipxl,inpix[];
    float image[], yst, yen;
    double dstart[],dstep[];
#endif
{
  int actsize;


  
  *istart = inpix[0] * f2i((yst - (float) dstart[1])/((float) dstep[1])) +1;
  SCFGET(fid, *istart, *ipxl, &actsize, (char *) image); 
  if (debug >= 2)
    printf("   ...1st pxl = %6d  Npxl = %6d  1st pxl value =%9.5f\n", *istart,
	   *ipxl,image[0]);
}



#ifdef __STDC__
  int f2i(float x)
#else
  int f2i(x)
    float x;
#endif
{
  int ix;
  ix = (int) (x +0.5);

  return ix;
}


















