/*                         */
/* ipolstok2d.c            */
/* Otmar Stahl, 14.6.1995  */
/*                         */

#include <math.h>
#include <stdio.h>
#include <string.h>

#include <midas_def.h>

#include <proto_fors.h>
#include <proto_polarutils.h>


#define LINEAR 0
#define CIRCULAR 1
#define MAXIMS 8
#define myabs(A) ((A)  >  0 ? (A) : (-A))
#define RAD1 (180.0/ acos(-1.0))

#ifndef PI
#define PI (acos(-1.0))
#endif

int main ()
     
{
  double start[2*MAXIMS], step[2*MAXIMS], ostep, ostart, eps;

  int iseq, frmcnt, stat, iav, uni, nulo, onpix, n, i,j,imno[3], imno_err[3];
  int polmode, null, tid, frame_col, ang_col,ipos[2],itmp;
  int ncol, nrow, nsort, acol, arow, iselect, offset;
  int imnol[MAXIMS], imno_errl[MAXIMS], npix[2*MAXIMS], naxis[MAXIMS];
  int phi_col;

  char catfil[64], output[64], frame[64], frame_err[64], cunit[64];
  char outima[64], outerr[64], append[1], mode[3],text[80];
  char conima[64];

  float *ao, *ae, *bo, *be, *ao_err, *ae_err, *bo_err, *be_err, *ima, *err; 
  float zeroang, *Q, *Qerr,*U, *Uerr,phi,tmp,*phic;

  static char  error1[] = "operands do not match in step size... ";
  static char  error2[] = "operands do not match in start value... ";
  static char  error3[] = "operands do not match in pixel number... ";

  size_t  istr;

/*         get into Midas and get parameters             */

  stat = SCSPRO ("stokes");

  stat = SCKGETC("IN_A",1,60,&iav,catfil);  
  stat = SCKGETC("INPUTC",1,3,&iav,mode);
  stat = SCKRDI("INPUTI",1,2,&iav,ipos,&uni,&null);
  stat = SCKRDR("INPUTR",1,1,&iav,&zeroang,&uni,&null);
  stat = SCKGETC("CONIMA",1,64,&iav,conima);

  polmode = LINEAR;
  if (!strncmp (mode, "CIR", 3) || !strncmp (mode, "cir", 3))
    polmode = CIRCULAR;

  /* loop through catalogue */

  stat = TCTOPN(catfil, F_I_MODE, &tid);
  stat = TCIGET (tid, &ncol, &nrow, &nsort, &acol, &arow);


  stat = TCCSER (tid, conima, &frame_col);
  stat = TCCSER (tid, ":RET_ANG", &phi_col);

  phic = fvector(0, 2*ipos[0] - 1);

  printf("...input frames - flux and standard deviation \n");
  iseq = -1;
  for (frmcnt=0; frmcnt<nrow; frmcnt++)           /* loop through table */
    {
      stat = TCSGET (tid, frmcnt+1, &iselect);            
      if (iselect) {                               /* selected entries only */
	iseq++;
	
	stat = TCERDC(tid, frmcnt+1, frame_col, frame, &null);
	stat = SCFOPN(frame,D_R4_FORMAT,0,F_IMA_TYPE,&imnol[iseq]); 
	
        stat = TCERDR(tid, frmcnt+1, phi_col, &phic[iseq], &null);
        phic[iseq] =  phic[iseq] * (float) ipos[1] + zeroang / 2.;
	
	
	
	
	istr = strlen(frame);
	/*	
	  strncpy(output,frame,istr);
	  sprintf(frame_err,"%sERR",output);
	  */
	frame_err[0] = '\0';
	strncat(frame_err,frame,istr);
	strcat(frame_err,"ERR\0");
	printf("      %s  %s \n",frame, frame_err);
       	stat = SCFOPN(frame_err,D_R4_FORMAT,0,F_IMA_TYPE,&imno_errl[iseq]); 
      }
    }

  /* read descriptors of first frame */

  stat = SCDRDI(imnol[0],"NAXIS",1,1,&iav,&naxis[0],&uni,&nulo);
  stat = SCDRDI(imnol[0],"NPIX",1,naxis[0],&iav,&npix[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"START",1,naxis[0],&iav,&start[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"STEP",1,naxis[0],&iav,&step[0],&uni,&nulo);
  stat = SCDGETC(imnol[0],"CUNIT",1,64,&iav,cunit);

  ostart = start[0];
  ostep = step[0];
  eps = 0.0001*step[0];
  onpix = npix[0];

  /* compare catalogue frames with first frame in catalogue */ 

  for (n=1; n < iseq; n++)
    {
      start[2*n] = 0.0;
      step[2*n] = 1.0;
      npix[2*n] = 1;
      
      stat = SCDRDI(imnol[n],"NAXIS",1,1,&iav,&naxis[n],&uni,&nulo);
      stat = SCDRDI(imnol[n],"NPIX",1,1,&iav,&npix[2*n],&uni,&nulo);
      stat = SCDRDD(imnol[n],"START",1,1,&iav,&start[2*n],&uni,&nulo);
      stat = SCDRDD(imnol[n],"STEP",1,1,&iav,&step[2*n],&uni,&nulo);
      
      /*  check step size, start values and pixel numbers ...  */

      if (myabs(step[2*n]-ostep) > eps) SCETER(5,error1);
      if (myabs(start[2*n]-ostart) > eps) SCETER(5,error2);
      if (npix[2*n] != onpix) SCETER(5,error3);
    }
  
  /* all o.k. now allocate space for arrays */

  ao = fvector(0, npix[0] - 1);
  ae = fvector(0, npix[0] - 1);
  ao_err = fvector(0, npix[0] - 1);
  ae_err = fvector(0, npix[0] - 1);

  if (npix[1] <= 0 )  npix[1] = 1;


  ima = fvector (0, npix[0] -1);
  err = fvector (0, npix[0] -1);








  if (polmode == CIRCULAR) {

  /* construct output file names by appending to catalogue name */

    append[0] = 'V';
    stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
		      outima,  outerr, &imno[0], &imno_err[0]);

    /* read four data frames and corresponding error frames */
    /* loop over rows, ipol only */
    bo = fvector(0, npix[0] - 1);
    be = fvector(0, npix[0] - 1);
    bo_err = fvector(0, npix[0] - 1);
    be_err = fvector(0, npix[0] - 1);

    for (nrow = 0; nrow < npix[1]; nrow++)
      {
	
	offset = nrow*npix[0]+1;
	stat = SCFGET(imnol[0],offset,npix[0],&nulo, (char *)ao);
	stat = SCFGET(imnol[1],offset,npix[0],&nulo, (char *)ae);
	stat = SCFGET(imnol[2],offset,npix[0],&nulo, (char *)bo);
	stat = SCFGET(imnol[3],offset,npix[0],&nulo, (char *)be);
      
	stat = SCFGET(imno_errl[0],offset,npix[0],&nulo, (char *)ao_err);
	stat = SCFGET(imno_errl[1],offset,npix[0],&nulo, (char *)ae_err);
	stat = SCFGET(imno_errl[2],offset,npix[0],&nulo, (char *)bo_err);
	stat = SCFGET(imno_errl[3],offset,npix[0],&nulo, (char *)be_err);

	stat = stokesv(ao,ae,bo,be,ao_err,ae_err,bo_err,be_err,ima,err,
		       npix[0]);
      
	stat = SCFPUT(imno[0], offset, npix[0], (char*)ima);
	stat = SCFPUT(imno_err[0], offset, npix[0], (char*)err);  
      }

    stat = SCFCLO(imno[0]);
    stat = SCFCLO(imno_err[0]);

    free_fvector(bo, 0, npix[0] - 1);  
    free_fvector(be, 0, npix[0] - 1);  
    free_fvector(bo_err, 0, npix[0] - 1);  
    free_fvector(be_err, 0, npix[0] - 1);  

  }











/* once again for linear polarisation */

  if (polmode == LINEAR) 
    {

      Q = fvector (0, npix[0] -1);
      U = fvector (0, npix[0] -1);
      Qerr = fvector (0, npix[0] -1);
      Uerr = fvector (0, npix[0] -1);

      /* for Stokes Q and U */

      append[0] = 'Q';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[0], &imno_err[0]);
      append[0] = 'U';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[1], &imno_err[1]);

      /* for Stokes I */

      append[0] = 'I';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
		    outima,  outerr, &imno[2], &imno_err[2]);

      tmp = 0.;
      itmp = 0;

      for (nrow = 0; nrow < npix[1]; nrow++)
	{

	  offset = nrow*npix[0]+1;
	  
	  for (i = 0; i < npix[0]; i++)
	    {
	      Q[i] = Qerr[i] = U[i] = Uerr[i] = ima[i] = err[i] = 0;
	    }
	  for (i = 0; i < ipos[0]; i++)
	    {


	      stat = SCFGET(imnol[i*2],offset,npix[0],&nulo, (char *)ao);
	      stat = SCFGET(imnol[i*2+1],offset,npix[0],&nulo, (char *)ae);
	      stat = SCFGET(imno_errl[i*2],offset,npix[0],&nulo, 
			    (char *)ao_err);
	      stat = SCFGET(imno_errl[i*2+1],offset,npix[0],&nulo, 
			    (char *)ae_err);



	      /*add the flux at this retarder position to the fourier sum*/ 
	      sfourier(ao,ao_err,ae,ae_err,phic[i*2],Q,Qerr,U,Uerr,npix[0],
		       ipos[0]);

	      /*add (ord+ext)/2*ipos[0] to Stokes I*/
	      stat = stokesi(ima,err,ao,ao_err,npix[0],iseq,0);
	      stat = stokesi(ima,err,ae,ae_err,npix[0],iseq,0);


	    }
	  tmp = tmp + 100.0 / ((float) npix[1]);
	  if ( tmp > 1.0){
	    itmp+=1;
	    sprintf (text,"   ...progress %3d percent",itmp);
	    SCTDIS (text,26);
	    tmp -= 1.0;
	  }

	  for (j = 0; j < npix[0]; j++)
	    {
	      Qerr[j] = (float) sqrt(Qerr[j]);
	      Uerr[j] = (float) sqrt(Uerr[j]);
	    }


	  /*Here it would be possible to call the linpol function if you like*/


	  stat = SCFPUT(imno[0], offset, npix[0], (char*)Q);
	  stat = SCFPUT(imno_err[0], offset, npix[0], (char*)Qerr);  
	  stat = SCFPUT(imno[1], offset, npix[0], (char*)U);
	  stat = SCFPUT(imno_err[1], offset, npix[0], (char*)Uerr);  

	  stat = stokesi(ima,err,ae,ae_err,npix[0],iseq,1);
	  
	  stat = SCFPUT(imno[2], offset, npix[0], (char*)ima);
	  stat = SCFPUT(imno_err[2], offset, npix[0], (char*)err);  

	}
      
      stat = SCFCLO(imno[0]);
      stat = SCFCLO(imno_err[0]);
      stat = SCFCLO(imno[1]);
      stat = SCFCLO(imno_err[1]);
      stat = SCFCLO(imno[2]);
      stat = SCFCLO(imno_err[2]);

      free_fvector(Q, 0, npix[0] - 1);    
      free_fvector(U, 0, npix[0] - 1);    
      free_fvector(Qerr, 0, npix[0] - 1);    
      free_fvector(Uerr, 0, npix[0] - 1);    

    }
  printf("\n");

  /*          epilogue ...                 */

  free_fvector(ao, 0, npix[0] - 1);  
  free_fvector(ae, 0, npix[0] - 1);  
  free_fvector(ao_err, 0, npix[0] - 1);  
  free_fvector(ae_err, 0, npix[0] - 1);  
  free_fvector(ima, 0, npix[0] - 1);    
  free_fvector(err, 0, npix[0] - 1);  
  
  SCSEPI();
  exit(0);
}






