/*                                                                 */
/* ipolstoktb.c                                                    */
/*                                                                 */
/* compute Stokes parameters for extracted objects                 */
/*                                                                 */
/* mode can be CIRCULAR or LINEAR                                  */
/* computed are Stokes I, Q and U (LINEAR) or V (CIRCULAR)         */
/* degree and angle of polarization for mode "LINEAR"              */
/*                                                                 */
/* Otmar Stahl, 14.6.1995                                          */
/* Thomas Szeifert                                                 */

#include <math.h>
#include <stdio.h>
#include <string.h>

#include <midas_def.h>
#include <proto_fors.h>
#include <proto_polarutils.h>

#define myabs(A) ((A)  >  0.0 ? (A) : (-A))
#define RAD1     (float) (180.0/ acos(-1.0))

#ifndef PI
#define PI       (float) acos(-1.0)
#endif

#define LINEAR 0
#define CIRCULAR 1
#define MAXIMS 64


/* tfourier: Stokes Q & U for a single table entry*/
#ifdef __STDC__
void tfourier( float , float , float , float , float ,
           float * , float *, float *, float *, int, int);
#else
void tfourier();
#endif

int main ()
{
  int stat, iav,  n, i, j, istr,nstep;
  int polmode, tid, col_flux[MAXIMS], col_err[MAXIMS], unit;
  int col_out[10], ncols, ncol, nrow, nsort, acol, arow, iselect, inull,ipos[2];
  char table[61], mode[4], colname[21];

  float *ao, *ae, *bo, *be, *ao_err, *ae_err, *bo_err, *be_err, *Q, *Qerr;
  float *U, *Uerr, *I, *Ierr, *V, *Verr, *pol, *polerr, *theta, *thetaerr; 
  float *buffima, *buffsig;
  float ext,ext_err,ord,ord_err, phi, zeroang;
/*         get into Midas and get parameters             */

  stat = SCSPRO ("stokes");

  stat = SCKGETC("IN_A",1,60,&iav,table);  
  stat = SCKGETC("INPUTC",1,3,&iav,mode);
  stat = SCKRDI("INPUTI",1,2,&iav,ipos,&unit,&inull);
  stat = SCKRDR("INPUTR",1,1,&iav,&zeroang,&unit,&inull);
  /* open table */
     
  stat = TCTOPN(table, F_IO_MODE, &tid);
  stat = TCIGET (tid, &ncol, &nrow, &nsort, &acol, &arow);






/* check mode (CIRCULAR or LINEAR) */

  polmode = LINEAR;
  if (!strncmp (mode, "CIR", 3) || !strncmp (mode, "cir", 3))
    polmode = CIRCULAR;


/* mode CIRCULAR, Stokes V */

  if (polmode == CIRCULAR) 
    {
      printf ("mode: %s ",mode);

      V = fvector (0, nrow - 1);
      Verr = fvector (0, nrow - 1);

/* read input columns, create output columns and map all columns */

      for (i = -1 ; i <= 1; i=i+2)
	{
	  istr = 4500*i;
	  if (istr < 0) istr = istr + 36000;

      	  sprintf(colname,":O%04iF",istr); /* NET suffix relaced by F*/
	  stat = TCCSER (tid, colname, &col_flux[i+1]);
	  sprintf(colname,":E%04iF",istr); /* NET suffix relaced by F*/
	  stat = TCCSER (tid, colname, &col_flux[i+2]);
	  sprintf(colname,":O%04iERR",istr);
	  stat = TCCSER (tid, colname, &col_err[i+1]);
	  sprintf(colname,":E%04iERR",istr);
	  stat = TCCSER (tid, colname, &col_err[i+2]);

	  printf(" istr = %04i ",istr);
	}

      printf("\n");


      stat = TCCSER(tid, ":V", &col_out[0]);
      stat = TCCSER(tid, ":Verr", &col_out[1]);

      if (col_out[0] == -1)
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "V", &col_out[0]);
      if (col_out[1] == -1)
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "Verr", &col_out[1]);


      stat = TCCMAP(tid,col_flux[0],(char **)&ao);  
      stat = TCCMAP(tid,col_flux[1],(char **)&ae);  
      stat = TCCMAP(tid,col_flux[2],(char **)&bo);  
      stat = TCCMAP(tid,col_flux[3],(char **)&be);  
      stat = TCCMAP(tid,col_err[0],(char **)&ao_err);  
      stat = TCCMAP(tid,col_err[1],(char **)&ae_err);  
      stat = TCCMAP(tid,col_err[2],(char **)&bo_err);  
      stat = TCCMAP(tid,col_err[3],(char **)&be_err);  


      /* circular polarimetry only! */
      stat = stokesv(ao,ae,bo,be,ao_err,ae_err,bo_err,be_err,
		    V,Verr,nrow); 


      for (i = 0 ; i < nrow ; i++){
	/*
	  printf("i = %2d  V = %8.5f  Verr = %8.5f \n",i,V[i],Verr[i]);
	  */
	TCEWRR(tid,i+1,col_out[0],&V[i]);	
	TCEWRR(tid,i+1,col_out[1],&Verr[i]);	
      }


      free_fvector(V, 0, nrow - 1);    
      free_fvector(Verr, 0, nrow - 1);  

    }    

/* mode LINEAR, Stokes Q and U*/

  else 
    {
      Q = fvector (0, nrow -1);
      U = fvector (0, nrow -1);
      Qerr = fvector (0, nrow -1);
      Uerr = fvector (0, nrow -1);
      pol = fvector (0, nrow -1);
      theta = fvector (0, nrow -1);
      polerr = fvector (0, nrow -1);
      thetaerr = fvector (0, nrow -1);
      
      for (i = 0 ; i < ipos[0] * 2; i=i+2)
	{
	  if (i != 0 && ipos[1] == -1){ /*flip retarder rotation direction*/
	    istr = 1125*( ipos[0] * 2 - i);
	  }
	  else{
	    istr = 1125*i; 
	  }
	  printf(" istr = %04i ",istr);

	  sprintf(colname,":O%04iF",istr);/* NET suffix relaced by F*/
	  stat = TCCSER (tid, colname, &col_flux[i]);
	  sprintf(colname,":O%04iERR",istr);
	  stat = TCCSER (tid, colname, &col_err[i]);
	  sprintf(colname,":E%04iF",istr);/* NET suffix relaced by F*/
	  stat = TCCSER (tid, colname, &col_flux[i+1]);
	  sprintf(colname,":E%04iERR",istr);
	  stat = TCCSER (tid, colname, &col_err[i+1]);
	}
      printf("\n");

      stat = TCCSER(tid, ":Q", &col_out[0]);
      stat = TCCSER(tid, ":Qerr", &col_out[1]);
      stat = TCCSER(tid, ":U", &col_out[2]);
      stat = TCCSER(tid, ":Uerr", &col_out[3]);
      

      if (col_out[0] == -1) 
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "Q", &col_out[0]);
      if (col_out[1] == -1) 
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "Qerr", &col_out[1]);
      if (col_out[2] == -1) 
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "U", &col_out[2]);
      if (col_out[3] == -1) 
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "Uerr", &col_out[3]);

      for (j = 0 ; j < nrow ; j++)
	{
	  TCSGET(tid,j+1,&iselect);
	  if (iselect == 1)
	    {
	      for (i = 0 ; i < ipos[0] ; i++)
		{
		  TCERDR(tid,j+1,col_flux[i*2],&ord,&inull);
		  TCERDR(tid,j+1,col_err[i*2],&ord_err,&inull);
		  TCERDR(tid,j+1,col_flux[i*2+1],&ext,&inull);
		  TCERDR(tid,j+1,col_err[i*2+1],&ext_err,&inull);

		  /*correct for zero angle of retarder*/ 
		  phi = 22.5 * ((float) i) + zeroang / 2.; 
		  if (phi < 0.0)
		    phi = phi + 22.5 * (float) ipos[0];
		  if (phi > 22.5 * (float) ipos[0])
		    phi = phi - 22.5 * (float) ipos[0];

		  /* fourier sum of the flux quotient at any retarder angle*/
		  tfourier(ord,ord_err,ext,ext_err,phi,Q,Qerr,U,Uerr,j,
			   ipos[0]);

		}
	      Qerr[j] = (float) sqrt(Qerr[j]);
	      Uerr[j] = (float) sqrt(Uerr[j]);
	      TCEWRR(tid,j+1,col_out[0],&Q[j]);
	      TCEWRR(tid,j+1,col_out[1],&Qerr[j]);
	      TCEWRR(tid,j+1,col_out[2],&U[j]);
	      TCEWRR(tid,j+1,col_out[3],&Uerr[j]);
	    }
	}
      
      
      /* mode LINEAR, degree and angle */

      stat = TCCSER(tid, ":pol", &col_out[4]);
      stat = TCCSER(tid, ":polerr", &col_out[5]);
      stat = TCCSER(tid, ":theta", &col_out[6]);
      stat = TCCSER(tid, ":thetaerr", &col_out[7]);
      if (col_out[4] == -1)
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "pol", &col_out[4]);
      if (col_out[5] == -1)
	TCCINI(tid, D_R4_FORMAT, 1, "F9.6", " ", "polerr", &col_out[5]);
      if (col_out[6] == -1)
	TCCINI(tid, D_R4_FORMAT, 1, "F9.4", " ", "theta", &col_out[6]);
      if (col_out[7] == -1)
	TCCINI(tid, D_R4_FORMAT, 1, "F9.4", " ", "thetaerr", &col_out[7]);

      stat = TCCMAP(tid,col_out[4],(char **)&pol);  
      stat = TCCMAP(tid,col_out[5],(char **)&polerr);  
      stat = TCCMAP(tid,col_out[6],(char **)&theta);  
      stat = TCCMAP(tid,col_out[7],(char **)&thetaerr);  

      nstep = 20; /*nstep = number of grid points for gaussian integral*/
      
      stat = linpol(Q, Qerr, U, Uerr, pol, polerr, theta, thetaerr,
		    nrow,nstep,1);


      /*
      for (j = 0 ; j < nrow ; j++)
	printf("%7.4f %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f \n",
	       Q[j],U[j],Qerr[j],Uerr[j],pol[j],theta[j],polerr[j],
	       thetaerr[j]);
	       */

      free_fvector(Q, 0, nrow - 1);
      free_fvector(U, 0, nrow - 1);
      free_fvector(Qerr, 0, nrow - 1);  
      free_fvector(Uerr, 0, nrow - 1);  

      /*
      free_fvector(pol, 0, nrow - 1);    
      free_fvector(theta, 0, nrow - 1);   
      free_fvector(polerr, 0, nrow - 1);
      free_fvector(thetaerr, 0, nrow - 1);
	*/

    }


  /* now Stokes I */


  stat = TCCSER(tid, ":I", &col_out[8]);
  stat = TCCSER(tid, ":Ierr", &col_out[9]);
  if (col_out[8] == -1)
    TCCINI(tid, D_R4_FORMAT, 1, "F9.0", " ", "I", &col_out[8]);
  if (col_out[9] == -1)
    TCCINI(tid, D_R4_FORMAT, 1, "F9.0", " ", "Ierr", &col_out[9]);


  I = fvector (0, nrow -1 );
  Ierr = fvector (0, nrow - 1);

  ncols = ipos[0]*2;
  if (polmode == CIRCULAR)
    ncols = 4;
  stat = stokesi(I,Ierr,buffima,buffsig,nrow,ncols,-1);         /* -1 = init */

  for (n=0; n < ncols; n++)                       /* loop all retarder angles*/
    {
      stat = TCCMAP(tid,col_flux[n],(char **)&buffima);  
      stat = TCCMAP(tid,col_err[n],(char **)&buffsig);  
      stat = stokesi(I,Ierr,buffima,buffsig,nrow,ncols,0);       /*  0 = add */
    }
  
  stat = stokesi(I,Ierr,buffima,buffsig,nrow,ncols,1);    /*  1 = sqrt(sigma)*/



  for (i = 0 ; i < nrow ; i++){
    TCEWRR(tid,i+1,col_out[8],&I[i]);	
    TCEWRR(tid,i+1,col_out[9],&Ierr[i]);	
  }


  /*  free space and epilogue ...   */

  free_fvector(I, 0, nrow - 1);
  free_fvector(Ierr, 0, nrow - 1);  


  TCTCLO(tid);

  SCSEPI();
  exit(0);
}

/*---------------------------------------------------------------------------*/

#ifdef __STDC__
void tfourier(float ao, float ao_err, float ae, float ae_err, 
	     float phi,float *Q ,float *Qerr, float *U, float *Uerr, int row, 
	     int ipos)
#else
void tfourier(ao,ao_err,ae,ae_err,phi,Q,Qerr,U,Uerr,row,ipos)
     float ao,ae,ao_err,ae_err, phi;
     float *Q,*Qerr,*U,*Uerr;
     int row,ipos;
#endif
{
  
  float buffer;
  /*  Compute Stokes parameter -- add the U and Q components of the 
      observations at discrete retarder angles. It's a Fouriertransformation
      in principle (see FORS-usermanual)
   */
 
  if   (ao - ae != 0)
    Q[row] = Q[row] + (2.0/ipos)*((ao-ae)/(ao+ae)) * 
      (float) cos((double) (4.0 * phi/RAD1));

  if   (ao - ae != 0)
    U[row] = U[row] + (2.0/ipos)*((ao-ae)/(ao+ae)) * 
      (float) sin((double) (4.0 * phi/RAD1));



  /* error calculation:*/
  /* valid for small ao_err/ao and ae_err/ae */

  /* relative error of the quotient (ao-ae)/(ao+ae): */
  if   (ao-ae != 0.0 && ao+ae != 0.0 )
    buffer = (float) sqrt((SQR(ao_err)+SQR(ae_err))/SQR(ao-ae)+(SQR(ao_err)+SQR(ae_err))/SQR(ao+ae));  
  else
    buffer = 1.0e+31; 
  
  /* sum square of absolute error: */ 
  Qerr[row] = Qerr[row] + SQR((2.0/ipos)*((ao-ae)/(ao+ae))*(float) sin(4.0*phi/RAD1)*buffer);
  /*cos not sin TS 1999-03-14 */
  Uerr[row] = Uerr[row] + SQR((2.0/ipos)*((ao-ae)/(ao+ae))*(float) cos(4.0*phi/RAD1)*buffer);

}






