/*                                                                         */
/*           Fit Gauss-Hermite function to line                            */
/*                                                                         */
/*           Otmar Stahl, Jan. 28, 1998                                    */

#include <math.h>
#include <midas_def.h>
#include <proto_fors.h>

void center_line (
#ifdef __STDC__
		  double *, double *, float *, int, int
#endif
		  );

void dhermite (
#ifdef __STDC__
	      double x, double p[], int np
#endif
	      );
double eval_dhermite (
#ifdef __STDC__
		      double x, double p[], int np
#endif
		      );

main()
{
  int i, j, iav, kun, knul;
  int nax, imno, npix, Npix[2];

  int pix1, pix2, ncoef, param[4];

  char image_i[60], image_o[60];
  char unit[64], inst[72], line[80];

  float *buff1, *buff2, *in_image, *out_image;

  double *Xgaus, *Ygaus;
  double Start[2], Step[2];

/*         get into Midas and get parameters             */

  SCSPRO ("fit_gh");

  SCKGETC ("IN_A", 1, 60, &iav, image_i);
  SCKGETC ("OUT_A", 1, 60, &iav, image_o);
  SCKRDI ("INPUTI", 1, 4, &iav, param, &kun, &knul);

/*          map input file                              */

  SCIGET (image_i, D_R4_FORMAT, F_I_MODE, F_IMA_TYPE,
	  1, &nax, Npix, Start, Step, inst, unit, (char **)&in_image, &imno);

  pix1 = param[0];
  pix2 = param[1];
  ncoef = param[2];
  
  npix = pix2 - pix1 + 1;

  Xgaus = dvector (1, npix);
  Ygaus = dvector (1, npix);

  for (i=1; i<=npix; i++)
    {
      j = pix1-2 + i;
      Xgaus[i] = j*Step[0] + Start[0];
      Ygaus[i] = in_image[j];
    }

  Npix[0] = npix;
  Start[0] = Xgaus[1];

  SCIPUT (image_o, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE,
	  1, Npix, Start, Step, inst, unit, (char **)&out_image, &imno);

  center_line (Xgaus, Ygaus, out_image, npix, ncoef);

/*          epilogue ...                 */

  free_dvector(Xgaus, 1, npix);
  free_dvector(Ygaus, 1, npix); 

  SCSEPI ();
  return (0);

}

/*---------------------------------------------------------------------------*/
void 
#ifdef __STDC__
center_line (double *Xgaus, double *Ygaus, float *image, int npix, int ncoef)
#else
     center_line (Xgaus, Ygaus, image, npix, ncoef)
     double *Xgaus, *Ygaus; 
     float *image;
     int   npix, ncoef;
#endif
{
  
  double **covar,  *A, *sig, *coef, chisq;
  
  int i, *ia;
  char line[80];

  A = dvector (1, 3);
  i = npix / 2;
  
  A[1] = Ygaus[i];
  A[2] = Xgaus[i];
  A[3] = 2.0;

  fit_gauss (Xgaus, Ygaus, npix, A);
  
  for (i=1; i<=npix; i++)
    {
      Xgaus[i] = (Xgaus[i] - A[2]) / A[3] ;
      Ygaus[i] = Ygaus[i] / A[1];
    }

  covar = (double **) dmatrix (1,npix, 1, npix); 
  sig = (double *) dvector (1, npix); 
  ia = (int *) ivector (1, ncoef); 
  coef = (double *) dvector (1, ncoef); 

  for (i = 1; i <= ncoef; i++) 
    ia[i] = 1; 
  for (i= 1; i <= npix; i++) 
    sig[i] = 1.0; 
  
  dlfit(Xgaus, Ygaus, sig, npix, coef, ia, ncoef, covar, 
	&chisq, dhermite); 

  for (i=1; i<=3; i++)
    {
      printf("%d %f\n",i,A[i]);
    }

  for (i=2; i<=ncoef; i++)
    {
      printf("%d %f\n",i,coef[i]);
    }
  
  for (i=1; i<=npix; i++)
    {
      image[i-1] = (float) eval_dhermite(Xgaus[i], coef, ncoef) * A[1];
    }
  
  free_dvector(A, 1, 3);
}

void 
#ifdef __STDC__
dhermite (double x, double p[], int np)
#else
     dhermite (x, p, np)
     double x, p[];
     int np;
#endif
{

  /* Gauss-Hermite fitting routine */

  int facul[10] = {1,1,2,6,24,120,720,5040,40320,362880};
  int powerof2[10] = {1,2,4,8,16,32,64,128,256,512};
  int i;

  double exp1, h[10];

  /* compute Gauss-Hermite functions from recursion relation 
     Up to 8 fit parameters are allowed */

  h[0] = 1.0;
  h[1] = 2 * x;
  
  for (i=2; i<=np+1; i++)
    h[i] = 2 * x  * h[i-1] - 2 * (i-1) * h[i-2];

  /* h1 and h2 are zero for best fitting gaussian */

  exp1 = exp( -(x*x) / 2.0 );
  
  p[1] = h[0] * exp1;
  
  for (i=2; i<=np; i++)
    {
      p[i] = exp1 * h[i+1] /  sqrt( (double) ( facul[i+1] * powerof2[i+1]));
    }
}

double 
#ifdef __STDC__
eval_dhermite (double x, double p[], int np)
#else
     eval_dhermite (x, p, np)
     double x, p[];
     int np;
#endif
{

  /* Gauss-Hermite evaluation routine */

  int facul[10] = {1,1,2,6,24,120,720,5040,40320,362880};
  int powerof2[10] = {1,2,4,8,16,32,64,128,256,512};

  double exp1, result, h[10];
  int i;

  h[0] = 1.0;
  h[1] = 2 * x;
  
  for (i=2; i<=np+1; i++)
    h[i] = 2 * x  * h[i-1] - 2 * (i-1) * h[i-2];

  exp1 = exp( -(x*x) / 2.0 );

  result = p[1] * h[0];

  for (i=2; i<=np; i++)
    {
      result += p[i] * h[i+1] /  sqrt( (double) ( facul[i+1] * powerof2[i+1]));
    }

  result *= exp1;
  return (float) result;
}
