/* 
   moscalifors.c

   Thomas Szeifert            LSW-Heidelberg           6. June 1998
   together with Walter Fuertig, Walter Seifert, Otmar Stahl 


   automatic wave length calibration for focal reducers with grisms
*/
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>

#include <tbldef.h>
#include <midas_def.h>
#include <proto_fors.h>





#ifndef PI
#define PI       acos(-1.0)
#endif
#define RAD1     (180.0 / PI)

#ifdef __STDC__
int interpol(float x1[], float y1[], int n1, float x2[], float y2[], int n2);
#else
int interpol();
#endif

void fitsel_poly (
#ifdef __STDC__
	double [], double [], int *, double *, double *, 
	double *, int , int , double *, int ,  double, int, double *
#endif
);

void
fitsel_poly_2D (
#ifdef __STDC__
        double [], double [], double [], int *, 
	double [], double [], double [], double [], 
	int, int, double *, int, double, int, int, int, double *
#endif
);

void 
mos_savedisp (
#ifdef __STDC__
	      char name[], double y[], float s[], int nrow, double coef[],
	      double start[],double step[], int fit_deg, int degx, int degxy, 
	      double dnull
#endif
);

double 
closest_to(
#ifdef __STDC__
	   double reference[], double what, double howclose, double dnull, 
	   int *jjj, int kkk
#endif
);

#ifdef __STDC__
int     
interpol(float x1[], float y1[], int n1, float x2[], float y2[], int n2)
#else
int     
interpol(x1, y1, n1, x2, y2, n2)
int n1,n2;
float x1[],y1[],x2[],y2[];
#endif
{
  int i,istart=0;
  float *buff;

  buff = fvector(1,n1);

  spline(x1, y1, n1, 1.0e30, 1.0e30, buff);

  for (i = 0 ; i < n2 ; i++)
    splint(x1, y1, buff, n1, x2[i], &y2[i]);

  free_fvector(buff,1,n1);
  return 0;
}


/*
  calculate the dispersion of the glass like in the Schott catalog
  (W.Fuertig, W.Seifert)
 */
int bindex
#ifdef __STDC__
(double lam, double coef[], double *gind)
#else
(lam, coef, gind)
double lam, coef[], *gind;
#endif
{
  double lam2;
  int i;

  lam2 = SQR(lam/10000.0);

  *gind = lam2 * coef[0] / ( lam2 - coef[1]);
  *gind += lam2 * coef[2] / ( lam2 - coef[3]);
  *gind += lam2 * coef[4] / ( lam2 - coef[5]);
  *gind = sqrt(*gind + 1.0);

  return 0;
}


/*
  calculate the position of a given wavelength on the chip meassured from
  the center (W.Fuertig, W.Seifert)
 */
/*int beam(double tilt, double phi, double *alf1, double gind,
	 double lambda,  double gr,  double ord, double fkam, 
	 double *yccd)*/
int beam
#ifdef __STDC__
(double tilt, double phi, double *alf1, double gind,
         double lambda,  double gr,  double ord, double fkam, 
         double *yccd)
#else
(tilt, phi, alf1, gind, lambda, gr, ord, fkam,  yccd)
double tilt, phi, *alf1, gind, lambda, gr, ord, fkam, *yccd;
#endif
{

  double alf2,bet2,alf3,bet3,betg,alfg, dnull = 0.0;

  lambda /= 1.0e+7;

  alf2 = *alf1 + tilt;
  bet2 = asin( sin(alf2)/gind );
  alf3 = phi - bet2;
  betg = ord * gr * lambda  - sin(alf3) * gind;
  if (fabs(betg) > 1.0)
    *yccd = -9999.0;
  else{
    betg = asin(betg);
    alfg = betg + phi - tilt;
    *yccd = tan(alfg) * fkam;
  }


  return 0;
}



/*
  find the clesest line in the WLC catalog (O.Stahl)
 */
double closest_to
#ifdef __STDC__
(double reference[], double what, double howclose, double dnull, int *jjj, 
 int kkk)
#else
     (reference, what, howclose, dnull,jjj,kkk)
     double reference[], what, howclose, dnull;
     int *jjj,kkk;
#endif
{
  int i=0, j=0 , k , l;
  double  x1, x2, max;
  
  x1 = what - howclose;
  x2 = what + howclose;
  
  i = 0;
  
  while (reference[i] < x1 || i == kkk)
    i++;

  if (i > kkk) return dnull;


  if (i != 0)
    j = i-1;
  else 
    return dnull;

  while ( (reference[j] < x2 || j == kkk) && j <= kkk)
    j++;

  if (i > kkk) return dnull;

  if ((j-i) == 1) /* identifications should be unique */
    {
      *jjj = i;
      return reference[i];
    }
  else 
    {
      return dnull;
    }
}



int main ()

{
  /*general*/
  int i,j,k,l,m=0,inull,actvals,kun,knul,debug,ibuff,select,selnrow;
  float fnull;
  double dnull;
  char line[80];
  int stat;
  /* dummies*/


  /* io-related*/
  int tid1, tid2, tid3, tid4, icol1[9],icol2[3],icol3[9],icol4[2];
  int lcol,lrow,lsort;
  int ncol,nrow,nsort,acol,arow;
  int mcol,mrow,msort,kcol,krow,ksort;
  int iparam[8],*rej2,rej1;
  int idno;
  int nslit=0,nyallrow=0,idcoef;
  int degy,degxy;
  int *sels1,*lower,*upper,*nyrow,*nline;
  int niter;

  char in_table1[61],in_table2[61],in_table3[61],in_table4[61],in_table5[61],
    out_table[61],mode[60],text[80];
   
  double *x1,*y1,*i1,*w1,*wc1,*r1,*w2,*f2,*x2;
  float *s4,*o4;
  double gr,phi,ncoef[6],gind,tilt,alf1=0,fkam,dparam[8],pxlsize;
  double deltax = 2048.0,dbuff;

  float *o1,*s1;
  float fparam[8],xbuff,ybuff;

  int fitdeg,iorder[2],isel,iall,point=0,poltyp;  
  int nmatch, prevnmatch=-1;

  double *xtmp, *ytmp, *itmp;
  double *dcoef;

  double start[2],step[2],center;
  int npix[2];
  
  int verif=-1,initsel=0;
  double alpha,stdres,tol;

  double newoffset;
  int nident;

  SCSPRO ("califors");
  TCMNUL(&inull,&fnull,&dnull);	/*  */
  

  SCKGETC ("IN_A",   1, 60, &actvals, in_table1);
  SCKGETC ("IN_B",   1, 60, &actvals, in_table2);
  SCKGETC ("IN_C",   1, 60, &actvals, in_table3);
  SCKGETC ("IN_D",   1, 60, &actvals, in_table4);
  SCKGETC ("IN_E",   1, 60, &actvals, in_table5);
  SCKGETC ("OUT_A",   1, 60, &actvals, out_table);

  SCKRDI ("INPUTI",   1, 8, &actvals, iparam,&kun,&knul);
  SCKRDR ("INPUTR",   1, 8, &actvals, fparam,&kun,&knul);
  SCKRDD ("INPUTD",   1, 8, &actvals, dparam,&kun,&knul);
  SCKRDI ("DEBUG",   1, 1, &actvals, &debug,&kun,&knul);


  SCKRDD ("START",   1, 2, &actvals, &start[0],&kun,&knul);
  SCKRDD ("STEP",   1, 2, &actvals, &step[0],&kun,&knul);
  SCKRDI ("NPIX",   1, 2, &actvals, &npix[0],&kun,&knul);

  SCKRDD ("CENTER",   1, 1, &actvals, &center,&kun,&knul);
  if (NINT(center) == -9999)
    center = start[0] + step[0] * (((double) npix[0])/2.0 - 1.0);

  tilt = dparam[0] / RAD1, fkam = dparam[1], pxlsize = dparam[2];

  fitdeg = iparam[0];
  degy = iparam[1];
  degxy = iparam[2];
  poltyp = iparam[3];

  alpha = (double) fparam[0];
  tol = (double) fparam[1];

  SCTPUT ("identification of spectral line from table to table ");
  SCTPUT ("------------");
  sprintf (line, "line table:         %s", in_table1);
  SCTPUT (line);
  sprintf (line, "line catalog:       %s", in_table2);
  SCTPUT (line);
  sprintf (line, "alpha = %f, tol = %f, fitdeg = %2d ", 
	   alpha,tol,fitdeg);
  SCTPUT (line);

  if (poltyp == 0)
    sprintf (line, "normal polynom");
  else if  (poltyp == 1)
    sprintf (line, "Chebichev polynom");
  else 
    sprintf (line, "Legendre polynom");
  SCTPUT (line);
  SCTPUT ("------------");




  /*

    READ THE INPUT TABLE with the locations of the mos slitlets

    */



  TCTOPN (in_table4, F_I_MODE, &tid4);

  TCIGET(tid4,&lcol, &lrow, &lsort, &acol, &arow);

  TCCSER (tid4, ":SLIT", &icol4[0]);
  TCCSER (tid4, ":XOFFSET", &icol4[1]);

  s4 = fvector(0,lrow);
  o4 = fvector(0,lrow);

  /* nslit = 0;*/
  for (i = 0; i < lrow; i++){
    /*
    TCSGET(tid3,j+1,&select);  
    if (select){
      TCERDR(tid4,i+1,icol4[0],&s4[nslit++],&knul);
      TCERDR(tid4,i+1,icol4[1],&o4[nslit],&knul);
    }
    */
    TCERDR(tid4,i+1,icol4[0],&s4[i],&knul);
    TCERDR(tid4,i+1,icol4[1],&o4[i],&knul);
  }

  TCTCLO(tid4);


  /*

    READ THE INPUT TABLE of observed line centers and wavelength 
    idenitfications

    */



  TCTOPN (in_table1, F_IO_MODE, &tid1);

  TCIGET(tid1,&mcol, &mrow, &msort, &acol, &arow);

  TCCSER (tid1, ":X", &icol1[0]);
  TCCSER (tid1, ":Y", &icol1[1]);
  TCCSER (tid1, ":IDENT", &icol1[2]);
  TCCSER (tid1, ":WAVE", &icol1[3]);
  TCCSER (tid1, ":WAVEC", &icol1[4]);
  TCCSER (tid1, ":RESIDUAL", &icol1[5]);

  TCCSER (tid1, ":SLIT", &icol1[6]);
  TCCSER (tid1, ":OFFSET", &icol1[7]);
  TCCSER (tid1, ":REJECT", &icol1[8]);


  if (icol1[0] == -1 || icol1[1] == -1 || icol1[6] == -1 ){
    sprintf(line,"Error: missing input column in table %s\n",in_table1);
    SCTPUT(line);
    SCETER(9, "Exiting...");
  }
  if (icol1[2] == -1)
    TCCINI (tid1,D_R8_FORMAT, 1, "F8.2", "Angstroem", "IDENT", &icol1[2]);
  if (icol1[3] == -1)
    TCCINI (tid1,D_R8_FORMAT, 1, "F8.2", "Angstroem", "WAVE", &icol1[3]);
  if (icol1[4] == -1)
    TCCINI (tid1,D_R8_FORMAT, 1, "F8.2", "Angstroem", "WAVEC", &icol1[4]);
  if (icol1[5] == -1)
    TCCINI (tid1,D_R8_FORMAT, 1, "F8.2", "Angstroem", "RESIDUAL", &icol1[5]);
  if (icol1[7] == -1)
    TCCINI (tid1,D_R4_FORMAT, 1, "F6.1", "Pixel", "OFFSET", &icol1[7]);
  if (icol1[8] == -1)
    TCCINI (tid1,D_I4_FORMAT, 1, "I4", "", "REJECT", &icol1[8]);



  
  x1 = dvector(0,mrow);       /*  X-AXES*/
  y1 = dvector(0,mrow);       /*  Y-AXES = ECHELLE-ORDER*/
  i1 = dvector(0,mrow);       /*  WAVE-LENGTH from column IDENT*/
  w1 = dvector(0,mrow);       /*  WAVE-LENGTH from column WAVE  (lab)*/
  wc1 = dvector(0,mrow);      /*  WAVE-LENGTH from column WAVEC (fit)*/
  r1 = dvector(0,mrow);       /*  RESIDUAL*/

  s1 = fvector(0,mrow);       /*  SLIT NUMBER*/
  o1 = fvector(0,mrow);       /*  OFFSET of SLIT*/



  /* changed to start arrays with index 1 for match  TS-180797*/

  for (i = 0; i < mrow; i++)
    { 
      xbuff = fnull;
      stat = TCERDR(tid1,i+1,icol1[0],&xbuff,&knul);
      if (xbuff != fnull && !stat)
	x1[i] = (double) xbuff;
      else
	x1[i] = dnull;
      TCERDR(tid1,i+1,icol1[1],&ybuff,&knul);
      y1[i] = (double) ybuff;
      TCERDR(tid1,i+1,icol1[6],&s1[i],&knul);
      w1[i] = wc1[i] = r1[i] = i1[i] = dnull;
      if (i > 0 && ( fabs(x1[i] - x1[i-1]) > 10.0))
	deltax = fabs(x1[i] - x1[i-1]) < deltax ? fabs(x1[i] - x1[i-1]) 
	  : deltax;
      if (deltax < 1){
	printf ("i = %d  x1[i] = %f x1[i-1] = %f deltax = %f \n",i,x1[i],x1[i-1],deltax);
	return;
      }
      for (k = 0 ; k < lrow ; k++)
	if (NINT(s4[k]-s1[i]) == 0)
	  o1[i] = o4[k];

      rej1 = inull;       

      TCEWRI (tid1,i+1,icol1[8],&rej1);

      TCEWRD (tid1,i+1,icol1[2],&i1[i]);
      TCEWRD (tid1,i+1,icol1[3],&w1[i]);
      TCEWRD (tid1,i+1,icol1[4],&wc1[i]);
      TCEWRD (tid1,i+1,icol1[5],&r1[i]);

      TCEWRR (tid1,i+1,icol1[7],&o1[i]);



    }




  if (debug > 0)
    printf("... minimum delta x in input table = %f \n",deltax);

  /*

    READ TABLE with phycal parameters of the grism

    */


  TCTOPN (in_table3, F_I_MODE, &tid3);

  TCIGET(tid3,&kcol, &krow, &ksort, &acol, &arow);

  TCCSER (tid3, "IDNO", &icol3[0]);
  TCCSER (tid3, "GR", &icol3[1]);
  TCCSER (tid3, "PHI", &icol3[2]);
  TCCSER (tid3, "N0", &icol3[3]);
  TCCSER (tid3, "N1", &icol3[4]);
  TCCSER (tid3, "N2", &icol3[5]);
  TCCSER (tid3, "N3", &icol3[6]);
  TCCSER (tid3, "N4", &icol3[7]);
  TCCSER (tid3, "N5", &icol3[8]);

  if (debug > 3)
    {
      for (j = 0 ; j < 9 ; j++)
	printf("%d ",icol3[j]);
	printf("\n");
    }

  for (j = 0; j < krow; j++)
    {
      TCSGET(tid3,j+1,&select);

      if (select)
	{

	  TCERDI(tid3,j+1,icol3[0],&idno,&knul);
	  TCERDD(tid3,j+1,icol3[1],&gr,&knul);
	  TCERDD(tid3,j+1,icol3[2],&phi,&knul);
	  phi /= RAD1;
	  
	  TCERDD(tid3,j+1,icol3[3],&ncoef[0],&knul);
	  TCERDD(tid3,j+1,icol3[4],&ncoef[1],&knul);
	  TCERDD(tid3,j+1,icol3[5],&ncoef[2],&knul);
	  TCERDD(tid3,j+1,icol3[6],&ncoef[3],&knul);
	  TCERDD(tid3,j+1,icol3[7],&ncoef[4],&knul);
	  TCERDD(tid3,j+1,icol3[8],&ncoef[5],&knul);

	  if (debug > 0)
	    printf("... grism %d  prism angle = %f deg, %f grooves/mm \n",
		   idno,phi*RAD1,gr);

	}
    }

  TCTCLO(tid3);




  /*

    READ THE INPUT CATALOG of laboratory wavelengths and peak intensities

    */



  TCTOPN(in_table2, F_I_MODE, &tid2);
  TCIGET(tid2, &ncol, &nrow, &nsort, &acol, &arow);
  
  TCCSER (tid2, ":WAVE", &icol2[0]);

  if (icol2[0] == -1){
    sprintf(line,"Error: missing input column in table %s\n",in_table2);
    SCTPUT(line);
    SCETER(9, "Exiting...");
  }
  


  w2 = dvector(0,nrow);           /* laboratory wavelength */
  x2 = dvector(0,nrow);           /*  */

  selnrow = 0;
  for (j = 0; j < nrow; j++)
    { 
      TCSGET(tid2,j+1,&select);
      if (select){ 
	TCERDD(tid2,j+1,icol2[0],&w2[selnrow],&knul);

	bindex(w2[selnrow],ncoef,&gind);
	if (debug > 2 )
	  printf(" lamda = %f  n(lambda) = %f \n", w2[selnrow], gind);

	beam(tilt,phi,&alf1,gind,w2[selnrow],gr,1.0,fkam,&x2[selnrow]);
	if (debug > 2)
	  printf("%f %f %f %f %f %f\n",tilt*RAD1,phi*RAD1,gind,w2[selnrow],
		 gr, fkam);

	if ( fabs(x2[selnrow]/pxlsize) < (2.0 * center)){
	  if (debug > 1)
	    printf("... optical guess:  lambda = %7.1f     x2 = %7.3f mm == %7.1f pxl\n",w2[selnrow], x2[selnrow], (x2[selnrow]/pxlsize) + center);
	}

	x2[selnrow] = (x2[selnrow] / pxlsize) + center;
	selnrow++;
      }
    }

  TCTCLO(tid2);





  /*
    find the closest possible identification and check out if it is not double
    */



  newoffset=0.0, nident=0;
  nslit=0;

  i = 0;
  while (i < mrow){

    if (i > 0)
      if (NINT(s1[i]) != NINT(s1[i-1]) )
	{
	  if (debug > 1)
	    printf("slit %2.0f  offset %7.4f -> offset to guess %7.4f pxl \n",s1[i-1],o1[i-1],newoffset);
	  newoffset=0.0,nident=0;
	}

    if (x1[i] != dnull && ((x1[i]-o1[i]) - newoffset) > x2[0] &&
	((x1[i]-o1[i]) - newoffset) < x2[selnrow-1])
      {
	dbuff = closest_to(x2, ((x1[i]-o1[i]) - newoffset) ,(1.5*deltax) , 
			   dnull, &j, selnrow);

      }
    else
      {
	dbuff = dnull;
      }
    /* if we have got an identification and if it's close to the center of
       the frame */
    if (dbuff != dnull && j >= 0 && j < selnrow) 
      {
	i1[i] = w2[j];

	if (debug > 1 && ( fabs(NINT(y1[i]) - 1500) < 3.0 || fabs(NINT(y1[i]) - 700) < 3.0) ) 
	  printf(" %8.2f %8.2f %8.2f %8.2f deltax = %4.1f newoffset = %4.1f\n", x1[i]-o1[i], x2[j], (x1[i]-o1[i])-x2[j], i1[i], deltax, newoffset);
	

	/* here we take the identification to estimate a new offset and
	   later we'll become better and better with our estimate */
	if (nident == 0) 
	  newoffset = ((x1[i]-o1[i])-x2[j]), ++nident;
	else{
	  newoffset += ((x1[i]-o1[i])-x2[j])/((double) (++nident));
	  newoffset /= (double) (1.0 + 1.0 / (double) nident);
	}
      }

    if ( NINT(s1[i]) !=  NINT(s1[i+1]))
      nslit++;
    i++;
  }

  if (nslit == 0) nslit++;



  /*
    find out how many slitlets are in the catalog and how many line have
    to be fitted and more similar things
   */



  sels1 = ivector (0,nslit); /* slitlet id number */
  lower = ivector (0,nslit); /* first line of a slitlet */
  upper = ivector (0,nslit); /* last  line of a slitlet */
  nyrow = ivector (0,nslit); /* number of y-positions in a slitlet */
  nline = ivector (0,nslit); 

  sels1[0] = s1[0];

  nslit = 0;
  lower[nslit] = 0;
  /*  lower[NINT(s1[i])] = i+1;*/
  for (i = 0 ; i <= mrow-1; i++)
    {
      if ( NINT(y1[i]) != NINT(y1[i+1]) && NINT(s1[i]) == NINT(s1[i+1]) )
	nyrow[nslit]++, nyallrow++;
      if ( NINT(s1[i]) !=  NINT(s1[i+1]) )
	{
	  lower[nslit+1] = i+1;
	  upper[nslit] = i;
	  sels1[nslit+1] = s1[i+1];
	  nslit++;
	}
    }


  if (nslit <= 0){
    upper[0] = mrow-1, lower[0] = 0, nslit=1, sels1[0]=1;
    printf ("lower = %d   upper = %d\n",lower[0],upper[0]);
  }



  /* 
     here we start with the line identification
     */

  
  if (debug > 3)
    printf ("%d %d %d %d %d \n",lower[0],upper[0],sels1[0],nyrow[0], nline[0]);

  itmp = dvector(0,mrow);
  xtmp = dvector(0,mrow); 
  ytmp = dvector(0,mrow); 

  rej2 = ivector(0,mrow);

  idcoef = (fitdeg+degy+degxy+1);
  dcoef = dvector (1, nyallrow * idcoef);
  


  for (i = 0 ; i < nslit; i++)
    {
      initsel = 0;

      j = lower[i] + 1;
      while (j <= upper[i])
	{
	  if ( NINT(y1[lower[i]]) !=  NINT(y1[j])) 
	    {
	      nline[i] = j - lower[i]; 
	      j = upper[i]+1;
	    }
	  j++;
	}
    
      if (debug > 1)
	printf ("%d %d %d %d %d\n",lower[i],upper[i],sels1[i],nyrow[i],
		nline[i]);

      /* 
	 do a first fit of data from the optical guess
	 */


      iall = upper[i] - lower[i] + 1;


      if (nyrow[i] > (degy + degxy + 1))
	{
	  fitsel_poly_2D (itmp, xtmp, ytmp, &isel, 
			  &i1[lower[i]],&wc1[lower[i]],
			  &x1[lower[i]],&y1[lower[i]],
			  iall, fitdeg, &dcoef[idcoef*i], debug, dnull, 
			  poltyp,degy,degxy, &stdres);
	}
      else
	{
	  fitsel_poly (itmp, xtmp, &isel, 
		       &i1[lower[i]],&wc1[lower[i]], &x1[lower[i]],
		       iall, fitdeg, &dcoef[idcoef*i], debug, dnull, 
		       poltyp, &stdres);
	}

      if (debug > 0)
	printf("\n... guess:  (slit %d) %d of %d lines identified -> stdev = %5.2f \n",sels1[i],isel,iall,stdres);
      
    
      
      
      /* reset the iteration */      
      for (j = lower[i] ; j < lower[i] + iall ; j++) 
	rej2[j] = NOT_PROCESSED;
      
      /* 
	 take the result of the fit to retry line identification:
	 */

      for (j = lower[i] ; j < lower[i] + iall ; j++) 
	if ((fabs(NINT(y1[i]) - 1500) < 3.0 || fabs(NINT(y1[i]) - 700) < 3.0) && debug > 1)
	  printf ("match v - x = %9.3f  ident = %9.3f  wavec = %9.3f  wave = %9.3f rej = %d\n",	  x1[j],i1[j], wc1[j], w1[j], rej2[j]);

      nmatch = match0 (verif,&w1[lower[i]],&wc1[lower[i]],&y1[lower[i]],
		      &r1[lower[i]], iall,w2,selnrow-1,alpha,&stdres,dnull,
		      &rej2[lower[i]]);

      for (j = lower[i] ; j < lower[i] + iall ; j++) 
	if ((fabs(NINT(y1[i]) - 1500) < 3.0 || fabs(NINT(y1[i]) - 700) < 3.0) && debug > 1)
	  printf ("match n - x = %9.3f  ident = %9.3f  wavec = %9.3f  wave = %9.3f rej = %d\n",	  x1[j],i1[j], wc1[j], w1[j], rej2[j]);




      niter = 0;
      initsel = 0;

      while ( !(prevnmatch == nmatch && initsel > 0 || niter > 20) )
	{
	
	  niter++;
	  
	  /* 
	     reject bad identifications or other bad things 
	     (not during 1st iteration) 
	     */


	  for (j = lower[i] ; j < lower[i] + iall ; j++)
	    {
	      /*
	      if (initsel == 0 && tol > 3 * stdres)
		if (fabs(w1[j]-wc1[j]) > 3 * stdres) 
		  w1[j]=dnull;
		else
		  if (fabs(w1[j]-wc1[j]) > tol) 
		    w1[j]=dnull;
	      */
	      if (initsel == 0)
		{
		  if (fabs(w1[j]-wc1[j]) > 2 * stdres) 
		    w1[j]=dnull;
		}
	      else
		if (fabs(w1[j]-wc1[j]) > tol) 
		  w1[j]=dnull;
	    }


	  initsel++;

	  /* 
	     fit the result again
	     */

	  if (nyrow[i] > (degy + degxy + 1))
	    {
	      fitsel_poly_2D (itmp, xtmp, ytmp, &isel, 
			      &w1[lower[i]],&wc1[lower[i]],
			      &x1[lower[i]],&y1[lower[i]], iall, 
			      fitdeg, &dcoef[idcoef*i], debug, dnull, 
			      poltyp,degy,degxy, &stdres);
	    }
	  else
	    {
	      fitsel_poly (itmp, xtmp, &isel, 
			   &w1[lower[i]],&wc1[lower[i]], &x1[lower[i]],
			   iall, fitdeg, &dcoef[idcoef*i], debug, dnull, 
			   poltyp, &stdres);
	    }
	  
	  if (debug > 0){
	    sprintf(text,"... match:  (slit %d) %d of %d lines identified -> stdev = %5.2f",sels1[i],nmatch,iall,stdres);
	    SCTDIS(text,80); 
	  }
	  /* 
	     retry line identification with Pascal Ballester's match-code
	     */

	  /* reset the iteration */      
	  for (j = lower[i] ; j < lower[i] + iall ; j++) 
	    rej2[j] = NOT_PROCESSED;
	  
	  prevnmatch = nmatch;
	  nmatch = match0 (verif,&w1[lower[i]],&wc1[lower[i]],&y1[lower[i]],
			   &r1[lower[i]], iall ,w2,selnrow-1,alpha,
			   &stdres,dnull, &rej2[lower[i]]);
	
	  for (j = lower[i] ; j < lower[i] + iall ; j++) 
	    if ((fabs(NINT(y1[i]) - 1500) < 3.0 || fabs(NINT(y1[i]) - 700) < 3.0) && debug > 1)
	      printf ("x = %9.3f  ident = %9.3f  wavec = %9.3f  wave = %9.3f rej = %d\n",		    x1[j],i1[j], wc1[j], w1[j], rej2[j]);
	
	
	
	
	} /* end of while loop */
      


      /* fit a last time */
      
      for (j = lower[i] ; j < lower[i] + iall ; j++) 
	{
	  r1[j] = w1[j] - wc1[j];


	  
	  	  
	  if (rej2[j] != SUCCESSFULL_MATCH || fabs(r1[j]) > tol){
	    w1[j] = dnull;
	    if (fabs(r1[j]) > tol) rej2[j] = RESIDUAL_GT_TOL;
	  }
	  
	  TCEWRI (tid1,j+1,icol1[8],&rej2[j]);
	}
      
      /* 
	 fit the results finally
	 */
      
      
      if (nyrow[i] > (degy + degxy + 1))
	{
	  fitsel_poly_2D (itmp, xtmp, ytmp, &isel, 
			  &w1[lower[i]],&wc1[lower[i]],
			  &x1[lower[i]],&y1[lower[i]], iall, 
			  fitdeg, &dcoef[idcoef*i], debug, dnull, 
			  poltyp,degy,degxy,&stdres);
	}
      else
	{
	  fitsel_poly (itmp, xtmp, &isel, 
		       &w1[lower[i]],&wc1[lower[i]], &x1[lower[i]],
		       iall, fitdeg, &dcoef[idcoef*i], debug, dnull, 
		       poltyp, &stdres);
	}
      
      for (j = lower[i] ; j < lower[i] + iall ; j++) 
	{
	  r1[j] = w1[j] - wc1[j];
	  TCEWRI (tid1,j+1,icol1[8],&rej2[j]);
	}

      if (debug > 0)
	printf("... final:  (slit %d) %d of %d lines identified -> stdev = %5.2f (niter=%d)\n",sels1[i],nmatch,iall,stdres,niter);
      
    }
    
  

  /* 
     write the resulting identification into linpos table
     */

  for (i = 0 ; i < mrow; i++)
    {

      if (x1[i] != dnull){
	TCEWRD (tid1,i+1,icol1[2],&i1[i]);
	TCEWRD (tid1,i+1,icol1[3],&w1[i]);
	TCEWRD (tid1,i+1,icol1[4],&wc1[i]);

	TCEWRD (tid1,i+1,icol1[5],&r1[i]);
      }
      else{
	TCEWRD (tid1,i+1,icol1[2],&dnull);
	TCEWRD (tid1,i+1,icol1[3],&dnull);
	TCEWRD (tid1,i+1,icol1[4],&dnull);

	TCEWRD (tid1,i+1,icol1[5],&dnull);
      }
    }


  if (debug > 0){
    printf ("\n... moscalifors; fit parameters: \n");
    
    for (i = 1; i <= (fitdeg+degy+degxy+1);i++)
      printf(" %15.10f",dcoef[i]);
	 
  if (debug > 2)
    printf ("\n... moscalifors... regular end of main\n");
  else if (debug > 0)
    printf ("\n");
  }


  TCTCLO(tid1);


  /* 
     write the resulting coeficients to output table:
   */

  mos_savedisp (out_table, y1, s1, mrow, dcoef, start, step, 
		fitdeg, degy, degxy,dnull);



  /* deallocate the memory */

  free_dvector (dcoef, 1, nyallrow * idcoef);

  free_fvector(s4,0,lrow);
  free_fvector(o4,0,lrow);


  free_dvector (xtmp,0,mrow);  
  free_dvector (ytmp,0,mrow);      
  free_dvector (itmp,0,mrow);


  free_dvector (x1,0,mrow);  
  free_dvector (y1,0,mrow);      

  free_dvector (i1,0,mrow);  
  free_dvector (w1,0,mrow);

  free_dvector (wc1,0,mrow);
  free_dvector (r1,0,mrow);      

  free_ivector (rej2,0,mrow);  

  free_dvector (x2,0,nrow);
  free_dvector (w2,0,nrow); 



  free_ivector (sels1,0,nslit);
  free_ivector (lower,0,nslit);
  free_ivector (upper,0,nslit);
  free_ivector (nyrow,0,nslit);
  free_ivector (nline,0,nslit);



  SCSEPI ();

  exit(0);
  return;
}
/*end of main*/





/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
fitsel_poly (double itmp[], double xtmp[], int *isel, double w1[], 
	     double wc1[], double x1[], int iall, int fitdeg, 
	       double *dcoef, int debug, double dnull,int poltyp, 
	       double *stdres)
#else
void
fitsel_poly (itmp, xtmp, isel, w1, wc1, x1, iall, fitdeg, dcoef, debug,dnull,
	     poltyp, stdres)
     int *isel,iall,fitdeg,debug,poltyp;
     double itmp[],xtmp[],w1[],wc1[],x1[];
     double *dcoef,dnull,*stdres;
#endif

{

  int i,l,point=0;
  int sel = 0;
      

  /* 
     READ_THIS:  

     the arrays of tid1 are given to this subroutine by pointers. Therefore
     point is 0 and we count from   i (= point) = 0   to   i = iall, but not
     from the first table row of the order as in main() 

     */    


  for (i = point ; i < point + iall ; i++) 
    {

      if (w1[i] > 0){
	xtmp[sel] = x1[i];
	itmp[sel] = w1[i];
	sel++;
      }

    }


  /* 
     FIT the dispersion and compute the wavelength for any
     lines the selected and the others of the given order
     */


  if (sel >= fitdeg + 1){
    if (debug > 1)
      printf("   ... fitsel_poly\n");
    dfit_poly (itmp,xtmp,sel,&wc1[point],&x1[point],iall,
	       fitdeg, &dcoef[point],poltyp);
  }
  else{
    dfit_poly (itmp,xtmp,sel,&wc1[point],&x1[point],iall,
	       (sel-1), &dcoef[point],poltyp);
    for (l = sel-2 ; l < fitdeg ; l++)
      dcoef[l] = dnull;
  }



  *isel = sel;
}


/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
fitsel_poly_2D (double itmp[], double xtmp[], double ytmp[], int *isel, 
		double w1[], double wc1[], double x1[], double y1[], 
		int iall, int fitdeg, double *dcoef, int debug, 
		double dnull,int poltyp, int degy, int degxy, double *stdres)
#else
void
fitsel_poly_2D (itmp, xtmp, ytmp, isel, w1, wc1, x1, y1, iall, fitdeg, dcoef, 
		debug, dnull,poltyp, degy, degxy, stdres)
     int *isel,iall,fitdeg,debug,poltyp,degy,degxy;
     double itmp[],xtmp[],ytmp[],w1[],wc1[],x1[],y1[],*stdres;
     double *dcoef,dnull;
#endif

{

  int i,l,point=0;
  int sel = 0, iii;
  double tmp = 0.0;

  /* 
     READ_THIS:  
     
     the arrays of tid1 are given to this subroutine by pointers. Therefore
     point is 0 and we count from   i (= point) = 0   to   i = iall, but not
     from the first table row of the order as in main() 
     
     */    


  for (i = point ; i < point + iall ; i++) 
    {
      
      if (w1[i] != dnull && x1[i] != dnull){
	xtmp[sel] = x1[i];
	ytmp[sel] = y1[i];
	itmp[sel] = w1[i];
	sel++;
      }

  }



  /* 
     FIT the dispersion and compute the wavelength for any
     lines the selected and the others of the given order
     */


  if (debug > 1)
    printf("   ... dfit_cali_2D\n");
  dfit_cali_2D (itmp,xtmp,ytmp,sel,
		&wc1[point],&x1[point],&y1[point],iall,
		fitdeg, &dcoef[point],poltyp,degy,degxy);



  iii = 0;

  for (i = point ; i < point + iall ; i++) 
    {
      if (x1[i] == dnull)
	wc1[i] = dnull;

      if (w1[i] != dnull){ 
	tmp += SQR(w1[i]-wc1[i]);
	iii++;
      }
    }

  tmp = sqrt(tmp/iii);

  *isel = sel;
  *stdres = tmp;

  return;
}






#ifdef __STDC__
void 
mos_savedisp (char name[], double y[], float s[], int nrow, double coef[],
	      double start[],double step[], int fit_deg, int degy, int degxy,
	      double dnull)
#else
void 
mos_savedisp (name, y, s, nrow, coef, start, step, fit_deg, degy, degxy,dnull)
char name[];
double y[], coef[], start[], step[], dnull;
float s[];
int nrow, fit_deg, degxy;
#endif
{
  int i,j,iypxl,istmp;
  int tide,nbline,ncoef,kun;
  int col[3],*ocol,nyrow=0,nslit=0;
  double *ocoef;
  double poly;
  char descname[81],colname[81];


  ocol = ivector(0,fit_deg);
  ocoef = dvector(0,fit_deg);

  if (TCTINI (name, F_TRANS, F_IO_MODE, 5L, 10L, &tide))
    SCTPUT ("**** Error while creating output table");
  nbline = 0;


  TCCSER (tide, ":SLIT", &col[0]);
  if (col[0] == (-1))
    TCCINI (tide, D_I4_FORMAT, 1, "I6", "Slit Number",
            "SLIT", &col[0]);
  TCCSER (tide, ":ROW", &col[1]);
  if (col[1] == (-1))
    TCCINI (tide, D_I4_FORMAT, 1, "I6", "Row Number",
            "ROW", &col[1]);
  TCCSER (tide, ":Y", &col[2]);
  if (col[2] == (-1))
    TCCINI (tide, D_R8_FORMAT, 1, "F8.2", "Y Value",
            "Y", &col[2]);


  for (i = 0; i <= fit_deg; i++)
    {
      sprintf (colname, ":COEF_%d\0", i+1);
      TCCSER (tide, colname, &ocol[i]);
      if (ocol[i] == (-1))
        TCCINI (tide, D_R8_FORMAT, 1, "F16.10", "Coefficients",
                colname, &ocol[i]);
    }

	 

  ncoef = fit_deg + degy + degxy + 1;

  for (j = 0; j < nrow-1; j++)
    if(y[j+1] != y[j])
      {
	nyrow++;

	for (i = 1; i <= (ncoef-(degy+degxy)); i++)
	  ocoef[i-1] = coef[i + nslit*ncoef];
	poly = y[j];
	for (i = 1; i <= degy; i++)
	  {
	    ocoef[0] += coef[(ncoef-degy)-degxy+i+nslit*ncoef]*poly;
	    poly = poly * y[j];
	  }
	poly = y[j];
	for (i = 1; i <= degxy; i++)
	  {
	    ocoef[1] += coef[ncoef-degxy+i+nslit*ncoef] * poly;
	    poly = poly * y[j];
	  }

	istmp = (NINT(s[j]));
	TCEWRI(tide, nyrow, col[0], &istmp);
	iypxl = NINT( (y[j] - start[1])/step[1] + 1);
	TCEWRI(tide, nyrow, col[1], &iypxl);
	TCEWRD(tide, nyrow, col[2], &y[j]);


	/*
	if (ocoef[1] > 0.0)
	  for (i=1; i<=ncoef-(degy+degxy); i++)
	    TCEWRD(tide, nyrow, ocol[i-1], &ocoef[i-1]);
	else
	  for (i=1; i<=ncoef-(degy+degxy); i++)
	    TCEWRD(tide, nyrow, ocol[i-1], &dnull);
	    */
	for (i=1; i<=ncoef-(degy+degxy); i++)
	  TCEWRD(tide, nyrow, ocol[i-1], &ocoef[i-1]);

	if(j == 0 || s[j+1] != s[j])
	  nslit++;
	if(nslit == 1 || s[j+1] != s[j]){
	  sprintf(descname,"coefslit%d",istmp);
	  SCDWRD(tide, descname, &coef[1+(nslit-1)*ncoef], 1, ncoef, &kun);
	}

      }


  i = fit_deg + 1;

  SCDWRI(tide, "DEGX", &i, 1, 1, &kun);
  SCDWRI(tide, "DEGXY", &degxy, 1, 1, &kun);
  SCDWRI(tide, "DEGY", &degy, 1, 1, &kun);

  
  SCDWRD (tide, "LNPIX", &coef[2], 1L, 1L, &kun);
  SCDWRI (tide, "LNDEG", &fit_deg, 1L, 1L, &kun);
  SCDWRI (tide, "LNCOE", &i, 1L, 1L, &kun);

  TCTCLO(tide);
  free_ivector(ocol,0,fit_deg);
  free_dvector(ocoef,0,fit_deg);

  return;
}







