/* @(#)mosdefine.c	8.5 (ESO-IPG) 11/30/94 15:51:43 */
/* @(#)mosdefine.c      6.1.1.4  (LSW)  07/02/94  13:48:49           */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* .COPYRIGHT   (C) 1994 Landessternwarte Heidelberg                 */
/* .IDENT        mosdefine.c                                         */
/* .AUTHOR       Sabine Moehler - LSW Heidelberg - source code       */
/* .AUTHOR       Thomas Szeifert - LSW Heidelberg - upgrade & debug  */
/* .KEYWORDS                                                         */
/* .PURPOSE      Search object with median: (val - med(val)) > thres */ 
/*               Center object with Gaussian fitting                 */
/*               (adapted from mosslin.c)                            */
/* .VERSION      1.0  07-FEB-1994  Implementation                    */
/*               2.0  TS-0497      Many bugs and problems eliminated */
/*-------------------------------------------------------------------*/

#include <math.h> 
#include <tbldef.h>
#include <midas_def.h>
#include <ok.h>

#include <proto_fors.h>


#define MAXLINES 1000
#define MAXSLITS 100

#define GAUSSIAN 1
#define MARK ((float) -9999.9)

int    inull, os_col, o1_col, o2_col, ni_col, s1_col, s2_col, ss_col;
int    Ntot,Npix[2],Method,NseqO,NseqS,upper[MAXSLITS],lower[MAXSLITS];
int    iav, kun, knul, Xbin,Iwin,Width,Tid;
int    min_dist, min_sky;
int    ot_col, obj_typ;

double Start[2],Step[2];

float  thresh,fnull;
char   qualif[5], o_typ[MAXSLITS+1];

int debug = 0;

#ifdef __STDC__
void  center_obj( float *, float *, int *, int ); 
void  search_obj( float *, int *, int, int *);
void  fit_obj( float *, int *, float *, int, int, float*);
void  def_sky( float *, float *, int , int *);
#else
void  center_obj(); 
void  search_obj();
void  fit_obj();
void  def_sky();
#endif

int
main()

{
  int   i, nax, tidslit, imno, select;
  int   col,nslits,nslits_sel,acol,arow,nsort,param[8], null, slit[MAXSLITS];
  int   sl_col, up_col, lo_col;
  
  char  cmethod[5],obj_image[61],slit_table[61];
  char  window_table[61],unit[64],inst[72],line[80],linem[80];

  float *buff,*image, rpar[8],flower,fupper,fbuff;
  double dnull;


  /* get into Midas and get parameters */

  SCSPRO("mosdefine");
   
  SCKGETC("IN_A",1,60,&iav,obj_image);
  SCKGETC("IN_B",1,60,&iav,slit_table);
  SCKGETC("OUT_A",1,60,&iav,window_table);

  SCKGETC("INPUTC",1,4,&iav,cmethod);
  SCKGETC ("QUALIF",  1, 4, &iav, qualif);

  SCKRDI("MIN_DIST",1,1,&iav,&min_dist,&kun,&knul);
  SCKRDI("MIN_SKY",1,1,&iav,&min_sky,&kun,&knul);
  SCKRDI("INPUTI",1,8,&iav,param,&kun,&knul);
  SCKRDR("INPUTR",1,8,&iav,rpar,&kun,&knul);

  /* map input file */
   
  strcpy (inst,"\0");
  strcpy (unit,"\0");
  SCIGET(obj_image,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,
	 2,&nax,Npix,Start,Step,inst,unit,(char **) &image,&imno);

  Method = GAUSSIAN;

  sprintf(linem,"   centering method:    Gaussian");
  SCTPUT(linem);
  
  /* read table MOS to get limits of slitlets  */
  
  TCTOPN(slit_table,F_I_MODE,&tidslit);

  TCMNUL(&inull, &fnull, &dnull);

  TCIGET(tidslit,&col,&nslits,&nsort,&acol,&arow);

  TCLSER(tidslit, "slit", &sl_col);
  TCLSER(tidslit, "ystart", &lo_col);
  TCLSER(tidslit, "yend", &up_col);
  if (toupper (qualif[0]) == 'S')
    TCLSER(tidslit, "ray_typ", &obj_typ);

  nslits_sel = 0;

  for (i=1;i<=nslits;i++) 
    {
      TCSGET(tidslit, i, &select);
      if (select)
        {
	  TCERDR(tidslit, i, sl_col, &fbuff, &null);
	  if (null) slit[nslits_sel] = inull;
	  slit[nslits_sel] = NINT(fbuff);

	  TCERDR(tidslit, i, lo_col, &flower, &null);
	  lower[nslits_sel] = NINT( (flower - Start[1]) / Step[1]); 
	  if (null) lower[nslits_sel] = fnull;

	  TCERDR(tidslit, i, up_col, &fupper, &null);
	  upper[nslits_sel] = NINT( (fupper - Start[1]) / Step[1]); 
	  if (null) upper[nslits_sel] = fnull;

	  if (lower[nslits_sel] < 0) lower[nslits_sel] = 0;
	  if (upper[nslits_sel] >= Npix[1]) 
	    upper[nslits_sel] = lower[nslits_sel] + (Npix[1]-1);


	  if (toupper (qualif[0]) == 'S')
	    TCERDC(tidslit, i, obj_typ, &o_typ[nslits_sel], &null);


	  /*	  
	  printf("slit %4d  lower = %6d upper = %6d\n",slit[nslits_sel],
		 lower[nslits_sel],upper[nslits_sel]); 
		 */

	  nslits_sel++;
	}
    }

  TCTCLO(tidslit);
  
  /* open output table */
  


  if (toupper (qualif[0]) == 'S')
    TCTINI(window_table,F_TRANS,F_O_MODE,8,MAXLINES,&Tid);
  if (toupper (qualif[0]) == 'M')
    TCTINI(window_table,F_TRANS,F_O_MODE,7,MAXLINES,&Tid);

  SCDWRD(Tid,"Pixel",Step,1,1,&kun);

  TCCINI(Tid,D_R4_FORMAT,1,"F6.0" , "None ","Obj_Slit",&os_col);
  TCCINI(Tid,D_R4_FORMAT,1,"F10.2", "Pixel","Obj_Strt",&o1_col);
  TCCINI(Tid,D_R4_FORMAT,1,"F10.2", "Pixel","Obj_End",&o2_col);
  TCCINI(Tid,D_R4_FORMAT,1,"F10.2", "Pixel","Net_Intens",&ni_col);
  if (toupper (qualif[0]) == 'S')
    TCCINI(Tid,D_C_FORMAT,1,"A1", "","Obj_Typ",&ot_col);
  TCCINI(Tid,D_R4_FORMAT,1,"F6.0" , "None ","Sky_Slit",&ss_col);
  TCCINI(Tid,D_R4_FORMAT,1,"F10.2", "Pixel","Sky_Strt",&s1_col);
  TCCINI(Tid,D_R4_FORMAT,1,"F10.2", "Pixel","Sky_End",&s2_col);
  

  /* set parameters */

  thresh  =  rpar[0];
  Width  =   param[0];
  Iwin  =   (int)((param[0]-1)/2.0);
  Xbin  =   param[1];
  
  SCTPUT("");
  sprintf(line,"   Input image: %s   slit table: %s   window table: %s",obj_image,slit_table,window_table);
  SCTPUT(line);
  SCTPUT(linem);
  sprintf(line,"   search window: width (pixels)              %i",Width); 
  SCTPUT(line);
  sprintf(line,"   detection threshold: (adu)                 %6.2f",thresh);
  SCTPUT(line);
  sprintf(line,"   minimum distance; object to sky: (pixels)  %d",min_dist);
  SCTPUT(line);
  sprintf(line,"   minimum number of CCD rows for valid sky:  %d rows",min_sky);
  SCTPUT(line);
  sprintf(line,"   size of median window: (pixels)            %i",Xbin);
  SCTPUT(line);
  
  
  Ntot = Npix[1]+1;
  
  /* center objects and derive sky limits */

  buff = fvector (0,Ntot);

  center_obj(image, buff, slit, nslits_sel);

  free_fvector(buff, 0, Ntot);


  NseqO--;
  NseqS--;
  
  if (debug > 0 ) printf("write descriptors\n");

  SCDWRI(Tid, "NOBJ", &NseqO, 1, 1, &kun);        
  SCDWRI(Tid, "NSKY", &NseqS, 1, 1, &kun);        
  
  if (debug > 0 ) printf("write keywords\n");

  SCKWRI("NOBJ",&NseqO, 1, 1, &kun);        
  SCKWRI("NSKY",&NseqS,1, 1, &kun);

  /* epilogue ... */



  if (debug > 0 ) printf("close tables\n");
  TCSINI(Tid);
  TCTCLO(Tid);
  
  /* can't close buff in case of context SPOL*/


  if (debug > 0 ) printf("end of main\n");

  SCSEPI();
  exit(0);
  return 0;
}

/*--------------------------------------------------------------------------*/
#ifdef __STDC__
void center_obj(float *image, float *rval, int *slit, int nslits)
#else
void center_obj( image, rval, slit, nslits)
  float *image,*rval;
  int   *slit,nslits;
#endif

{

   int    nact, nslit, j, k, jj, objlist[MAXLINES];
   int    center, isum, ibuff = 0;
   float  *sval, sum, obj[4],sky[3],dum,*tmp;
   double scan_pos;
   char   text[80];


   NseqO = 1;
   NseqS = 1;
   SCKRDD("SCAN", 1, 1, &iav, &scan_pos, &kun, &knul); 
   center = (int) ((scan_pos-Start[0])/Step[0]); /* transform world coordinates
                                                 to pixels */
   if ( debug > 0) 
     printf ("   center_obj around position %f = %d pixel\n",scan_pos,center);

   if (debug > 1)
     printf("\nopen arays with length Ntot = %d\n",Ntot);

   sval = fvector(0,Ntot);
   tmp = fvector(0,Ntot);

   /* loop over all slitlets */
   for (nslit=0; nslit < nslits; nslit++) 
     {
       if (slit[nslit] > inull)
	 {
	   obj[0] = (float) slit[nslit];
	 
	   /* median on XBIN columns at chosen position */
	   for (j=0;j<=Ntot;j++) rval[j] = (float) 0.0;


	   for (k=lower[nslit]; k<=upper[nslit]; k++) 
	     {
	       isum = 0;

	       for (j=center-(int)(Xbin/2.0);j<=center+(int)(Xbin/2.0);j++) 
		 {           

		   /* jj = j + Npix[0] * ( k - (int) Start[1]);*/
		   jj = j + Npix[0] * k;

		   if (debug > 2 && k == upper[nslit])
		     printf ("%4d %4d %6d %4d\n",j,k,jj,isum);

		   tmp[isum] = (float) (image[jj]);
		   isum++;

		 }


	       rval[k] = (float) heap_median (isum, tmp);
	       sval[k] = rval[k];

	     }


	   /* now search for object ...  */
	   if (debug > 0)
	     printf("main search \n");
	   search_obj(rval,objlist,nslit,&nact);


	   /* ... fit object ...  */
	   if (debug > 0)
	     printf("main fit \n");
	   if (nact > 0)
	     fit_obj(rval,objlist,obj,nact,nslit, sval);	
	   
    	   sprintf(text,"      slit %2d  - %2d objects", slit[nslit],nact);
	   SCTPUT(text);
	   ibuff = nact;
	   
	   /* ... and define sky */
	   if (debug > 0)
	     printf("main sky \n");
	   def_sky(sval, sky, nslit, slit);

	 }

     }

   /*can't close these vectors*/


   free_fvector(tmp,0,Ntot);
   free_fvector(sval,0,Ntot);

   if (debug > 0)
     printf("... end of center_obj \n");
   return;
}
/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void search_obj( float *rval, int *objlist, int nslit, int *nact)
#else
void search_obj( rval, objlist, nslit, nact )
  float *rval;
  int   *objlist,nslit,*nact;
#endif

{
  int number = 0,j,i,imax;
  float med,diff,max_intens,Thresh;
  
  if ( debug > 0) printf ("search_obj");

  if (debug > 1)
    printf("loop %d %d \n",lower[nslit]+Iwin,upper[nslit]-Iwin);

  for (j=lower[nslit]+Iwin;j<=upper[nslit]-Iwin;j++) 
    {
      

      med = heap_median(Width,&rval[j-Iwin]);


      diff = rval[j]- med;
	      /*subtract median*/
      if (med<0.) med = -med;
      Thresh = (thresh>0.) ? thresh : -thresh*med;


      if (diff > Thresh) 
	{
	  imax = j;
	  max_intens = rval[j];

	  for (i=j-Iwin;i<=j+Iwin;i++){ /* compute max in 2*Iwin+1 */ 
	    if (rval[i] > max_intens ){
	      max_intens = rval[i]; 
	      imax = i;
	    }
	  }

	  if (number == 0){ /* take a first object */
	    objlist[number] = imax; /* a star is born */
	    number++;
	  }
	  else if (imax - objlist[number-1] > Iwin){ /* last one is far away */
	    objlist[number] = imax; /* a star is born */
	    number++;
	  }
	}


    }
  *nact = number;
  if ( debug > 0) printf ("...\n");

  return;
}
/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void fit_obj( float *rval, int *objlist, float *obj, int nact, int nslit,
              float *sval)
#else
void fit_obj( rval, objlist, obj, nact, nslit, sval )
  float *rval,*obj,*sval;
  int   *objlist,nact,nslit;
#endif
{
  int    j,s1,s2,k,jj,ii,index;
  double aold[3], obj_lim, fraction;
  float  back, med; 
  double *xgaus, *ygaus, *a;
  
  if ( debug > 0) printf ("fit_obj ...");

  
  if ( debug > 0) printf ("arrays - length = %d\n",Width);

  xgaus = dvector(1,Width);
  ygaus = dvector(1,Width);	
  a     = dvector(1,3);

  for (jj=0; jj<3; jj++) aold[jj] = 0.0;

  for (jj=0;jj<nact;jj++) 
    {
      index = objlist[jj];

      if ( debug > 0) printf ("fit_obj loop %d objects index = %d\n",jj,index);

      /* Determine local background (but ensure to take it within the 
	 slitlet) */

      if (index-2*Iwin >= lower[nslit] && index+2*Iwin <= upper[nslit])
        { 
	  med = heap_median(Iwin,&rval[index-2*Iwin]); 
	  back = heap_median(Iwin,&rval[index+Iwin]);
	  if (back > med) back = med;
        }
      else if (index-2*Iwin <= lower[nslit]) 
	back = med = heap_median(Iwin,&rval[index-2*Iwin]);   
      else if (index+2*Iwin >= upper[nslit]) 
	back = med = heap_median(Iwin,&rval[index+Iwin]);
      else
	back = med = heap_median( (upper[nslit]-lower[nslit]+1), &rval[0]); 


      /* GAUSSIAN fit of object profile */

      /* guess values for fitnol.c */
      a[1] = (double) (rval[index] - back);    /*fcen*/
      a[2] = (double) (objlist[jj]+1.0);       /*xcen*/
      a[3] = (double) (2*Step[1]);               /*sigm*/
      for (k = 1 , j = index-Iwin ; j<= index+Iwin; j++, k++) 
        {
	  xgaus[k] = (double) (objlist[jj]+k-Iwin);
	  ygaus[k] = (double) (rval[j] - back);
        }	

      fit_gauss(xgaus,ygaus,Width,a);  

      if (debug > 0)
	printf("I = %6.1f  Y = %6.1f FVWM = %6.1f\n",a[1], a[2], a[3]);
      /* Avoid finding the same object several times  */
      /* new object if difference in x > 0.5 -- otherwise many double 
	 identifications  TS*/
      if ( (int) (a[2] - aold[1] + 0.5) > 0)
        { 
	  for (ii=0; ii<=2; ii++) aold[ii] = a[ii+1];

/* object limits are defined at position where I(x) = fraction * I(x0) */
/* We use the Gaussian fit to compute the integration limits */
	  SCKRDD("INT_LIM",1,1,&iav,&fraction,&kun,&knul);
	  obj_lim = sqrt(-2*log(fraction));

	  obj[1] = (float)((int) ((a[2]-obj_lim*a[3])+0.5) - 1);/*lower limit*/
	  if (obj[1] < (float) lower[nslit]) 
	    obj[1] = (float) lower[nslit];

	  obj[2] = (float)((int)  (a[2]+obj_lim*a[3] +0.5)-1);/*upper limit*/
	  if (obj[2] > (float) upper[nslit]) 
	    obj[2] = (float) upper[nslit];

	  obj[3] = (float) a[1];                     /* net peak intensity */

/* mark object's position for later sky determination (ensure that to stay 
   within slitlet) */

	  s1 = ((int) obj[1]) - min_dist;
	  s2 = ((int) obj[2]) + min_dist - 1; 
	  for (j = s1; j<= s2; j++) 
	    {
	      if (j < lower[nslit]) s1++;
	      if (j > upper[nslit]) s2--;
	    } 
	  for (j = s1; j<= s2; j++) sval[j] = MARK;



	  TCEWRR(Tid, NseqO, os_col, &obj[0]);
	  obj[1] = obj[1] * Step[1] + Start[1]; 
	  obj[2] = obj[2] * Step[1] + Start[1]; 
	  TCEWRR(Tid, NseqO, o1_col, &obj[1]);
	  TCEWRR(Tid, NseqO, o2_col, &obj[2]);
	  TCEWRR(Tid, NseqO, ni_col, &obj[3]);

	  if (toupper (qualif[0]) == 'S')
	    TCEWRC(Tid, NseqO, ot_col, &o_typ[nslit]);
	  NseqO++;  
        }
    }                 

  free_dvector(xgaus, 1, Width);
  free_dvector(ygaus, 1, Width);
  free_dvector(a, 1, 3);

  if ( debug > 0) printf ("...\n");

  return;
}
/*---------------------------------------------------------------------------*/
/*
It's a little unclear to me what def_sky is doing.  
In star window sval is set to sval[i] = MARK - here we check for sval > MARK
but one could do this much easier!!! (TS0497)
*/
#ifdef __STDC__
void def_sky(float *sval, float *sky, int nslit, int *slit)
#else
void def_sky( sval, sky, nslit, slit )
  float *sval,*sky;
  int   nslit,*slit;
#endif

{
  int   skystrt,skyend,nsky,j,slitend,found;
  char  text[80];

  if ( debug > 0) printf ("def_sky");

  slitend = upper[nslit]; 
  skystrt = (int) 0;
  skyend = (int) 0;
  nsky = (int) 0;
  found = (int) 0;

  for (j=lower[nslit];j<=upper[nslit];j++) 
    {
      if (sval[j] > MARK && j < slitend)
        {
	  if (nsky == 0) skystrt = j;
	  nsky++;    
        }
      else if ((sval[j] <= MARK || j == slitend) && nsky!=0)
        {
	  skyend = j;
	  nsky++;
	  if (nsky > min_sky) 
	    {
	      sky[0] = (float) (skystrt * Step[1] + Start[1]);
	      sky[1] = (float) (skyend * Step[1] + Start[1]);
	      sky[2] = (float) slit[nslit];
	      TCEWRR(Tid, NseqS, s1_col, &sky[0]);
	      TCEWRR(Tid, NseqS, s2_col, &sky[1]);
	      TCEWRR(Tid, NseqS, ss_col, &sky[2]);
	      NseqS++;
	      nsky = 0;
	      found = 1;
	    }
	  else if (nsky <= min_sky) nsky = 0;
        }
    }


  if (nsky <= min_sky && found == 0)
    {
      sprintf(text, "   No sky found in slitlet %d", slit[nslit]);
      SCTPUT(text);
    }

  if ( debug > 0) printf ("... end of def_sky\n");

  return;
}








