/* 
   mosrebifors.c

   Thomas Szeifert            LSW-Heidelberg           21. February 1999
   together with Walter Fuertig, Walter Seifert, Otmar Stahl 


   rebinning of 2d spectra for


   Thomas Szeifert, LSW Heidelberg, 1999-02-21
   warning : have changed all to normalized coordinates for numerical reasons
*/
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>

#include <tbldef.h>
#include <midas_def.h>
#include <proto_fors.h>



#define LINEAR		0
#define QUADRATIC	1
#define SPLINE		2

int debug = 0;



int
main()
{
  int i,j,k,l;
  int kact,knul,kun,arow,acol,kbyte,kno;
  char in_table1[81],in_table2[81],out_frame[81],ext_frame[81],in_frame[81],
    cunit[48], text[81],ctype[80];

  int method;

  /* mos slitlet table */
  int tid1, icol,irow,isort,icol1[4], upper, lower;
  float *s1,*o1,*ys1,*ye1;
  
  /* mos fit parameters */
  int tid2,degx,degy,degxy,poltyp,fitdeg,*sel1,ilog;
  double *dcoef,*x1,*y1,*w1,*w2,*y2,tmp,centerxy[3];
  char descname[81];

  /* input & output image */
  int fid1,fid2,npix1[2],npix2[2];
  double start1[2],step1[2],start2[2],step2[2];
  float rebpar[4];
  float *f1,*f2;

  /* extracted images */
  int nobj, *slit, *ypos,naxis=1,numdim;


  SCSPRO("mosrebifors");

  if (debug > 0)
    printf("read keywords\n");

  SCKGETC ("IN_A",   1, 80, &kact, in_frame);
  if (debug > -1)
    printf("input image          %s\n",in_frame);

  SCKGETC ("OUT_A",   1, 80, &kact, out_frame);
  if (debug > -1)
    printf("output image         %s\n",out_frame);

  SCKRDI( "INPUTI", 1, 1, &kact, &method, &kun, &knul );

  SCKGETC ("IN_B",   1, 80, &kact, in_table2);
  if (debug > -1)
    printf("coeficients in table %s\n",in_table2);

  SCKGETC ("IN_C",   1, 80, &kact, in_table1);
  if (debug > -1)
    printf("slitlets in table    %s\n",in_table1);





  if (debug > 1)
    printf("read table %s\n",in_table1);

  /*
    
    READ THE INPUT TABLE with the locations of the mos slitlets
    
  */

  TCTOPN (in_table1, F_I_MODE, &tid1);
  TCIGET (tid1,&icol, &irow, &isort, &acol, &arow);

  TCCSER (tid1, ":SLIT", &icol1[0]);
  TCCSER (tid1, ":XOFFSET", &icol1[1]);
  TCCSER (tid1, ":YSTART", &icol1[2]);
  TCCSER (tid1, ":YEND", &icol1[3]);

  s1 = fvector(0,irow);
  o1 = fvector(0,irow);
  ys1 = fvector(0,irow);
  ye1 = fvector(0,irow);
  sel1 = ivector(0,irow);


  for (i = 0; i < irow; i++){
    TCSGET(tid1,i+1,&sel1[i]);  
    TCERDR(tid1,i+1,icol1[0],&s1[i],&knul);
    TCERDR(tid1,i+1,icol1[1],&o1[i],&knul);
    TCERDR(tid1,i+1,icol1[2],&ys1[i],&knul);
    TCERDR(tid1,i+1,icol1[3],&ye1[i],&knul);
  }

  TCTCLO(tid1);

  if (debug > 1)
    printf("read table %s\n",in_table2);

  /*
    
    READ THE INPUT TABLE with the wave length calibration
    
  */

  TCTOPN (in_table2, F_I_MODE, &tid2);

  /*
  SCDRDI(tid2, "DEGX", 1, 1, &kact, &degx, &kun, &knul);
  SCDRDI(tid2, "DEGXY", 1, 1, &kact, &degxy, &kun, &knul);
  SCDRDI(tid2, "DEGY", 1, 1, &kact, &degy, &kun, &knul);
  SCDRDI(tid2, "POLTYPE", 1, 1, &kact, &poltyp, &kun, &knul);
  */

  if (SCDFND(tid2, "NORMCOORD", ctype, &kno, &kbyte) == 0)
    {
      SCDRDI(tid2, "NORMCOORD", 1, 1, &kact, &ilog, &kun, &knul);
      SCDRDD(tid2, "CENTERXY", 1, 3, &kact, centerxy, &kun, &knul );
      printf("mosrebifors: coordinates are transformed to -1,1 interval\n");
      SCDRDI(tid2, "NORMCOORD", 2, 1, &kact, &degx, &kun, &knul);
      SCDRDI(tid2, "NORMCOORD", 3, 1, &kact, &degxy, &kun, &knul);
      SCDRDI(tid2, "NORMCOORD", 4, 1, &kact, &degy, &kun, &knul);
      SCDRDI(tid2, "NORMCOORD", 5, 1, &kact, &poltyp, &kun, &knul);
    }
  else
    {
      ilog = 0;
      centerxy[0] = centerxy[1] = 0.0;
      centerxy[2] = 1.0;
      printf ("incompatible coeficient table - let's crash now\n");
      SCSEPI();
      exit(1);
    }


  fitdeg = degx + degxy + degy;

  dcoef = dvector(1,fitdeg*irow);

  for (i = 0 ; i < irow ; i++)
    {
      if (sel1[i]){
	sprintf(descname,"coefslit%1.0f",s1[i]);
	SCDRDD( tid2, descname, 1, fitdeg, &kact, &dcoef[fitdeg*i+1], &kun, 
		&knul );
      }
    }






  SCFOPN (in_frame,D_R4_FORMAT,0,F_IMA_TYPE,&fid1);
  SCDRDI (fid1,"NPIX",1,2,&kact,npix1,&kun,&knul);
  SCDRDD (fid1,"START",1,2,&kact,start1,&kun,&knul);
  SCDRDD (fid1,"STEP",1,2,&kact,step1,&kun,&knul);


  /*  
  SCKRDD( "REBSTRT", 1, 1, &kact, &start2[0], &kun, &knul );
  start2[1] =  start1[1];
  SCKRDD( "REBSTP", 1, 1, &kact, &step2[0], &kun, &knul );
  step2[1] =  step1[1];
  SCKRDD( "REBEND", 1, 1, &kact, &tmp, &kun, &knul );
  */

  SCKRDR( "INPUTR", 1, 3, &kact, rebpar, &kun, &knul );
  start2[0] = rebpar[0];
  start2[1] =  start1[1];
  step2[0] = rebpar[2];
  step2[1] =  step1[1];
  tmp = rebpar[1];
  npix2[0] = NINT( (tmp - start2[0]) / step2[0] ) + 1;

  printf("REBSTRT = %f REBEND = %f REBSTP = %f \n",start2[0],tmp,step2[0]);

  f1 = fvector(0,npix1[0]);
  x1 = dvector(0,npix1[0]);
  y1 = dvector(0,npix1[0]);
  w1 = dvector(0,npix1[0]);

  f2 = fvector(0,npix2[0]);
  w2 = dvector(0,npix2[0]);

  for ( i = 0; i < npix1[0]; i++ )
    {
      x1[i] = ((start1[0]) + step1[0] * i);
      if (ilog == 1)
	x1[i] = (x1[i]  - centerxy[0]) / centerxy[2];
    }
  for ( i = 0; i < npix2[0]; i++ )
    w2[i] = (start2[0] + step2[0] * i);

  SCDFND(fid1, "YPOS", ctype, &nobj, &kbyte);
  if (ctype[0] != ' ')
    {
      /*

	rebin extracted spectra
	
      */

      printf("nobj = %d  kbyte = %d\n",nobj,kbyte);
      npix2[1] = 1;
      
      ypos = ivector(0,nobj);
      slit = ivector(0,nobj);
      
      SCDRDI(fid1, "YPOS", 1, nobj, &kact, ypos, &kun, &knul);
      SCDRDI(fid1, "SLIT", 1, nobj, &kact, slit, &kun, &knul);
      
      for (j = 0 ; j < nobj ; j++){
	/*
	sprintf(ext_frame,"%s_%1d_%1d\0",out_frame,slit[j],ypos[j]);
	*/
	sprintf(ext_frame,"%s%04i\0",out_frame,j+1);
	printf("%s \n",ext_frame);

	SCFCRE (ext_frame, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, 
		npix2[0]*npix2[1], &fid2);

	for ( k = 0; k < npix1[0]; k++ ){
	  y1[k]  = step1[1] * ypos[j] + start1[1];
	  if (ilog == 1)
	    y1[k] = (y1[k] - centerxy[1]) / centerxy[2]; 
	}

	printf("coefslit%1.0f %d\n",slit[j],j);
	sprintf(descname,"COEFSLIT%1d\0",slit[j]);
	SCDRDD( tid2, descname, 1, fitdeg, &kact, &dcoef[1], &kun, 
		&knul );

	if (poltyp == 0)
	  eval_disp_2D_rebi (x1,  y1, w1, npix1[0], &dcoef[0], fitdeg, 
			     "POLY",degy,degxy);
	else if (poltyp == 1)
	  eval_disp_2D_rebi(x1,  y1, w1, npix1[0], &dcoef[0], fitdeg, 
			    "CHEB",degy,degxy);
	else
	  eval_disp_2D_rebi(x1,  y1, w1, npix1[0], &dcoef[0], fitdeg, 
			    "LEGE",degy,degxy);
	
	SCFGET( fid1, j*npix1[0]+1, npix1[0], &knul, (char*) f1 );
	rebin( w1, w2, f1, f2, npix1[0], npix2[0] , step2[0], method);
	SCFPUT( fid2, 1, npix2[0], (char*) f2 );

	/*	SCDCOP( fid1, fid2, 3, 0 );*/ 
	SCDWRD( fid2, "START", start2, 1, 1, &kun );
	SCDWRD( fid2, "STEP", step2, 1, 1, &kun );
	SCDWRI( fid2, "NPIX", npix2, 1, 1, &kun );
	SCDWRI( fid2, "NAXIS", &naxis, 1, 1, &kun );	
	SCDCOP( fid1, fid2, 4, "IDENT");
	SCDCOP( fid1, fid2, 4, "LHCUTS");
	SCDWRC (fid2, "CUNIT", 1, cunit, 1, 48, &kun);

	SCFCLO(fid2);


	strcat(ext_frame,"ERR\0");
	printf("%s\n",ext_frame);

	SCFCRE (ext_frame, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, 
		npix2[0]*npix2[1], &fid2);

	SCFGET( fid1, j*npix1[0]+1 + nobj*npix1[0], npix1[0], &knul, 
		(char*) f1 );
	rebin( w1, w2, f1, f2, npix1[0], npix2[0] , step2[0], method);
	SCFPUT( fid2, 1, npix2[0], (char*) f2 );

	/*	SCDCOP( fid1, fid2, 3, 0 );*/ 
	SCDWRD( fid2, "START", start2, 1, 1, &kun );
	SCDWRD( fid2, "STEP", step2, 1, 1, &kun );
	SCDWRI( fid2, "NPIX", npix2, 1, 1, &kun );
	SCDWRI( fid2, "NAXIS", &naxis, 1, 1, &kun );
	SCDCOP( fid1, fid2, 4, "IDENT");
	SCDCOP( fid1, fid2, 4, "LHCUTS");
	SCDWRC (fid2, "CUNIT", 1, cunit, 1, 48, &kun);

	SCFCLO(fid2);

      }
		

    }
  else
    {
      /*
	
	rebin 2D spectra
	
      */

      npix2[1] =npix1[1];

      SCFCRE (out_frame, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, npix2[0]*npix2[1] ,
	      &fid2);


      for (i = 0; i < irow; i++){
	if (sel1[i]){
	  lower =  NINT((ys1[i]-start1[1])/step1[1])+1;
	  upper =  NINT((ye1[i]-start1[1])/step1[1])+1;
	  
	  sprintf (text,"slit %2.0f lower %d upper %d ",s1[i],lower ,upper);
	  SCTDIS(text,80);
	  
	  for (j = lower; j <= upper; j++){
	    
	    for ( k = 0; k < npix1[0]; k++ ){
	      y1[k]  = step1[1] * j + start1[1];
	      if (ilog == 1)
		y1[k] = (y1[k] - centerxy[1]) / centerxy[2]; 
	    }
	    
	    
	    if (poltyp == 0)
	      eval_disp_2D_rebi (x1,  y1, w1, npix1[0], &dcoef[fitdeg*i], 
				 fitdeg, "POLY",degy,degxy);
	    else if (poltyp == 1)
	      eval_disp_2D_rebi(x1,  y1, w1, npix1[0], &dcoef[fitdeg*i], fitdeg, 
				"CHEB",degy,degxy);
	    else
	      eval_disp_2D_rebi(x1,  y1, w1, npix1[0], &dcoef[fitdeg*i], fitdeg, 
				"LEGE",degy,degxy);
	    
	    SCFGET( fid1, (j-1)*npix1[0]+1, npix1[0], &knul, (char*) f1 );
	    rebin( w1, w2, f1, f2, npix1[0], npix2[0] , step2[0], method);
	    SCFPUT( fid2, (j-1)*npix2[0]+1, npix2[0], (char*) f2 );
	    
	  }
	  
	}
      }
      printf("\n");
      SCDCOP( fid1, fid2, 3, 0 ); 
      SCDWRD( fid2, "START", start2, 1, 2, &kun );
      SCDWRD( fid2, "STEP", step2, 1, 2, &kun );
      SCDWRI( fid2, "NPIX", npix2, 1, 2, &kun );
      SCDCOP( fid1, fid2, 4, "NAXIS");
      SCDCOP( fid1, fid2, 4, "IDENT");
      SCDCOP( fid1, fid2, 4, "LHCUTS");
      SCDWRC (fid2, "CUNIT", 1, cunit, 1, 48, &kun);

      SCFCLO(fid2);
    }


  TCTCLO(tid2);  
  SCFCLO(fid1);

  
  SCSEPI();
  exit(0);
}

















