/* @(#)mosrebin1d.c	8.6 (ESO-IPG) 11/30/94 15:56:47 */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.IDENT        mosrebin1d.c
.MODULE       main program -- mosrebin1D.exe
.LANGUAGE     C
.AUTHOR       Cristian Levin - ESO La Silla
              Otmar Stahl, LSW, rebin "image" of extracted spectra !!!
.PURPOSE      linear, quadratic an spline rebinning.
.KEYWORDS     rebinning.
.VERSION 1.0  1-Apr-1991   Implementation
.ENVIRONMENT  UNIX
------------------------------------------------------------*/

#include <tbldef.h>
#include <midas_def.h>

#include <proto_fors.h>
#include <proto_mos.h>

#define MAXINTENS   99999.0
#define MININTENS  -99999.0 

#define LINEAR    0
#define QUADRATIC 1
#define SPLINE    2

char    Coerbr[61];
double  Starti, Stepi;
double  Starto, Endo, Stepo;
int     Method, NpixOut, Npix[2], kun, knul;

float   Cutsi[2];
float   Imax = MININTENS, Imin = MAXINTENS;


main ()

{

  char inframe[60], rootima[40], outima[60], output[61];
  int dunit;
  char number[6];

  int status, actvals, null;	/* Std Interfaces variables */
  int naxis, ypos[101], slit[101], i, id;
  int  dnull, imno, imno2;

  double *x, *win, *wout;
  float *yin, *yout, rebpar[3];

  SCSPRO ("mosrebin1d");

  status = SCKGETC ("IN_A", 1, 60, &actvals, inframe);
  if (status)
    SCTPUT ("Error while reading IN_A");

  status = SCKGETC ("IN_B", 1, 60, &actvals, Coerbr);
  if (status)
    SCTPUT ("Error while reading IN_B");

  status = SCKGETC ("OUT_A", 1, 40, &actvals, rootima);
  if (status)
    SCTPUT ("Error while reading OUT_A");

  status = SCKRDI ("INPUTI", 1, 1, &actvals, &Method, &kun, &knul);

  /* open "line-by-line" frame and read descriptors */

  if ( SCFOPN( inframe, D_R4_FORMAT, 0, F_IMA_TYPE, &imno ) != 0 ) {
    sprintf( output, "Frame %s invalid...", inframe );
    SCTPUT( output );
    return;
  }
  
  SCDRDI( imno, "NAXIS", 1, 1, &actvals, &naxis, &dunit, &dnull );
  SCDRDI( imno, "NPIX", 1, 2, &actvals, Npix, &dunit, &dnull );
  if ( naxis == 1 )
    Npix[1] = 1;
  SCDRDD( imno, "START", 1, 1, &actvals, &Starti, &dunit, &dnull );
  SCDRDD( imno, "STEP", 1, 1, &actvals, &Stepi, &dunit, &dnull );
  SCDRDR( imno, "LHCUTS", 1, 2, &actvals, Cutsi, &dunit, &dnull );
  
  
  /* ypos is the original y-position of the spectra (before extraction) */
  
  SCDRDI (imno, "YPOS", 1, NINT(Npix[1]/2), &actvals, ypos, &dunit, &dnull);
  SCDRDI (imno, "SLIT", 1, NINT(Npix[1]/2), &actvals, slit, &dunit, &dnull);

  for (i = 0 ; i <  NINT(Npix[1]/2) ; i++)
    printf("%d %d\n",ypos[i],slit[i]);

/*    if ( ! file_exists(Coerbr, ".tbl") )
      SCETER( 9 , "Coefficients table couldn't be opened. Stop.\n"); */

  TCTOPN(Coerbr, F_IO_MODE, &id);

  SCDRDR( id, "REBPAR", 1, 3, &actvals, rebpar, &dunit, &dnull );
  Starto = (double) rebpar[0];
  Endo   = (double) rebpar[1];
  Stepo  = (double) rebpar[2];
  
  if ( fabs(Stepo) > 1.0)  
    NpixOut = (int) ((Endo - Starto) / Stepo + 0.5);
  else
    NpixOut = 0;

  TCTCLO( id );
  
  mos_initdisp(Coerbr, "OLD", 0);
  
  x    = dvector(0, Npix[0] - 1);
  win  = dvector(0, Npix[0] - 1);
  yin  = fvector(0, Npix[0] - 1);
  /*    yout = fvector(0, NpixOut - 1); */
  

  /* TS 250898 check if REBSTRT ... are defined */
  if (NpixOut <= 2){
    status = mos_readdisp(ypos[0],slit[0]);
    x[0] = Starti; 
    x[1] = Starti + Stepi * (Npix[0]-1);
    mos_eval_disp(x, win, 2);
    Starto = win[0];
    Endo   = win[1];
    printf ("y = %d slit = %d -> %f %f %f %f\n", ypos[0],slit[0], x[0],  x[1] ,win[0], win[1]); 
    Stepo  = (Endo - Starto) / (double) Npix[0];
    NpixOut = (int) ((Endo - Starto) / Stepo + 0.5);

    Starto =  Starto - ((Stepo / 10.0) * (double) NpixOut) ;
    Endo   =  Endo   + ((Stepo / 10.0) * (double) NpixOut) ;
    NpixOut = (int) ((Endo - Starto) / Stepo + 0.5);

    printf("... set rebin parameters to %f,%f,%f\n",Starto,Endo,Stepo); 

    SCKWRD("REBSTRT", &Starto, 1, 1, &dunit);
    SCKWRD("REBEND", &Endo, 1, 1, &dunit);
    SCKWRD("REBSTP", &Stepo, 1, 1, &dunit);


  }


  for ( i = 0; i < Npix[0]; i++ )
    x[i]  = (double) (Starti + Stepi * i);

  wout = dvector(0, NpixOut - 1);
  for ( i = 0; i < NpixOut; i++ )
    wout[i] = (double) (Starto + Stepo * i);

  /* loop over all rows, and images  */

  for (i = 0; i < Npix[1]/2; i++)
    {

      strcpy (outima, rootima);
      sprintf (number, "%04i", i + 1);
      strcat (outima, number);
      sprintf (output,"%s", outima);
      SCTPUT(output);

      SCFGET(imno, i*Npix[0]+1, Npix[0], &null, (char *)yin);

      status = mos_readdisp(ypos[i],slit[i]);

      if (status == 0) {
	mos_eval_disp(x, win, Npix[0]);      
	
	SCIPUT(outima, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, 
	       1, &NpixOut, &Starto, &Stepo, 
	       "artificial Image", "raw flux        Angstrom", (char **)&yout, &imno2);
	
	rebin (win, wout, yin, yout, Npix[0], NpixOut, Stepo, Method);
	
	SCFCLO(imno2);
      }

      /* same for the standard deviation */

      strcpy (outima, rootima);
      sprintf (number, "%04iERR", i + 1);
      strcat (outima, number);
      sprintf (output,"%s", outima);
      SCTPUT(output);
      
      SCFGET(imno, (i+Npix[1]/2)*Npix[0]+1, Npix[0], &null, (char *)yin);
	
      status = mos_readdisp(ypos[i],slit[i]);
      
      if (status == 0) {
	mos_eval_disp(x, win, Npix[0]);      
	
	SCIPUT(outima, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, 
	       1, &NpixOut, &Starto, &Stepo, 
	       "artificial Image", "raw flux        Angstrom", (char **)&yout, &imno2);
	
	rebin (win, wout, yin, yout, Npix[0], NpixOut, Stepo, Method);
	
	SCFCLO(imno2);
      }
    }

  finishdisp();

  /* Close everything and good bye */

  free_dvector(x, 0, Npix[0] - 1);
  free_dvector(win, 0, Npix[0] - 1);
  free_dvector(wout, 0, NpixOut - 1);
  free_fvector(yin, 0, Npix[0] - 1);

  SCFCLO (imno), SCSEPI ();
}













