/*
overslitlet.c  

overplots slitlet positions of the FORS MOS segment into the overlay chanel 
of the display window. An image has to be displayed first. We presume you have
a suitable table with slitlet positions! 



                              Thomas Szeifert - LSW Heidelberg - June 10, 1997

*/
#ifndef PI
#define PI  (float) acos(-1.0)
#endif

#include <math.h> 
#include <midas_def.h>
/*
  #include <plot_def.h>
  */
#include <tbldef.h>


#define STRGLEN      80
#define inull         0
#define KEY_EXIT     32        /* AGL key code for middle mouse buton */

 
int debug;



#ifdef __STDC__
void overslit (int *, float *, float *, float *, int, float, float, int, 
	       float *);
#else
void overslit ();
#endif


main()
{
  /* general: */
  int iav,kun,knul;
  int allcol,allrow;
  int i,j,iselect;

  /* related to keywords: */
  int ipar[8];
  char star_table[STRGLEN],slit_table[STRGLEN],qualif[4];
  float fpar[8],angle,length;

  /* related to star table: */
  int tidstar,nrow,ncol,nsort,icol[3],nsel=0,*idstar;
  float *xstar,*ystar;

  /* related to slitlet table: */
  int tidslit,mrow,mcol,msort,jcol[4],*idslit;
  float *xslit,*yslit,*wslit;

  /* related to plot parameters: */
  char plcmeta[STRGLEN],devnam[STRGLEN];
  int actval,plmode=-1,icolour;
  float ssize=2.0;
  float  wcframx[4], wcframy[4],clp[4];

  SCSPRO("overslitlet");




  /*
    
    Here we read the keywords:

  */

  SCKRDI("DEBUG",1,1,&iav,&debug,&kun,&knul);

  if (debug > 0) printf ("... read keywords \n");

  SCKGETC("IN_A",1,STRGLEN,&iav,star_table);
  SCKGETC("IN_B",1,STRGLEN,&iav,slit_table);

  SCKRDI("INPUTI",1,8,&iav,ipar,&kun,&knul);
  SCKRDR("INPUTR",1,8,&iav,fpar,&kun,&knul);
  SCKRDR("ROT_ANG",1,1,&iav,&angle,&kun,&knul);
  SCKRDR("SLIT_LENGTH",1,1,&iav,&length,&kun,&knul);

  SCKGETC ("QUALIF",  1, 4, &iav, qualif);




  /*
    
    Here we read or set the plot parameters:

  */

  if (debug > 0) printf ("... read plot parameters \n");


  wcframx[0] = fpar[0];
  wcframx[1] = fpar[2];
  wcframx[2] = fpar[4];
  wcframx[3] = fpar[6];

  wcframy[0] = fpar[1];
  wcframy[1] = fpar[3];
  wcframy[2] = fpar[5];
  wcframy[3] = fpar[7];


  PCKWRR ("XWNDL", 4, wcframx);
  PCKWRR ("YWNDL", 4, wcframy);


  icolour=4;
  PCKWRI("COLOUR",1,&icolour);


  /*  look! there are hidden parameters! if ACTION != DISP CLPL ist changed
      by PCOPEN and world coordinates won't fit the display data */
  PCKRDR ("CLPL", 4 ,&actval, clp);
  if (debug > 2)
    printf ("CLPL = %10.3f %10.3f %10.3f %10.3f\n",clp[0],clp[1],clp[2],
	    clp[3]);



  /* parameter names are ok - documentation (midas manual) is wrong*/
  PCKRDC ("DNAM", 60, &actval, devnam);  
  PCKRDC ("PLNAM", STRGLEN, &actval, plcmeta);


  if (debug > 0) printf ("... plcmeta = %s  devnam = %s\n",plcmeta,devnam); 

  if (debug > 1) printf("PCOPEN\n");

  PCOPEN (" ",plcmeta, 0, &plmode);
  

  if (debug > 2 ){
    PCKRDR ("CLPL", 4 ,&actval, clp);
    printf ("CLPL = %10.3f %10.3f %10.3f %10.3f\n",clp[0],clp[1],clp[2],
	    clp[3]);
  }


  PCFRAM( wcframx, wcframy, "X_axes", "Yaxes"); 
    

  PCKWRR ("SSIZE", 1, &ssize); 








  /*
    
    Here we read the table with the star's positions:

  */

  if (debug > 0) printf ("... read star table \n");

  TCTOPN(star_table,F_I_MODE,&tidstar);
  TCIGET(tidstar,&ncol,&nrow,&nsort,&allcol,&allrow);

  idstar =  (int *) osmmget((nrow+1)*sizeof(int));
  xstar =  (float *) osmmget((nrow+1)*sizeof(float));
  ystar =  (float *) osmmget((nrow+1)*sizeof(float));


  TCLSER(tidstar, "ICENT", &icol[0]);
  TCLSER(tidstar, "XCEN", &icol[1]);
  TCLSER(tidstar, "YCEN", &icol[2]);


  if (debug > 1) printf ("... read selected \n");
  for (i=1;i<=nrow;i++) 
    {
      if (debug > 3) printf ("... read row %d of %d\n",i,nrow);
      TCSGET(tidstar, i, &iselect);
      if (iselect){
	TCERDI(tidstar, i, icol[0], &idstar[nsel] , &knul);
	TCERDR(tidstar, i, icol[1], &xstar[nsel], &knul);
	TCERDR(tidstar, i, icol[2], &ystar[nsel], &knul);
	nsel++;
      }
    }
 
  TCTCLO(tidstar);

  if (debug > 2) {
    for (i=0;i<nsel;i++)
      {
	printf ("      idstar = %4d  xstar = %7.1f  ystar = %7.1f  \n",
		idstar[i],xstar[i],ystar[i]);
      }
  }

  if (debug > 0) printf ("... plot selected stars (ESO  blau)\n");
  if (debug > 1)    printf("PCDATA\n");
  PCDATA(2,0,0,xstar,ystar,0.0,nsel);
  


  osmmfree((char *) idstar);
  osmmfree((char *) xstar);
  osmmfree((char *) ystar);





  /*
    
    Here we read the table with the slitlet's positions:

  */


  if (debug > 0) printf ("... read slitlet table \n");

  TCTOPN(slit_table,F_I_MODE,&tidslit);
  TCIGET(tidslit,&mcol,&mrow,&msort,&allcol,&allrow);

  idslit =  (int *) osmmget((mrow+1)*sizeof(int));
  xslit =  (float *) osmmget((mrow+1)*sizeof(float));
  yslit =  (float *) osmmget((mrow+1)*sizeof(float));
  wslit = (float *) osmmget((mrow+1)*sizeof(float));

  TCLSER(tidslit, "IDSLIT", &jcol[0]);
  TCLSER(tidslit, "XSLIT", &jcol[1]);
  TCLSER(tidslit, "YSLIT", &jcol[2]);
  TCLSER(tidslit, "WSLIT", &jcol[3]);


  if (debug > 1) printf ("... read table entries \n");
  for (j=1;j<=mrow;j++) 
    {
	TCERDI(tidslit, j, jcol[0], &idslit[j-1] , &knul);
	TCERDR(tidslit, j, jcol[1], &xslit[j-1] , &knul);
	TCERDR(tidslit, j, jcol[2], &yslit[j-1] , &knul);
	TCERDR(tidslit, j, jcol[3], &wslit[j-1] , &knul);
    }
 
  TCTCLO(tidslit);

  if (debug > 2) {
    for (j=0;j<mrow;j++)
      {
	printf ("      idslit = %4d  xslit = %7.1f  yslit = %7.1f  wslit = %7.1f  \n",	idslit[j],xslit[j],yslit[j],wslit[j]);
      }
  }

  icolour = 5;
  PCKWRI("COLOUR",1,&icolour);

  if (debug > 0) printf ("... plot slitlets       (FORS gelb): \n");

  overslit ( idslit, xslit, yslit, wslit, mrow, angle, length, icolour, fpar);


  osmmfree((char *) idslit);
  osmmfree((char *) xslit);
  osmmfree((char *) yslit);
  osmmfree((char *) wslit);





  PCCLOS(); 

  return SCSEPI();
}


#ifdef __STDC__
void overslit (int *idslit, float *xslit, float *yslit, float *wslit, 
	       int mrow, float angle, float length, int icolour, float *fpar)
#else
     void overslit (idslit, xslit, yslit, wslit, mrow, angle, length, icolour,
		    fpar)
     int *idslit,mrow, icolour;
     float *xslit,*yslit,*wslit,angle,length,*fpar;
#endif
{
  float x[8],y[8],tmp;
  int i;
  
  tmp = 10.0;


  if (angle < 0.0)   angle = angle + 360.0;
  if (angle > 360.0) angle = angle - 360.0;
  angle = PI * angle / 180.0; 

  for (i = 0; i < mrow; i++)
    {
      x[1] = xslit[i] - wslit[i]/2.0 * cos (angle) - length/2.0 *  sin (angle);
      y[1] = yslit[i] - wslit[i]/2.0 * sin (angle) + length/2.0 *  cos (angle);
      x[2] = xslit[i] - wslit[i]/2.0 * cos (angle) + length/2.0 *  sin (angle);
      y[2] = yslit[i] - wslit[i]/2.0 * sin (angle) - length/2.0 *  cos (angle);

      x[0] = x[1] - tmp * length * cos (angle);
      y[0] = y[1] - tmp * length * sin (angle);
      x[3] = x[2] - tmp * length * cos (angle);
      y[3] = y[2] - tmp * length * sin (angle);
      
      x[5] = xslit[i] + wslit[i]/2.0 * cos (angle) - length/2.0 *  sin (angle);
      y[5] = yslit[i] + wslit[i]/2.0 * sin (angle) + length/2.0 *  cos (angle);
      x[6] = xslit[i] + wslit[i]/2.0 * cos (angle) + length/2.0 *  sin (angle);
      y[6] = yslit[i] + wslit[i]/2.0 * sin (angle) - length/2.0 *  cos (angle);

      x[4] = x[5] + tmp * length * cos (angle);
      y[4] = y[5] + tmp * length * sin (angle);
      x[7] = x[6] + tmp * length * cos (angle);
      y[7] = y[6] + tmp * length * sin (angle);
      

      if (debug > 1)
	printf ("%7.1f %7.1f %7.1f %7.1f %7.1f %7.1f \n", xslit[i],yslit[i],
		x[1],y[1],x[2],y[2]); 
      PCDATA (0,1,0,&x[0],&y[0],0.0,4);
      PCDATA (0,1,0,&x[4],&y[4],0.0,4);
    }
}

