/*
overslitlet.c  overplots slitlet positions of the FORS MOS segment



in construction - Thomas Szeifert - LSW Heidelberg - June 4, 1997

*/
#include <stdlib.h>
#include <string.h>
#include <idinumd.h>


#include <math.h> 
#include <midas_def.h>
#include <plot_def.h>
#include <tbldef.h>


#define STRGLEN      80
#define inull         0
#define KEY_EXIT     32        /* AGL key code for middle mouse buton */
#define NOT_OK    -9999
 
int debug;

/*
  #ifdef __STDC__
  void overslit (int *, float *, float *, float *, int, float, float, int);
  #else
  void overslit ();
  #endif
*/

main()
{
  /* general: */
  int iav,kun,knul;
  int allcol,allrow;
  int i,j,iselect;
  /*  float tmp[4],ump[4];*/

  /* related to keywords: */
  int ipar[8];
  char star_table[STRGLEN],slit_table[STRGLEN],qualif[4];
  float fpar[8],angle,length;

  /* related to star table: */
  int tidstar,nrow,ncol,nsort,icol[3],nsel=0,*idstar;
  float *xstar,*ystar;

  /* related to slitlet table: */
  int tidslit,mrow,mcol,msort,jcol[4],*idslit;
  float *xslit,*yslit,*wslit;

  /* related to plot parameters: */
  char plcmeta[STRGLEN],devnam[STRGLEN];
  int actval,plmode=-1,icolour;
  float ssize=2.0;
  float  wcframx[4], wcframy[4];

  SCSPRO("overslitlet");




  /*
    
    Here we read the keywords:

  */

  SCKRDI("DEBUG",1,1,&iav,&debug,&kun,&knul);

  if (debug > 0) printf ("... read keywords \n");

  SCKGETC("IN_A",1,STRGLEN,&iav,star_table);
  SCKGETC("IN_B",1,STRGLEN,&iav,slit_table);

  SCKRDI("INPUTI",1,8,&iav,ipar,&kun,&knul);
  SCKRDR("INPUTR",1,8,&iav,fpar,&kun,&knul);
  SCKRDR("ROT_ANG",1,1,&iav,&angle,&kun,&knul);
  SCKRDR("SLIT_LENGTH",1,1,&iav,&length,&kun,&knul);

  SCKGETC ("QUALIF",  1, 4, &iav, qualif);




  /*
    
    Here we read or set the plot parameters:

  */

  if (debug > 0) printf ("... read plot parameters \n");

  
  PCKRDR ("XAXES", 4, &actval, wcframx);
  PCKRDR ("YAXES", 4, &actval, wcframy);


  icolour=4;
  PCKWRI("COLOUR",1,&icolour);

  
    
  PCKRDC ("DNAME", 60, &actval, devnam);  
  PCKRDC ("PLNAME", STRGLEN, &actval, plcmeta);



  if (debug > 0) printf ("... plcmeta = %s  devnam = %s\n",plcmeta,devnam); 

  if (debug > 1) printf("PCOPEN\n");
  /*
    PCOPEN (" "," ", 1, &plmode);
    */
  /*TS*/
  PCOPEN (" "," ", 0, &plmode);
  PCFRAM( wcframx, wcframy, "X_axes", "Yaxes"); 


  PCKWRR ("SSIZE", 1, &ssize); 








  /*
    
    Here we read the table with the star's positions:

  */

  if (debug > 0) printf ("... read star table \n");

  TCTOPN(star_table,F_I_MODE,&tidstar);
  TCIGET(tidstar,&ncol,&nrow,&nsort,&allcol,&allrow);

  idstar =  (int *) osmmget((nrow+1)*sizeof(int));
  xstar =  (float *) osmmget((nrow+1)*sizeof(float));
  ystar =  (float *) osmmget((nrow+1)*sizeof(float));


  TCLSER(tidstar, "ICENT", &icol[0]);
  TCLSER(tidstar, "XCEN", &icol[1]);
  TCLSER(tidstar, "YCEN", &icol[2]);


  if (debug > 1) printf ("... read selected \n");
  for (i=1;i<=nrow;i++) 
    {
      if (debug > 3) printf ("... read row %d of %d\n",i,nrow);
      TCSGET(tidstar, i, &iselect);
      if (iselect){
	TCERDI(tidstar, i, icol[0], &idstar[nsel] , &knul);
	TCERDR(tidstar, i, icol[1], &xstar[nsel], &knul);
	TCERDR(tidstar, i, icol[2], &ystar[nsel], &knul);
	nsel++;
      }
    }
 
  TCTCLO(tidstar);

  if (debug > 2) {
    for (i=0;i<nsel;i++)
      {
	printf ("      idstar = %4d  xstar = %7.1f  ystar = %7.1f  \n",
		idstar[i],xstar[i],ystar[i]);
      }
  }

  if (debug > 0) printf ("... plot selected stars \n");
  icolour = 5;
  PCKWRI("COLOUR",1,&icolour);
  if (debug > 1)    printf("PCDATA\n");
  PCDATA(2,0,0,xstar,ystar,0.0,nsel);



  osmmfree((char *) idstar);
  osmmfree((char *) xstar);
  osmmfree((char *) ystar);





  /*
    
    Here we read the table with the slitlet's positions:

  */


  if (debug > 0) printf ("... read slitlet table \n");

  TCTOPN(slit_table,F_I_MODE,&tidslit);
  TCIGET(tidslit,&mcol,&mrow,&msort,&allcol,&allrow);

  idslit =  (int *) osmmget((mrow+1)*sizeof(int));
  xslit =  (float *) osmmget((mrow+1)*sizeof(float));
  yslit =  (float *) osmmget((mrow+1)*sizeof(float));
  wslit = (float *) osmmget((mrow+1)*sizeof(float));

  TCLSER(tidslit, "IDSLIT", &jcol[0]);
  TCLSER(tidslit, "XSLIT", &jcol[1]);
  TCLSER(tidslit, "YSLIT", &jcol[2]);
  TCLSER(tidslit, "WSLIT", &jcol[3]);


  if (debug > 1) printf ("... read table entries \n");
  for (j=1;j<=mrow;j++) 
    {
	TCERDI(tidslit, j, jcol[0], &idslit[j-1] , &knul);
	TCERDR(tidslit, j, jcol[1], &xslit[j-1] , &knul);
	TCERDR(tidslit, j, jcol[2], &yslit[j-1] , &knul);
	TCERDR(tidslit, j, jcol[3], &wslit[j-1] , &knul);
    }
 
  TCTCLO(tidslit);

  if (debug > 2) {
    for (j=0;j<mrow;j++)
      {
	printf ("      idslit = %4d  xslit = %7.1f  yslit = %7.1f  wslit = %7.1f  \n",	idslit[j],xslit[j],yslit[j],wslit[j]);
      }
  }

  /*
    overslit ( idslit, xslit, yslit, wslit, mrow, angle, length, icolour);
  */

  osmmfree((char *) idslit);
  osmmfree((char *) xslit);
  osmmfree((char *) yslit);
  osmmfree((char *) wslit);





  PCCLOS(); 

  return SCSEPI();
}

/*
  #ifdef __STDC__
  void overslit (int *idslit, float *xslit, float *yslit, float *wslit, 
  int mrow, float angle, float length, int icolour)
  #else
  void overslit (idslit, xslit, yslit, wslit, mrow, angle, length, icolour)
  int *idslit,mrow, icolour;
  float *xslit,*yslit,*wslit,angle,length;
  #endif
  {
  
  
  }
*/
