/* median.c 

   von Thomas Szeifert & Monika Darr  -  LSW  -  Heidelberg  -  July 5, 1997

   Compute the median of frames stored in an image catalog.
   We use a modified median routine of the numerical recipes. 
   The frames are read row by row to avoid codes with large size buffers.
   The number of input frames is only limited by the size of the RAM and
   the swap disk, but although the code is very fast one should take into
   accont that it takes a long time to read the data from the hard disk.
   Therefore more then a few 100 MB of input files won't make fun

   On old HP workstation with 50Mhz we can process 40 files of 1MB in less
   then 30 CPU seconds using HP ANSI c-compiler with options -OAe, type:
   make -k C_OPT=-OAe         on directory  ./src
*/

/*
  CHANGES: By some reason we can't open more then 50 files so we open
           the files for any row and close them again now

	   try to use fits format files - but it doesn't work yet
	   TS-16.5.98
 */



#include <midas_def.h>
#include <proto_fors.h>

#define STRGLEN 80
 


main()

{
  char ina[STRGLEN],inb[STRGLEN],cat[STRGLEN],fname[STRGLEN],cbuff[STRGLEN];
  char text[26];

  int imnoa,imnob,*imnoc,ibuff,iparam[8],itmp=0;
  int stat,nval,l,anzahl,idnul,unit,iav;
  int npix[2],naxis,size;
  int i,j,k,debug;
  unsigned *ibuff1;

  register float *rbuff1,*rbuff2,*medbuff,tmp=0;
  
  double start[2],step[2];




  SCSPRO ("median");




  stat = SCKRDI ("INPUTI", 1, 8, &iav, iparam, &unit, &idnul);
  debug = iparam[0];

  if (debug > 1)
    SCTPUT ("... read keywords");

  stat = SCKGETC ("IN_C",1,STRGLEN,&nval,cat); /* Midas Katalog der Input ..*/
  stat = SCKGETC ("IN_A",1,STRGLEN,&nval,ina);  /* Output File fuer MEDIAN */
  stat = SCKGETC ("IN_B",1,STRGLEN,&nval,inb);  /* Output File fuer SIGMA */
  if (debug > 0)
    printf ("catalog %s \n",cat);

  stat = SCCSHO (cat,&anzahl,&l);            /* Eintraege im Katalog */

  if (debug > 0)
  printf ("number of files %4d - last entry %4d \n",anzahl,l);

  stat = SCCFND (cat,1,fname);             /* 1. Bild im Katalog */

  if (debug > 1)
    printf ("first file %s \n",fname);




  /* 
     open buffers rbuff1 for the row we'll read and for transposed input array
     rbuff2[1,...,anzahl] which contains pixel values of the N files and 
     the next pixel values in rbuff2[(anzahl+1)+1,...,(anzahl+1)+anzahl]
     We keep one pixel empty because the numerical recipies start with the
     first entry of the array rbuff2[1]!!!
  */

  if (debug > 1)
    SCTPUT ("... open arrays");

  if (file_exists(fname,".bdf"))
    {
      stat = SCFOPN (fname,D_R4_FORMAT,0,F_IMA_TYPE,&ibuff); 
      stat = SCDRDI (ibuff,"NPIX",1,2,&nval,npix,&unit,&idnul);
    }
  else if (file_exists(fname,".fits"))
    {
      printf (" file %s, fits format not supported yet \n",fname);
      return;

      stat = SCFOPN (fname,D_UI2_FORMAT,0,F_IMA_TYPE,&ibuff); 
      stat = SCDRDI (ibuff,"NPIX",1,2,&nval,npix,&unit,&idnul);
    }


  naxis=2;

  imnoc = (int *) osmmget(anzahl * sizeof(int));
  ibuff1 = (unsigned *) osmmget((npix[0]+1) * sizeof(unsigned));
  rbuff1 = (float *) osmmget((npix[0]+1) * sizeof(float));
  rbuff2 = (float *) osmmget((anzahl+1) * (npix[0]+1) * sizeof(float));
  medbuff = (float *) osmmget ((npix[0]+1) * sizeof(float));




  if (debug > 1)
    SCTPUT ("... create output file");
  
  /* create output file */

  size = npix[0]*npix[1]+1;
  SCFCRE (ina, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, size ,&imnoa);

  if (debug > 1)
    printf ("... copy descriptors of %s to the output file %s \n",fname,ina);

  SCDCOP (ibuff, imnoa, 1, "");
  SCFCLO (ibuff);




  /* copy input data to a buffer sorted by the files as first and by the rows 
     as second criterium for an elegant use of the median routine -- 
     then compute the median -- and write the result into the output file
     on the disk
  */

  if (debug > 0)
    printf ("... do the job with %4.0f MB \n",(size*4*anzahl)/1.0e+6);

  for (k = 0 ; k < npix[1] ; k++)
    {

      for (i = 0 ; i < anzahl ; i++)
	{
	  stat = SCCFND (cat,i+1,fname);  
	  if (debug > 2)
	    printf ("... open file %4d\n",imnoc[i]);

	  if (file_exists(fname,".bdf"))
	    {
	      stat = SCFOPN (fname,D_R4_FORMAT,0,F_IMA_TYPE,&ibuff); 
	      stat = SCFGET (imnoc[i],npix[0]*k+1,npix[0],&iav,(char *)rbuff1);
	      for ( j = 0 ; j < npix[0] ; j++)
		{
		  rbuff2[(anzahl+1)*j + (i+1)] = rbuff1[j]; 
		  if (debug > 2)
		    printf("%4d %4d %4d %8.1f\n",i,j,(anzahl+1)*j + (i+1),
			   rbuff1[j]);
		} 
	    }
	  else if (file_exists(fname,".fits"))
	    {
	      printf (" file %s, fits format not supported yet \n",fname);
	      return;

	      stat = SCFOPN (fname,D_UI2_FORMAT,0,F_IMA_TYPE,&ibuff); 
	      stat = SCFGET (imnoc[i],npix[0]*k+1,npix[0],&iav,(char *)ibuff1);
	      for ( j = 0 ; j < npix[0] ; j++)
		{
		  rbuff2[(anzahl+1)*j + (i+1)] = (float) ibuff1[j]; 
		  if (debug > 2)
		    printf("%4d %4d %4d %8.1f\n",i,j,(anzahl+1)*j + (i+1),
			   rbuff1[j]);
		} 

	    }

	  /*
	    stat = SCFOPN (fname,D_R4_FORMAT,0,F_IMA_TYPE,&imnoc[i]); 	  
	    stat = SCFGET (imnoc[i],npix[0]*k+1,npix[0],&iav, (char *)rbuff1);
	    */

	  /* j starts with 1 - see numerical recipes */

	  /*
	  for ( j = 0 ; j < npix[0] ; j++)
	    {
	      rbuff2[(anzahl+1)*j + (i+1)] = rbuff1[j]; 
	      if (debug > 2)
		printf("%4d %4d %4d %8.1f\n",i,j,(anzahl+1)*j + (i+1),
		       rbuff1[j]);
	    } 
	    */

	  stat = SCFCLO (imnoc[i]);	  
	} 
  

      /*
	Use this code as a template:

	it's very easy to write a new application that computes a mathematical
	function of the files in the catalog. Examples are mean values,
	kappa/sigma clipping ...

	First write the function and the definitions of the function in the
	header of this file. 
	Then change only the following line if you want to use your 
	mathematical function but not the median given below.

	take the following routine as an example 
	(== piksrt median -> see below):

	for (j = 0 ; j < npix[0] ; j++)
	  medbuff[j] = median (&rbuff2[(anzahl+1)*j],anzahl);
      */ 

      for (j = 0 ; j < npix[0] ; j++)
	medbuff[j] = median2 (&rbuff2[(anzahl+1)*j],anzahl);


      SCFPUT(imnoa,npix[0]*k+1,npix[0], (char *) medbuff);


      tmp = tmp + 100.0 / ((float) npix[1]);
      if ( tmp > 1.0){
        itmp+=1;
        sprintf (text,"... progress %3d percent",itmp+1);
        SCTDIS (text,26);
        tmp -= 1.0;
      }
    }


  
  if (debug > 1)
    SCTPUT ("... close  files");

  /* close all input files and stop the work*/

  osmmfree((char *) imnoc);
  osmmfree((char *) rbuff1);
  osmmfree((char *) rbuff2);
  osmmfree((char *) medbuff);


  SCFCLO (imnoa);

  if (debug > 1)
    SCTPUT ("... stop the work");

  SCSEPI();
  return 0;
}






