#include <ctype.h>
#include <string.h>
#include <stdio.h>

#include <tbldef.h>
#include <midas_def.h>
#include <proto_fors.h>





#ifndef PI
#define PI       acos(-1.0)
#endif
#define RAD1     (180.0 / PI)

#ifdef __STDC__
int     
interpol(double x1[], double y1[], int n1, double *x2, double *y2, int n2)
#else
int     
interpol(x1, y1, n1, x2, y2, n2)
int n1,n2;
double x1[],y1[],*x2,*y2;
#endif
{
  int i,istart=0;
  double *buff;

  buff = dvector(1,n1);

  dspline(x1, y1, n1, 1.0e30, 1.0e30, buff);

  for (i = 0 ; i < n2 ; i++)
    dsplint(x1, y1, buff, n1, x2[i], &y2[i]);

  free_dvector(buff,1,n1);
  return 0;
}

void 
fitsel_poly (
#ifdef __STDC__
	double *, double *, int *, double *, double *, 
	double *, int , int , double *, int ,  double, int, double *
#endif
);

void 
mos_savedisp (
#ifdef __STDC__
	      char name[], double y[], float s[], int nrow, double coef[],
	      double start[],double step[], int fit_deg, int degx, int degxy, 
	      double dnull
#endif
);

double 
closest_to(
#ifdef __STDC__
	   double reference[], double what, double howclose, double dnull, 
	   int *jjj, int kkk
#endif
);



/*
  calculate the dispersion of the glass like in the Schott catalog
  (W.Fuertig, W.Seifert)
 */
int bindex
#ifdef __STDC__
(double lam, double coef[], double *gind)
#else
(lam, coef, gind)
double lam, coef[], *gind;
#endif
{
  double lam2;
  int i;

  lam2 = SQR(lam/10000.0);

  *gind = lam2 * coef[0] / ( lam2 - coef[1]);
  *gind += lam2 * coef[2] / ( lam2 - coef[3]);
  *gind += lam2 * coef[4] / ( lam2 - coef[5]);
  *gind = sqrt(*gind + 1.0);

  return 0;
}


/*
  calculate the position of a given wavelength on the chip meassured from
  the center (W.Fuertig, W.Seifert)
 */
/*int beam(double tilt, double phi, double *alf1, double gind,
	 double lambda,  double gr,  double ord, double fkam, 
	 double *yccd)*/
int beam
#ifdef __STDC__
(double tilt, double phi, double *alf1, double gind,
         double lambda,  double gr,  double ord, double fkam, 
         double *yccd)
#else
(tilt, phi, alf1, gind, lambda, gr, ord, fkam,  yccd)
double tilt, phi, *alf1, gind, lambda, gr, ord, fkam, *yccd;
#endif
{

  double alf2,bet2,alf3,bet3,betg,alfg, dnull = 0.0;

  lambda /= 1.0e+7;

  alf2 = *alf1 + tilt;
  bet2 = asin( sin(alf2)/gind );
  alf3 = phi - bet2;
  betg = ord * gr * lambda  - sin(alf3) * gind;
  if (fabs(betg) > 1.0)
    *yccd = -9999.0;
  else{
    betg = asin(betg);
    alfg = betg + phi - tilt;
    *yccd = tan(alfg) * fkam;
  }


  return 0;
}



/*
  find the clesest line in the WLC catalog (O.Stahl)
 */
double closest_to
#ifdef __STDC__
(double reference[], double what, double howclose, double dnull, int *jjj, 
 int kkk)
#else
     (reference, what, howclose, dnull,jjj,kkk)
     double reference[], what, howclose, dnull;
     int *jjj,kkk;
#endif
{
  int i=0, j=0;
  double  x1, x2;
  
  x1 = what - howclose;
  x2 = what + howclose;
  
  i = 0;
  
  while (reference[i] < x1 || i == kkk)
    i++;

  if (i > kkk) return dnull;


  if (i != 0)
    j = i-1;
  else 
    return dnull;

  while ( (reference[j] < x2 || j == kkk) && j <= kkk)
    j++;

  if (i > kkk) return dnull;

  if ((j-i) == 1) /* identifications should be unique */
    {
      *jjj = i;
      return reference[i];
    }
  else
    {
      return dnull;
    }   
}



int main ()

{
  /*general*/
  int i,j,k,l,m=0,inull,actvals,kun,knul,debug,ibuff,select,selnrow;
  float fnull;
  double dnull;
  char line[80];
  int stat;
  /* dummies*/


  /* io-related*/
  int tid1, tid2, tid3, tid4, icol1[9],icol2[3],icol3[9],icol4[2];
  int lcol,lrow,lsort;
  int ncol,nrow,nsort,acol,arow;
  int mcol,mrow,msort,kcol,krow,ksort;
  int iparam[8],*rej2,rej1;
  int idno;
  int nslit=0,nyallrow=0,idcoef;
  int degy,degxy;
  int *sels1,*lower,*upper,*nyrow,*nline;
  int niter;

  char in_table1[60],in_table2[60],in_table3[60],in_table4[60],out_table[60],
    mode[60],text[80];
   
  double *x1,*y1,*i1,*w1,*wc1,*r1,*w2,*f2,*x2,*y2,*wc2;
  float *s4,*o4,*y4;
  double gr,phi,ncoef[6],gind,tilt,alf1=0.0,fkam,dparam[8],pxlsize;
  double deltax = 2048.0,dbuff;

  float *o1,*s2;
  float fparam[8],xbuff,ybuff;

  int fitdeg,iorder[2],isel,iall,point=0,poltyp;  
  int nmatch, prevnmatch=-1;

  double *xtmp, *itmp;
  double *dcoef;

  double start[2],step[2],center;
  int npix[2];
  
  int verif=-1,initsel=0;
  double alpha,stdres=0,tol;

  double newoffset;
  int nident;

  /* for focal length of the camera:*/
  int tid5,fcol,frow,fsort,icol5[2];
  double *flambda,*fcam,fcaminter;
  char in_table5[61];

  SCSPRO ("califors");
  TCMNUL(&inull,&fnull,&dnull);	/*  */
 


  SCKGETC ("IN_A",   1, 60, &actvals, in_table1);
  SCKGETC ("IN_B",   1, 60, &actvals, in_table2);
  SCKGETC ("IN_C",   1, 60, &actvals, in_table3);
  SCKGETC ("IN_D",   1, 60, &actvals, in_table4);
  SCKGETC ("IN_E",   1, 60, &actvals, in_table5);
  SCKGETC ("OUT_A",   1, 60, &actvals, out_table);

  SCKRDI ("INPUTI",   1, 8, &actvals, iparam,&kun,&knul);
  SCKRDR ("INPUTR",   1, 8, &actvals, fparam,&kun,&knul);
  SCKRDD ("INPUTD",   1, 8, &actvals, dparam,&kun,&knul);
  SCKRDI ("DEBUG",   1, 1, &actvals, &debug,&kun,&knul);


  SCKRDD ("START",   1, 2, &actvals, &start[0],&kun,&knul);
  SCKRDD ("STEP",   1, 2, &actvals, &step[0],&kun,&knul);
  SCKRDI ("NPIX",   1, 2, &actvals, &npix[0],&kun,&knul);

  SCKRDD ("CENTER",   1, 1, &actvals, &center,&kun,&knul);
  if (NINT(center) <= -9990)
    {
      center = start[0] + step[0] * (((double) npix[0])/2.0 - 1.0);
    }


  tilt = dparam[0] / RAD1, fkam = dparam[1], pxlsize = dparam[2];

  fitdeg = iparam[0];
  degy = iparam[1];
  degxy = iparam[2];
  poltyp = iparam[3];

  alpha = (double) fparam[0];
  tol = (double) fparam[1];

  SCTPUT ("create raw calibration table table ");
  SCTPUT ("------------");
  sprintf (line, "alpha = %f, tol = %f, fitdeg = %2d ", 
	   alpha,tol,fitdeg);
  SCTPUT (line);

  if (poltyp == 0)
    sprintf (line, "normal polynom");
  else if  (poltyp == 1)
    sprintf (line, "Chebichev polynom");
  else 
    sprintf (line, "Legendre polynom");
  SCTPUT (line);
  SCTPUT ("------------");




  /*

    READ THE INPUT TABLE with the locations of the mos slitlets

    */


  if (debug > 1)
    printf ("open MOS table with slitlet and offsets\n");

  TCTOPN (in_table4, F_I_MODE, &tid4);

  TCIGET(tid4,&lcol, &lrow, &lsort, &acol, &arow);

  TCCSER (tid4, ":SLIT", &icol4[0]);
  TCCSER (tid4, ":XOFFSET", &icol4[1]);
  TCCSER (tid4, ":YSTART", &icol4[2]);
  TCCSER (tid4, ":YEND", &icol4[3]);

  s4 = fvector(0,lrow);
  o4 = fvector(0,lrow);
  y4 = fvector(0,2*lrow);

  for (i = 0; i < lrow; i++){
    TCERDR(tid4,i+1,icol4[0],&s4[i],&knul);
    TCERDR(tid4,i+1,icol4[1],&o4[i],&knul);
    TCERDR(tid4,i+1,icol4[2],&y4[i*2],&knul);
    TCERDR(tid4,i+1,icol4[3],&y4[i*2+1],&knul);
  }

  TCTCLO(tid4);






  /*

    READ TABLE with phycal parameters of the grism

    */

  if (debug > 1)
    printf ("READ TABLE with phycal parameters of the grism\n");


  TCTOPN (in_table3, F_I_MODE, &tid3);

  TCIGET(tid3,&kcol, &krow, &ksort, &acol, &arow);

  TCCSER (tid3, "IDNO", &icol3[0]);
  TCCSER (tid3, "GR", &icol3[1]);
  TCCSER (tid3, "PHI", &icol3[2]);
  TCCSER (tid3, "N0", &icol3[3]);
  TCCSER (tid3, "N1", &icol3[4]);
  TCCSER (tid3, "N2", &icol3[5]);
  TCCSER (tid3, "N3", &icol3[6]);
  TCCSER (tid3, "N4", &icol3[7]);
  TCCSER (tid3, "N5", &icol3[8]);

  if (debug > 3)
    {
      for (j = 0 ; j < 9 ; j++)
	printf("%d ",icol3[j]);
	printf("\n");
    }

  for (j = 0; j < krow; j++)
    {
      TCSGET(tid3,j+1,&select);

      if (select)
	{

	  TCERDI(tid3,j+1,icol3[0],&idno,&knul);
	  TCERDD(tid3,j+1,icol3[1],&gr,&knul);
	  TCERDD(tid3,j+1,icol3[2],&phi,&knul);
	  phi /= RAD1;
	  
	  TCERDD(tid3,j+1,icol3[3],&ncoef[0],&knul);
	  TCERDD(tid3,j+1,icol3[4],&ncoef[1],&knul);
	  TCERDD(tid3,j+1,icol3[5],&ncoef[2],&knul);
	  TCERDD(tid3,j+1,icol3[6],&ncoef[3],&knul);
	  TCERDD(tid3,j+1,icol3[7],&ncoef[4],&knul);
	  TCERDD(tid3,j+1,icol3[8],&ncoef[5],&knul);

	  if (debug > 0)
	    printf("... grism %d  prism angle = %f deg, %f grooves/mm \n",
		   idno,phi*RAD1,gr);

	}
    }

  TCTCLO(tid3);



  /*

    READ TABLE with the focal length of the camera

    */


  TCTOPN (in_table5, F_I_MODE, &tid5);

  TCIGET(tid5,&fcol, &frow, &fsort, &acol, &arow);

  TCCSER (tid5, "LAMBDA", &icol5[0]);
  TCCSER (tid5, "FCAM", &icol5[1]);

  flambda = dvector(0,frow);
  fcam = dvector(0,frow);

  for (j = 0; j < frow; j++)
    {
      TCSGET(tid5,j+1,&select);

      if (select)
	{
	  TCERDD(tid5,j+1,icol5[0],&flambda[j],&knul);
	  TCERDD(tid5,j+1,icol5[1],&fcam[j],&knul);
	}
    }



  /*

    nrow 

    */


  nrow = (int) ((10000.0 - 3500.0) / 100.0) + 1;  


  w2 = dvector(0,nrow*2*lrow);           /* laboratory wavelength */
  x2 = dvector(0,nrow*2*lrow);           /* x position */
  y2 = dvector(0,nrow*2*lrow);           /* y position */
  s2 = vector(0,nrow*2*lrow);           /* slit */
  
  selnrow = 0;
  for (i = 0; i < lrow; i++)
    {

      for (j = 0; j < nrow; j++)
	{ 
	  selnrow = 2*i*nrow+j;
	  w2[selnrow] = (double)(3200.0 + 100.0 * j);
	  w2[selnrow+nrow] = (double)(3200.0 + 100.0 * j);
	  
	  bindex(w2[selnrow],ncoef,&gind);
	  if (debug > 2 )
	    printf(" lamda = %f  n(lambda) = %f \n", w2[selnrow], gind);
	  
	  /*interpolate the color dependent focal length of the camera*/
	  interpol(flambda,fcam,frow,&w2[selnrow],&fkam,1);

	  beam(tilt,phi,&alf1,gind,w2[selnrow],gr,1,fkam,&x2[selnrow]);
	  x2[selnrow+nrow] = x2[selnrow];
	  if (debug > 2)
	    printf("%f %f %f w=%f x=%f %f %f\n",tilt*RAD1,phi*RAD1,gind,w2[selnrow],x2[selnrow],gr,fkam);
	  
	  
	  if ( fabs(x2[selnrow]/pxlsize) < (1.25 * center)){
	    if (debug > 1)
	      printf("... optical guess:  lambda = %7.1f     x2 = %7.3f mm == %7.1f pxl\n",w2[selnrow], x2[selnrow], (x2[selnrow]/pxlsize) + center);
	  }
      
	  x2[selnrow] = (x2[selnrow] / pxlsize) + center + o4[i];
	  x2[selnrow+nrow] = (x2[selnrow+nrow] / pxlsize) + center + o4[i];

	  y2[selnrow] = y4[i*2];
	  y2[selnrow+nrow] = y4[i*2+1];

	  s2[selnrow] = s4[i];
	  s2[selnrow+nrow] = s4[i];

	}

    }


  idcoef = (fitdeg+1);
  dcoef = dvector (0, 2*lrow*idcoef);
  wc2 = dvector(0, 2*lrow*nrow); 
  itmp = dvector(0, 2*lrow*nrow);
  xtmp = dvector(0, 2*lrow*nrow); 


  if (debug > 1)
    printf ("start fitsel_poly with %d entries \n",selnrow);


  /* for all mos slits */

  for (i = 0 ; i < 2*lrow ; i++){
    
    for (j = 0 ; j < 2*lrow*nrow; j++)
	itmp[j] = xtmp[j] = (double) 0.0;

    fitsel_poly (itmp, xtmp, &isel, 
		 &w2[i*nrow],&wc2[i*nrow],&x2[i*nrow], nrow, 
		 fitdeg, &dcoef[(fitdeg+1)*i], debug, dnull, 
		 poltyp, &stdres);


    /*
    if ( (int) dcoef[(fitdeg+1)*i+1] > 9500 )
      for (j= i*nrow; j < nrow*(i+1) ; j++)
	printf ("x2 %f y2 %f s2 %f w2 %f wc2 %f \n",x2[j],y2[j],s2[j],w2[j],wc2[j]);
    */


    if (debug > 1)
    printf ("fitsel_poly slit %f y %f stdres = %7.2f \n",s2[i*nrow],y2[i*nrow],stdres);

    if (debug > 1){
      for (j = (fitdeg+1)*i+1; j <= (fitdeg+1)*(i+1); j++)
	printf ("%f ",dcoef[j]);
      printf ("\n");
    }    
  }


  mos_savedisp (out_table, y2, s2, 2*lrow*nrow, dcoef, start, step, 
		fitdeg, degy, degxy,dnull);




  free_fvector(s4,0,lrow);
  free_fvector(o4,0,lrow);
  free_fvector(y4,0,2*lrow);

  free_dvector (dcoef,0, 2*lrow*idcoef);
  free_dvector(itmp,0,selnrow);
  free_dvector(xtmp,0,selnrow); 

  free_dvector(wc2,0,selnrow); 
  free_dvector(w2,0,nrow*2*lrow);           /* laboratory wavelength */
  free_dvector(x2,0,nrow*2*lrow);           /* x position */
  free_dvector(y2,0,nrow*2*lrow);           /* y position */
  free_vector(s2,0,nrow*2*lrow);           /* slit */


  SCSEPI ();
  exit(0);
  return;
}
/*end of main*/





/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void
fitsel_poly (double itmp[], double xtmp[], int *isel, double w1[], 
	     double wc1[], double x1[], int iall, int fitdeg, 
	       double *dcoef, int debug, double dnull,int poltyp, 
	       double *stdres)
#else
void
fitsel_poly (itmp, xtmp, isel, w1, wc1, x1, iall, fitdeg, dcoef, debug,dnull,
	     poltyp, stdres)
     int *isel,iall,fitdeg,debug,poltyp;
     double itmp[],xtmp[],w1[],wc1[],x1[];
     double *dcoef,dnull,*stdres;
#endif

{

  int i,l;
  
  int sel = 0;
  *stdres = 0.0;
    
  if (debug > 2)
    printf ("fitsel_poly has started\n");

  /* 
     READ_THIS:  

     the arrays of tid1 are given to this subroutine by pointers. Therefore
     point is 0 and we count from   i (= point) = 0   to   i = iall, but not
     from the first table row of the order as in main() 

     */    


  for (i = 0; i < iall ; i++) 
    {

      if (w1[i] > 1.0e-5){
	xtmp[sel] = x1[i];
	itmp[sel] = w1[i];
	sel++;
      }

    }


  /* 
     FIT the dispersion and compute the wavelength for any
     lines the selected and the others of the given order
     */


  if (sel >= fitdeg + 1){
    if (debug > 1)
      printf("   ... fitsel_poly\n");
    dfit_poly (itmp,xtmp,sel,wc1,x1,iall,
	       fitdeg, dcoef,poltyp);
  }
  else{
    dfit_poly (itmp,xtmp,sel,wc1,x1,iall,
	       (sel-1), dcoef,poltyp);
    for (l = sel-2 ; l < fitdeg ; l++)
      dcoef[l] = dnull;
  }

  *stdres=0.0;

  for  (i = 0; i < iall ; i++)
    *stdres += pow( wc1[i]-w1[i] ,2.0);

  *stdres = pow(*stdres/iall,0.5);

  *isel = sel;

  return;
}





#ifdef __STDC__
void 
mos_savedisp (char name[], double y[], float s[], int nrow, double coef[],
	      double start[],double step[], int fit_deg, int degy, int degxy,
	      double dnull)
#else
void 
mos_savedisp (name, y, s, nrow, coef, start, step, fit_deg, degy, degxy,dnull)
char name[];
double y[], coef[], start[], step[], dnull;
float s[];
int nrow, fit_deg, degxy;
#endif
{
  int i,j,iypxl,istmp;
  int tide,nbline,ncoef,kun;
  int col[3],*ocol,nyrow=0,nslit=0;
  double *ocoef;
  double poly;
  char descname[81],colname[81];


  ocol = ivector(0,fit_deg);
  ocoef = dvector(0,fit_deg);

  if (TCTINI (name, F_TRANS, F_IO_MODE, 5L, 10L, &tide))
    SCTPUT ("**** Error while creating output table");
  nbline = 0;


  TCCSER (tide, ":SLIT", &col[0]);
  if (col[0] == (-1))
    TCCINI (tide, D_I4_FORMAT, 1, "I6", "Slit Number",
            "SLIT", &col[0]);
  TCCSER (tide, ":ROW", &col[1]);
  if (col[1] == (-1))
    TCCINI (tide, D_I4_FORMAT, 1, "I6", "Row Number",
            "ROW", &col[1]);
  TCCSER (tide, ":Y", &col[2]);
  if (col[2] == (-1))
    TCCINI (tide, D_R8_FORMAT, 1, "F8.2", "Y Value",
            "Y", &col[2]);


  for (i = 0; i <= fit_deg; i++)
    {
      sprintf (colname, ":COEF_%d\0", i+1);
      TCCSER (tide, colname, &ocol[i]);
      if (ocol[i] == (-1))
        TCCINI (tide, D_R8_FORMAT, 1, "F16.10", "Coefficients",
                colname, &ocol[i]);
    }

	 

  ncoef = fit_deg + 1;

  for (j = 0; j < nrow; j++)
    if(y[j+1] != y[j])
      {
	nyrow++;

	istmp = (NINT(s[j]));
	TCEWRI(tide, nyrow, col[0], &istmp);
	iypxl = NINT( (y[j] - start[1])/step[1] + 1);
	TCEWRI(tide, nyrow, col[1], &iypxl);
	TCEWRD(tide, nyrow, col[2], &y[j]);


	for (i=1; i<=ncoef; i++)
	  TCEWRD(tide, nyrow, ocol[i-1], &coef[(nyrow-1)*(fit_deg+1)+i]);

	if(j == 0 || s[j+1] != s[j])
	  nslit++;
	if(nslit == 1 || s[j+1] != s[j]){
	  sprintf(descname,"coefslit%d",istmp);
	  SCDWRD(tide, descname, &coef[1+(nslit-1)*ncoef], 1, ncoef, &kun);
	}

      }


  i = fit_deg + 1;

  SCDWRI(tide, "DEGX", &i, 1, 1, &kun);
  SCDWRI(tide, "DEGXY", &degxy, 1, 1, &kun);
  SCDWRI(tide, "DEGY", &degy, 1, 1, &kun);

  
  SCDWRD (tide, "LNPIX", &coef[2], 1L, 1L, &kun);
  SCDWRI (tide, "LNDEG", &fit_deg, 1L, 1L, &kun);
  SCDWRI (tide, "LNCOE", &i, 1L, 1L, &kun);

  TCTCLO(tide);
  free_ivector(ocol,0,fit_deg);
  free_dvector(ocoef,0,fit_deg);

  return;
}



















