! @(#)spolcalib.prg	1.7 (ESO-IPG) 2/17/94 17:28:46
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++
!.COPYRIGHT   (C) 1993 European Southern Observatory
!.IDENT       spolcalib.prg
!.AUTHORS     Pascal Ballester (ESO/Garching) 
!             Cristian Levin   (ESO/La Silla)
!             Sabine Moehler   (LSW)
!.KEYWORDS    Spectropolarimetry
!.PURPOSE     fit dispersion relation
!.VERSION     1.0  Package Creation  17-MAR-1993  
!             1.1  adapted for SPOL   20-AUG-1993   
!
! CALIB/SPOL tolerance degree method central_wavelength, av._dispersion
!               (A/pix) plot_option  display_option try spol_table
! The program consists of two parts:
! 1) The sorting part, where it reads from spol_table the tables with line and
!    with slitlet positions
! 2) The calibration part where the actual calibration is performed
!-------------------------------------------------------------------------------
!
CROSSREF tol deg mtd cen,disp plo disp spol_tab try
!-------------------------------------------------------------------------------
! Parameters ...
!---------------------------------------------------
! ... for calibration part
!---------------------------------------------------
DEFINE/PARAM  P1  {TOL}                  NUMB   "Tolerance "
DEFINE/PARAM  P2  {WLCORD(1)}            NUMB   "Degree     "
DEFINE/PARAM  P3  {WLCMET}               CHAR   "Mode (I[dent]/L[inear]/F[ors])(C[onst.]/V[ar.]/F[it]) (Default: FC) :"
DEFINE/PARAM  P4  {WCENTER},{AVDISP},0.  NUMB   "Center wav., dispersion:"
DEFINE/PARAM  P5  {PLOTC}                CHAR   "Plot of residuals (Y/N):"
DEFINE/PARAM  P6  50                     NUMB   "Display (100 = detailed, 50 = partial. Default = 50)"
!---------------------------------------------------
! ... for sorting part
!-------------------------------------------------------------------------------
DEFINE/PARAM P7 ? CHAR "First try/All frames (Def. F):"
DEFINE/PARAM P8 {SPOL_TAB}      TABLE   "Enter input table:"

DEFINE/MAXPAR 8

!-------------------------------------------------------------------------------
! Local keys ...
!---------------------------------------------------
! ... for calibration part
!---------------------------------------------------
DEFINE/LOCAL KEYSEL/C/1/64 - 
DEFINE/LOCAL COUNT/I/1/1 0.
DEFINE/LOCAL DISP/I/1/1 0.
!---------------------------------------------------
! ... for sorting part
!-------------------------------------------------------------------------------
DEFINE/LOCAL NUMROW/I/1/1 0.
DEFINE/LOCAL I/I/1/1 0.
DEFINE/LOCAL II/I/1/1 0.
DEFINE/LOCAL J/I/1/1 0.
DEFINE/LOCAL GETFF/C/1/1 X
DEFINE/LOCAL TRY/C/1/1 F
WRITE/KEY TMP/C/1/80 " " ALL
WRITE/KEY SLITTABL/C/1/80 " " ALL

!-------------------------------------------------------------------------------
! Check if TRY is explicitly given (otherwise assume First try)
!-------------------------------------------------------------------------------
IF P7(1:1) .ne. " " THEN
  WRITE/KEY TRY/C/1/1 {P7}
ENDIF 

!-------------------------------------------------------------------------------
! Get number of rows in spol_tab
!-------------------------------------------------------------------------------
COPY/DK {P8}.tbl TBLCONTR/I/4/4 NUMROW
!-------------------------------------------------------------------------------
! Get low and high row numbers for science frames in spol_tab
!-------------------------------------------------------------------------------
!@@ spolcount {P8}
COUNT/SPOL {P8}
!-------------------------------------------------------------------------------
! Check whether same flat or individual flats have been used for slitlet search
!-------------------------------------------------------------------------------
COPY/DK {P8}.tbl TBLGETFF/C/1/1 GETFF 
!-------------------------------------------------------------------------------
! Initialize column to hold names of linfit tables
!-------------------------------------------------------------------------------
WRITE/KEYW IN_A   {P8}
WRITE/KEY INPUTC/C/1/8 {LINFIT}  
RUN STD_EXE:spolcolini
!*******************************************************************************
!*****************            SORTING PART                **********************
!*******************************************************************************
!-------------------------------------------------------------------------------
! Get tables with slitlets and line positions
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
! Same flat for all frames == one slitlet table for all frames
!---------------------------------------------------
IF GETFF(1:1) .eq. "S" .or. GETFF(1:1) .eq. "s" THEN
!---------------------------------------------------
! Get slitlet table for all science frames
!---------------------------------------------------
  DO II = 1 {NUMROW}
    COPY/TK {P8} {SLIT_TAB} @{II} TMP/C/1/20
    IF TMP(1:1) .ne. " " THEN
      WRITE/KEY SLITTABL {TMP}
    ENDIF
  ENDDO
!---------------------------------------------------
! Get LINPOS table(s) for science frames
!---------------------------------------------------
  DO II = 1 {NUMROW}
    COPY/TK {P8} {LINPOS} @{II} IN_A/C/1/80
    IF IN_A(1:1) .ne. " " THEN
      COPY/DK {IN_A} WLC/C/1/80 WLC
!---------------------------------------------------
! Perform calibration for actual table 
!---------------------------------------------------
      GOTO CALIB 
!---------------------------------------------------
! Continue sorting and write name of output table to 
! column {LINFIT} of spol_tab
!---------------------------------------------------
CONTS:
      WRITE/TABLE {P8} {LINFIT} @{II} {OUT_A}.tbl 
!---------------------------------------------------
! Try only one LINPOS table for first try
!---------------------------------------------------
      IF TRY(1:1) .eq. "F" .or. TRY(1:1) .eq. "f"  THEN
        GOTO endsort   
      ENDIF
    ENDIF
!---------------------------------------------------
! End of DO loop over LINPOS table(s) for one 
! science frame
!---------------------------------------------------
  ENDDO  
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
! Individual flat for each frame == one slitlet table for each SC frame
!---------------------------------------------------
ELSEIF GETFF(1:1) .eq. "I" .or. GETFF(1:1) .eq. "i" THEN
!---------------------------------------------------
! Get slitlet table for each science frame
!---------------------------------------------------
  DO J = 1 {N_SC}  
    DO II = SC_STRT({J}) SC_END({J})
      COPY/TK {P8} {SLIT_TAB} @{II} TMP/C/1/20
      IF TMP(1:1) .ne. " " THEN
        WRITE/KEY SLITTABL {TMP}
      ENDIF
    ENDDO
!---------------------------------------------------
! Get LINPOS table(s) for science frame
!---------------------------------------------------
    DO II = SC_STRT({J}) SC_END({J})
      COPY/TK {P8} {LINPOS} @{II} IN_A/C/1/80
      IF IN_A(1:1) .ne. " " THEN
        COPY/DK {IN_A} WLC/C/1/80 WLC
!---------------------------------------------------
! Perform calibration for actual table
!---------------------------------------------------
        GOTO CALIB ! perform calibration for actual table
!---------------------------------------------------
! Continue sorting and write name of output table to 
! column {LINFIT} of spol_tab
!---------------------------------------------------
CONTI:  
        WRITE/TABLE {P8} {LINFIT} @{II} {OUT_A}.tbl 
!---------------------------------------------------
! Try only one LINPOS table for first try
!---------------------------------------------------
        IF TRY(1:1) .eq. "F" .or. TRY(1:1) .eq. "f"  THEN
          GOTO endsort    ! Try only one WLC frame for first try
        ENDIF
      ENDIF
!---------------------------------------------------
! End of DO loop over LINPOS table(s) for one 
! science frame
!---------------------------------------------------
    ENDDO  
!---------------------------------------------------
! End of DO loop over SC frames 
!-------------------------------------------------------------------------------
  ENDDO  
ENDIF
GOTO ENDSORT
!*******************************************************************************
!*****************          CALIBRATION PART              **********************
!*******************************************************************************
CALIB:
!-------------------------------------------------------------------------------
! check if linpos and linecat are tables
!-------------------------------------------------------------------------------
!VERIFY/MOS {LINPOS} TAB
!VERIFY/MOS {LINECAT} TAB
!
!-------------------------------------------------------------------------------
! get parameters
!-------------------------------------------------------------------------------
!WRITE/KEYW TOL    {P1}
!WRITE/KEYW WLCMET {P3}
!WRITE/KEYW WLCORD {P2} 
!WRITE/KEYW PLOTC  {P5}
WRITE/KEYW DISP   {P6}
!
COMP/KEY OUTPUTI = M$INDEXB(IN_A,".")-1
WRITE/KEY OUT_A {IN_A(1:{OUTPUTI(1)})}{SUF_FIT}
WRITE/KEYW IN_B   {LINECAT}
WRITE/KEYW IN_C/C/1/80   {SLITTABL}
!WRITE/KEYW INPUTC {WLCMET}
WRITE/KEYW INPUTC {P3}
!
INPUTR(1) = ALPHA
INPUTR(2) = MAXDEV
!INPUTR(3) = TOL
WRITE/KEY INPUTR/R/3/3 {P1} 
!
INPUTI(10) = DISP
!
!-------------------------------------------------------------------------------
! get size of WLC frame used for LINPOS table
!-------------------------------------------------------------------------------
COPY/DK {wlc} START/D/1/2 START/D/1/2
COPY/DK {wlc} STEP/D/1/2 STEP/D/1/2
COPY/KEYW  START/D/1/2   INPUTD/D/1/2  
COPY/KEYW  STEP/D/1/2    INPUTD/D/3/4  
!-------------------------------------------------------------------------------
! set integer parameters
!-------------------------------------------------------------------------------
WRITE/KEY INPUTI/I/1/1 {P2}
!INPUTI(1) = WLCORD(1)
!INPUTI(2) = WLCORD(2)
INPUTI(2) = NPIX(1)+START(1) ! total number of pixels along dispersion axis
INPUTI(3) = WLCNITER(1)
INPUTI(4) = WLCNITER(2)
INPUTI(5) = YSTART
!-------------------------------------------------------------------------------
! Linear Fit
!---------------------------------------------------
IF INPUTC(1:1) .EQ. "L" THEN

!---------------------------------------------------
! Check size of WLC 
!---------------------------------------------------
  IF NPIX(1) .EQ. 0 THEN
    COPY/DK {WLC} NPIX NPIX/I/1/2 
  ENDIF
  IF NPIX(1) .EQ. 0 THEN
    WRITE/OUT "No wavelength calibration frame defined - frame size unknown!"
    GOTO END
  ENDIF
  
  DEFINE/LOCAL WDEF/D/1/3  {P4}
  INPUTD(5) = WDEF(1) - (NPIX(1)+START(1)-1)/2.*WDEF(2) ! Coeffs in world 
  INPUTD(6) = WDEF(2)                                   ! coordinates system
  INPUTD(7) = WDEF(3)
ENDIF
!
!-------------------------------------------------------------------------------
! FORS Fit 
!---------------------------------------------------
IF INPUTC(1:1) .EQ. "F" THEN
SET/FORMAT I2   

!---------------------------------------------------
! Check grism
!---------------------------------------------------
  IF GRISM .EQ. 0 THEN
    WRITE/OUT "No grism defined!"
    GOTO END
  ENDIF
!---------------------------------------------------
! Check size of WLC 
!---------------------------------------------------
  IF NPIX(1) .EQ. 0 THEN
    COPY/DK {WLC} NPIX NPIX/I/1/2 
  ENDIF
  IF NPIX(1) .EQ. 0 THEN
    WRITE/OUT "No wavelength calibration frame defined - frame size unknown!"
    GOTO END
  ENDIF

  INPUTD(5) = FORS{GRISM}(1) - (NPIX(1)+START(1)-1)/2.*FORS{GRISM}(2) ! Coeffs
  INPUTD(6) = FORS{GRISM}(2)                      ! in world coordinates system
  INPUTD(7) = 0.
ENDIF
!
RUN STD_EXE:moscalib
!
!-------------------------------------------------------------------------------
! Plot Option
!---------------------------------------------------
COPY/DK {IN_A} TSELTABL KEYSEL
!IF PLOTC(1:1) .eq. "Y" THEN
IF P5(1:1) .eq. "Y" THEN
 DO I = 1 {NSLIT} 
   IF CAL({I}:{I}) .gt. 0 THEN
   COMP/KEY OUTPUTI = M$INDEXB(IN_A,".")-1
     RESPLOT/SPOL {IN_A(1:{OUTPUTI(1)})} {I}  
   ELSEIF CAL({I}:{I}) .lt. 0 THEN
     WRITE/OUT "No dispersion relation for slitlet no. {I}"
   ENDIF
 ENDDO
ELSE 
 GOTO END
ENDIF
END:
!---------------------------------------------------
! Restore selection status of LINPOS table
!-------------------------------------------------------------------------------
IF KEYSEL(1:1) .eq. "-" THEN
  SELECT/TABLE {IN_A} all
ELSE
  COPY/KD KEYSEL/c/1/64 {IN_A} TSELTABL/c/1/64
ENDIF
!-------------------------------------------------------------------------------
! Return to sorting part
!-------------------------------------------------------------------------------
IF GETFF(1:1) .eq. "S" .or. GETFF(1:1) .eq. "s" THEN
  GOTO CONTS
ELSEIF GETFF(1:1) .eq. "I" .or. GETFF(1:1) .eq. "i" THEN
  GOTO CONTI
ENDIF
ENDSORT:
