/* @(#)mosfollow.c      6.1.1.4  (LSW)  07/02/94  13:48:49           */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* .COPYRIGHT   (C) 1994 Landessternwarte Heidelberg                 */
/* .IDENT        follow.c                                            */
/* .AUTHOR       Otmar Stahl - LSW Heidelberg                        */
/* .KEYWORDS                                                         */
/* .PURPOSE      Search object with median: (val - med(val)) > thres */
/*               Center object with Gaussian fitting                 */
/*               Follow the object and fit polynomial to y = y(x)    */
/*               (adapted from mosdefine.c)                          */
/* .VERSION      1.0  07-FEB-1994  Implementation                    */
/*-------------------------------------------------------------------*/

#include <math.h>
#include <tbldef.h>
#include <midas_def.h>
#include <ok.h>

/*
#include <proto_mutil.h>
#include <proto_nrutil.h>
#include <proto_fitnol.h>
#include <dlfit.h>
*/
#include <proto_fors.h>

#define MAXOBJECTS 1000
#define MAXSLITS 100
#define MAXDIFF 2.0
#define GAUSSIAN 1

int os_col, o1_col, o2_col, ni_col, x_col;
int Ntot, Npix[2], Method, NseqO, NseqS, upper[MAXSLITS], lower[MAXSLITS];
int Xbin, Xstep, Iwin, Width, Tid;
double  Start[2], Step[2], *Xgaus, *Ygaus, *A;
float thresh;

#ifdef __STDC__
void find_objects (float[], float[], int[], int, int *, int, float[]);
void fit_obj (float[], int, float[], int, int);
void follow (float[], float[], int, int, float[], int);
/*
  float own_median (int, float[], float *);
*/
void average(float[], float[], int, int, int);
#else
void find_objects ();
void fit_obj ();
void follow ();
/*
  float own_median ();
*/
void average();
#endif

main ()

{
  int i, nax, tidmos, imno, selected, kun, knul, iav;
  int col, nslits, acol, arow, nsort, param[4], null, slit[MAXSLITS];
  int sl_col, up_col, lo_col, dunit, inull;

  int nact, nslit, objlist[MAXOBJECTS];
  int center;
  float obj[5], scan_pos, rnull;
  double dnull;

  char method[21], obj_image[60], mos_table[60];
  char window_table[60], unit[64], inst[72], line[80], linem[80];

  float *buff, *image, rpar[2];

/* get into Midas and get parameters */

  SCSPRO ("follow");

  SCKGETC ("IN_A", 1, 60, &iav, obj_image);
  SCKGETC ("IN_B", 1, 60, &iav, mos_table);
  SCKGETC ("OUT_A", 1, 60, &iav, window_table);
  SCKGETC ("INPUTC", 1, 3, &iav, method);

  SCKRDI ("INPUTI", 1, 4, &iav, param, &kun, &knul);
  SCKRDR ("INPUTR", 1, 2, &iav, rpar, &kun, &knul);

/* map input file */

  SCIGET (obj_image, D_R4_FORMAT, F_I_MODE, F_IMA_TYPE,
	  2, &nax, Npix, Start, Step, inst, unit, (char **) &image, &imno);

  Method = GAUSSIAN;
  sprintf (linem, "centering method:    Gaussian");

/* read table MOS to get limits of slitlets  */

  TCMNUL(&inull,&rnull,&dnull);

  TCTOPN (mos_table, F_I_MODE, &tidmos);
  TCIGET (tidmos, &col, &nslits, &nsort, &acol, &arow);
  TCLSER (tidmos, "slit", &sl_col);
  TCLSER (tidmos, "ystart", &lo_col);
  TCLSER (tidmos, "yend", &up_col);

  for (i = 1; i <= nslits; i++)
    {
      TCSGET (tidmos, i, &selected);
      if (selected)
	{
	  TCERDI (tidmos, i, sl_col, &slit[i - 1], &null);
	  if (null)
	    slit[i - 1] = inull;
	  TCERDI (tidmos, i, lo_col, &lower[i - 1], &null);
	  if (null)
	    lower[i - 1] = inull;
	  else
	    lower[i - 1] = (lower[i-1]-Start[1])/Step[1] + 1;
	  TCERDI (tidmos, i, up_col, &upper[i - 1], &null);
	  if (null)
	    upper[i - 1] = inull;
	  else
	    upper[i - 1] = (upper[i-1]-Start[1])/Step[1] + 1;
	}
    }
  TCTCLO (tidmos);

/* open output table */

  TCTINI (window_table, F_TRANS, F_O_MODE, 7, MAXOBJECTS, &Tid);
  SCDWRD (Tid, "Pixel", Step, 1, 1, &kun);
  TCCINI (Tid, D_R4_FORMAT, 1, "F6.0", "None ", "Obj_Slit", &os_col);
  TCCINI (Tid, D_R4_FORMAT, 1, "F10.2", "Pixel", "Y_pos", &o1_col);
  TCCINI (Tid, D_R4_FORMAT, 1, "F10.2", "Pixel", "X_pos", &x_col);
  TCCINI (Tid, D_R4_FORMAT, 1, "F10.2", "Pixel", "Width", &o2_col);
  TCCINI (Tid, D_R4_FORMAT, 1, "E12.3", "Pixel", "Net_Intens", &ni_col);

/* set parameters */

  thresh = rpar[0];
  Width = param[0];
  Iwin = (param[0] - 1) / 2;
  Xbin = param[1];
  Xstep = param[2];

  SCTPUT ("search object ");
  SCTPUT ("------------\n");
  sprintf (line, "Input image:         %s ", obj_image);
  SCTPUT (line);
  sprintf (line, "Input table:         %s ", mos_table);
  SCTPUT (line);
  sprintf (line, "Output table:        %s\n ", window_table);
  SCTPUT (line);
  SCTPUT ("input parameters: ");
  sprintf (line, "search window:       %i pixels", Width);
  SCTPUT (line);
  sprintf (line, "detection threshold: %6.2f DN", thresh);
  SCTPUT (line);
  SCTPUT (linem);
  sprintf (line, "\naverage on:          %i scan lines", Xbin);
  SCTPUT (line);
  sprintf (line, "Step in X :          %i scan lines\n", Xstep);
  SCTPUT (line);

  Ntot = Start[1] + Step[1] * Npix[1];
  buff = (float *) osmmget (Ntot * sizeof (float));
  Xgaus = dvector (1, Width);
  Ygaus = dvector (1, Width);
  A = dvector (1, 3);

/* center objects */

  NseqO = 1;
  NseqS = 1;

  SCKRDR ("SCAN", 1, 1, &iav, &scan_pos, &kun, &knul);
  center = (int) ((scan_pos - Start[0]) / Step[0]);    

/* loop over all slitlets */

  for (nslit = 0; nslit <= nslits - 1; nslit++) 
    {
      if (slit[nslit] > inull)
	{
	  
	  sprintf(line,"Working on slitlet: %04i",nslit+1);
	  SCTPUT(line);
	  
	  obj[0] = slit[nslit];
	  
	  /* now search for objects ...  */
	  
	  find_objects (image, buff, objlist, nslit, &nact, 
			center, obj);
	  
	}
    }


  SCDWRI (Tid, "NSLIT",&nslits, 1, 1, &dunit);

/* epilogue ... */

  TCSINI (Tid);  TCTCLO (Tid);  SCSEPI ();

}

/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void 
find_objects (float image[], float rval[], int objlist[], int nslit, 
	      int *nact, int center, float obj[])
#else
void 
find_objects (image, rval, objlist, nslit, nact, center, obj)
     float image[], rval[], obj[];
     int objlist[], nslit, *nact, center;
#endif

{
  int number = 0, j, i, imax;
  float med, diff, max_intens, Thresh;

  average ( image, rval, center, 0, Ntot-1 );

  for (j = lower[nslit] - 1 + Iwin; j < upper[nslit] - 1 - Iwin; j++)
    { 
      med = heap_median (Width, &rval[j - Iwin]);
      diff = rval[j] - med;
      if (med < 0.0)
	med = -med;
      Thresh = (thresh > 0.0) ? thresh : -thresh * med;
        
      if (diff > Thresh)
	{
	  imax = j;
	  max_intens = rval[j];
	  for (i = j - Iwin; i <= j + Iwin; i++)
	    if (rval[i] > max_intens)
	      {
		max_intens = rval[i];
		imax = i;
	      }
	  objlist[number] = imax;
	  number++;
	}
    }

  i = 0;
  while (i < number - 1)
    {
      if ((objlist[i + 1] - objlist[i]) < Iwin)
        {
          number--;
          for (j = i + 1; j < number; j++)
            {
              objlist[j] = objlist[j + 1];
            }
          i--;                  /* check again (more than two detections!) */
        }
      i++;
    }
  
  printf("%i object(s) found\n",number);

  *nact = number;
  
  /* now follow the spectrum */  

  for ( i = 1 ; i <= number; i++) 
    {
      follow (image, rval, objlist[i], center, obj, nslit);
    }
  
}


/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void 
follow (float image[], float rval[], int objpos, int center, float obj[], 
int nslit)
#else
void 
follow (image, rval, objpos, center, obj, nslit)
     float image[], rval[], obj[];
     int objpos, center, nslit;
#endif
{
  int  jj,  index;

  index = objpos;
  for (jj = center; jj > 0 ; jj=jj-Xstep)
    { 
      average (image, rval, jj, index-Iwin, index+Iwin); 
      fit_obj (rval, index, obj, nslit, jj);
      index = (int) A[2];
    }

  index = objpos;
  for (jj = center+Xstep; jj < Npix[0] ; jj=jj+Xstep)
    { 
      average (image, rval, jj, index-Iwin, index+Iwin); 
      fit_obj (rval, index, obj, nslit, jj);
      index = (int) A[2];
    }
}

/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void 
fit_obj (float rval[], int index, float obj[], int nslit, int xpos)
#else
void 
fit_obj (rval, index, obj, nslit, xpos)
     float rval[], obj[];
     int index, nslit, xpos;
#endif
{
  int j, k;
  float back, med;
  char line[80];

/* Determine local background (but ensure to take it within the slitlet) */

  if (index - Iwin - 10 > lower[nslit] && index + Iwin + 10 < upper[nslit])
    {
      med = heap_median (Width, &rval[index - Iwin - 10]);
      back = (med + heap_median (Width, &rval[index + Iwin + 10])) / 
	(float) 2.0;
    }
  else if (index - Iwin - 10 <= lower[nslit])
    back = heap_median (Width, &rval[index + Iwin + 10]);
  else if (index + Iwin + 10 >= upper[nslit] - 1)
    back = heap_median (Width, &rval[index - Iwin - 10]);
  med = heap_median (Width, &rval[index]);
  if (back > med)
    back = med;

/* GAUSSIAN fit of object profile */

  A[1] = rval[index];
  A[2] = index + 1;
  A[3] = Step[1];
  
  obj[4] = (float) A[1]; /* net peak intensity */
  obj[3] = (float) A[3]; /* width */
  obj[1] = (float) A[2]; /* Y position */ 
  obj[2] = (float) xpos; /* X position */

  for (k = 1, j = index - Iwin; j <= index + Iwin; j++, k++)
    {
      Xgaus[k] = index + k - Iwin;
      Ygaus[k] = rval[j] - back;
    }

  fit_gauss (Xgaus, Ygaus, Width, A);
  
  obj[4] = (float) A[1]; /* net peak intensity */
  obj[3] = (float) A[3]; /* width */
  if ((fabs( (float) A[2] - obj[1])) < MAXDIFF ) 
    {
      obj[1] = (float) A[2]; /* Y position */
    }
  else
    {
      sprintf(line,"problem detected at xposition: %04d",xpos);
      SCTPUT(line);
      A[2] = obj[1];
    }
  obj[2] = (float) xpos; /* X position */
  obj[1] = (obj[1]-1.0)*Step[1]+Start[1];
  obj[2] = (obj[2]-1.0)*Step[0]+Start[0];

  TCEWRR (Tid, NseqO, os_col, &obj[0]);
  TCEWRR (Tid, NseqO, o1_col, &obj[1]);
  TCEWRR (Tid, NseqO, o2_col, &obj[3]);
  TCEWRR (Tid, NseqO, ni_col, &obj[4]);
  TCEWRR (Tid, NseqO, x_col,  &obj[2]);
  NseqO++;
}

/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void 
average (float image[], float rval[], int xcenter, int yfirst, int ylast)
#else
void  
average (image, rval, xcenter, yfirst, ylast)
     float image[], rval[];
     int xcenter, yfirst, ylast;
#endif

{
float sum;
int j, jj, k;


for (j = yfirst ; j <= ylast; j++)
  rval[j] = 0.0;

for (k = yfirst; k < ylast; k++) 
  if (k >= 0 && k < Npix[1] ) {
    {
      sum = 0;
      for (j = xcenter-Xbin/2-1; j <= xcenter+Xbin/2-0.5; j++)
	{     
	  jj = j + Npix[0] * k ;
	  rval[k] = rval[k] + image[jj];	
	  sum++;
	}
      rval[k] = rval[k] / sum;
    }
  }
  else {
    rval[k] = 0.0;
  }
}
