      PROGRAM LPOTAB
C
C++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
C
C.IDENTIFICATION
C  Program LPOTAB                      Version 1.0         01/02/1994
C  J. R. Walsh                         ST-ECF, ESO.
C
C.KEYWORDS
C  Spectropolarimetry; Linear polarization; Image to table conversion
C
C.PURPOSE
C  Forms a set of total intensity, linear polarization and polarization
C  position angle frames and their errors into a single table with
C  columns for:
C  X value, Intensity and error, Linear Polarization & error, Polarization
C                                     Position Angle and error 
C
C.ALGORITHM
C  Read six images, copy to table format
C  
C.INPUT/OUTPUT
C  The following keywords are used:
C
C  TOTINT/C/1/60       Input total intensity data frame 
C  TOTERR/C/1/60       Input total intensity error data frame 
C  LPOL/C/1/60         Input total intensity data frame 
C  LPOLERR/C/1/60      Input total intensity error data frame 
C  PA/C/1/60           Input total intensity data frame 
C  PAERR/C/1/60        Input total intensity error data frame 
C  LPTAB/C/1/60        Output table of I, p and theta and errors
C------------------------------------------------------------------------
C
      IMPLICIT NONE
C  
      INTEGER MADRID(1)            ! Psuedo virtual memory pointer
      INTEGER KNULL,KUNIT(1)       ! Unit no. and no. of null values
      INTEGER IAV                  ! No. of elements - data frames
      INTEGER STAT1,STAT           ! Error status I/O return flags
      INTEGER NAXIS1,NAXIS2        ! No. of dimensions - data frames
      INTEGER NAXIS3,NAXIS4        ! No. of dimensions - data frames
      INTEGER NAXIS5,NAXIS6        ! No. of dimensions - data frames
      INTEGER NPIX1(2),NPIX2(2)    ! No. of pixels per axis - data frames
      INTEGER NPIX3(2),NPIX4(2)    ! No. of pixels per axis - data frames
      INTEGER NPIX5(2),NPIX6(2)    ! No. of pixels per axis - data frames
      INTEGER PNT1,PNT2            ! Virtual memory pointer - input data frames
      INTEGER PNT3,PNT4            ! Virtual memory pointer - input data frames
      INTEGER PNT5,PNT6            ! Virtual memory pointer - input data frames
      INTEGER IMNO1,IMNO2          ! File no. of data - input data frames
      INTEGER IMNO3,IMNO4          ! File no. of data - input data frames
      INTEGER IMNO5,IMNO6          ! File no. of data - input data frames
      INTEGER I                    ! Array index
      INTEGER TABID                ! Table identifier
      INTEGER COLX                 ! Column number for X values
      INTEGER COLI                 ! Column number for total intensity values
      INTEGER COLIERR              ! Column number for total intensity error values
      INTEGER COLLP                ! Column number for linear polarization values
      INTEGER COLLPERR           ! Column number for linear pol. error values
      INTEGER COLPA                ! Column number for position angle values
      INTEGER COLPAERR             ! Column number for position angle error values
C
      DOUBLE PRECISION START1(2)   ! Start values for axis
      DOUBLE PRECISION START2(2)   ! Start values for axis
      DOUBLE PRECISION START3(2)   ! Start values for axis
      DOUBLE PRECISION START4(2)   ! Start values for axis
      DOUBLE PRECISION START5(2)   ! Start values for axis
      DOUBLE PRECISION START6(2)   ! Start values for axis
      DOUBLE PRECISION STEP1(2)    ! Step values for axis
      DOUBLE PRECISION STEP2(2)    ! Step values for axis
      DOUBLE PRECISION STEP3(2)    ! Step values for axis
      DOUBLE PRECISION STEP4(2)    ! Step values for axis
      DOUBLE PRECISION STEP5(2)    ! Step values for axis
      DOUBLE PRECISION STEP6(2)    ! Step values for axis
      DOUBLE PRECISION XVAL(4096)  ! Array of values for X axis
C
      CHARACTER*64 IDENT1,IDENT2   ! ASCII identification for frame
      CHARACTER*64 IDENT3,IDENT4   ! ASCII identification for frame
      CHARACTER*64 IDENT5,IDENT6   ! ASCII identification for frame
      CHARACTER*64 CUNIT1,CUNIT2   ! Unit of each axis and data
      CHARACTER*64 CUNIT3,CUNIT4   ! Unit of each axis and data
      CHARACTER*64 CUNIT5,CUNIT6   ! Unit of each axis and data
      CHARACTER*60 TOTINT          ! Name of input intensity image
      CHARACTER*60 TOTERR          ! Name of input intensity error image
      CHARACTER*60 LPOL            ! Name of input linear polarization image
      CHARACTER*60 LPOLERR         ! Name of input linear polarization error image
      CHARACTER*60 PA              ! Name of input position angle image
      CHARACTER*60 PAERR           ! Name of input position angle error image
      CHARACTER*60 LPTAB           ! Name of output linear polarization table
      CHARACTER*8 UNITX            ! Name of units for X values in output table
      CHARACTER*8 LABELX           ! Label for X values in output table
C
C  Include MIDAS definitions for the ST interfaces
C
      INCLUDE 'MID_INCLUDE:ST_DEF.INC'
C
C  Associate MADRID to the common block VMR for FORTRAN 77 pointer emulation
C
      COMMON/VMR/MADRID
C
C  Include MIDAS definitions for data values
C
      INCLUDE 'MID_INCLUDE:ST_DAT.INC'
C
C  Initialize IDENT and CUNIT for STIGET
C
      DATA IDENT1 /' '/
      DATA IDENT2 /' '/
      DATA IDENT3 /' '/
      DATA IDENT4 /' '/
      DATA IDENT5 /' '/
      DATA IDENT6 /' '/
      DATA CUNIT1 /' '/
      DATA CUNIT2 /' '/
      DATA CUNIT3 /' '/
      DATA CUNIT4 /' '/
      DATA CUNIT5 /' '/
      DATA CUNIT6 /' '/
C
C  Set up MIDAS environment and enable automatic error abort
C
      CALL STSPRO('LPOTAB')
C
C  Get name of first input data frame for total intensity 
C  Map as an image with pixels in floating point format
C
      CALL STKRDC('TOTINT',1,1,60,IAV,TOTINT,KUNIT,KNULL,STAT1)
      CALL STIGET(TOTINT,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,2,
     :            NAXIS1,NPIX1,START1,STEP1,IDENT1,CUNIT1,PNT1,
     :            IMNO1,STAT1)
C
C     Abort if the dimension of the image is not 1
C
      IF (NAXIS1.NE.1) THEN
        CALL STETER(1,' 1st input frame not 1-dimensional')
        GO TO 990
      ENDIF
C
C  Get name of second input data frame for total intensity error
C  Map as an image with pixels in floating point format
C
      CALL STKRDC('TOTERR',1,1,60,IAV,TOTERR,KUNIT,KNULL,STAT1)
      CALL STIGET(TOTERR,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,2,
     :            NAXIS2,NPIX2,START2,STEP2,IDENT2,CUNIT2,PNT2,
     :            IMNO2,STAT1)
C
C     Abort if the dimension of the image is not 1
C
      IF (NAXIS2.NE.1) THEN
        CALL STETER(1,' 2nd input frame not 1-dimensional')
        GO TO 990
      ENDIF
C
C  Check that dimensions of both frames the same
C
      IF (NPIX1(1).NE.NPIX2(1)) THEN
        CALL STETER(1,' Input images not same size')
        GO TO 990
      ENDIF
C  
C  Get name of third input data frame for linear polarization
C  Map as an image with pixels in floating point format
C
      CALL STKRDC('LPOL',1,1,60,IAV,LPOL,KUNIT,KNULL,STAT1)
      CALL STIGET(LPOL,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,2,
     :            NAXIS3,NPIX3,START3,STEP3,IDENT3,CUNIT3,PNT3,
     :            IMNO3,STAT1)
C
C     Abort if the dimension of the image is not 1
C
      IF (NAXIS3.NE.1) THEN
        CALL STETER(1,' 3rd input frame not 1-dimensional')
        GO TO 990
      ENDIF
C
C  Get name of fourth input data frame for linear polarization
C  error
C  Map as an image with pixels in floating point format
C
      CALL STKRDC('LPOLERR',1,1,60,IAV,LPOLERR,KUNIT,KNULL,STAT1)
      CALL STIGET(LPOLERR,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,2,
     :            NAXIS4,NPIX4,START4,STEP4,IDENT4,CUNIT4,PNT4,
     :            IMNO4,STAT1)
C
C     Abort if the dimension of the image is not 1
C
      IF (NAXIS4.NE.1) THEN
        CALL STETER(1,' 4th input frame not 1-dimensional')
        GO TO 990
      ENDIF
C
C  Check that dimensions of 4th frame same as 2nd
C
      IF (NPIX2(1).NE.NPIX4(1)) THEN
        CALL STETER(1,' Input images not same size')
        GO TO 990
      ENDIF
C  
C  Get name of fifth input data frame for polarization
C  position angle error
C  Map as an image with pixels in floating point format
C
      CALL STKRDC('PA',1,1,60,IAV,PA,KUNIT,KNULL,STAT1)
      CALL STIGET(PA,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,2,
     :            NAXIS5,NPIX5,START5,STEP5,IDENT5,CUNIT3,PNT5,
     :            IMNO5,STAT1)
C
C     Abort if the dimension of the image is not 1
C
      IF (NAXIS5.NE.1) THEN
        CALL STETER(1,' 5th input frame not 1-dimensional')
        GO TO 990
      ENDIF
C
C  Get name of sixth input data frame for polarization
C  position angle error
C  Map as an image with pixels in floating point format
C
      CALL STKRDC('PAERR',1,1,60,IAV,PAERR,KUNIT,KNULL,STAT1)
      CALL STIGET(PAERR,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,2,
     :            NAXIS6,NPIX6,START6,STEP6,IDENT6,CUNIT6,PNT6,
     :            IMNO6,STAT1)
C
C     Abort if the dimension of the image is not 1
C
      IF (NAXIS6.NE.1) THEN
        CALL STETER(1,' 6th input frame not 1-dimensional')
        GO TO 990
      ENDIF
C
C  Check that dimensions of 6th frame same as 2nd
C
      IF (NPIX2(1).NE.NPIX6(1)) THEN
        CALL STETER(1,' Input images not same size')
        GO TO 990
      ENDIF
C
C  From the X value array from the START and STEP values
C
      DO I=1,NPIX1(1),1
        XVAL(I)=START1(1) + DBLE(I-1)*STEP1(1)
      ENDDO     
      IF (XVAL(1).GT.1000.0) THEN
C
C  Assume units are Angstroms
C
        UNITX='Angstrom'
        LABELX='X'
      ELSE
        UNITX='Pixel'
        LABELX='X' 
      ENDIF
C
C  Get the name of the output Intensity, Polarization, Position
C  Angle and error table
C
      CALL STKRDC('LPTAB',1,1,60,IAV,LPTAB,KUNIT,KNULL,STAT)
C
C  Initialize the table with 7 columns
C
      CALL TBTINI(LPTAB,F_RECORD,F_O_MODE,7,NPIX1(1),TABID,STAT)
C
C  Create the X column
C
      CALL TBCINI(TABID,D_R4_FORMAT,1,'F12.5',UNITX,LABELX,
     :            COLX,STAT)
C
C  Create the total intensity column
C
      CALL TBCINI(TABID,D_R4_FORMAT,1,'F12.5',' ','Tot_Int',
     :            COLI,STAT)
C
C  Create the total intensity error column
C
      CALL TBCINI(TABID,D_R4_FORMAT,1,'F12.5',' ','Tot_Int_Err',
     :            COLIERR,STAT)
C
C  Create the Linear polarization column
C
      CALL TBCINI(TABID,D_R4_FORMAT,1,'F9.5',' ','Lin_Pol',
     :            COLLP,STAT)
C
C  Create the Linear polarization error column
C
      CALL TBCINI(TABID,D_R4_FORMAT,1,'F9.5','%','Lin_Pol_Err',
     :            COLLPERR,STAT)
C
C  Create the polarization position angle column
C
      CALL TBCINI(TABID,D_R4_FORMAT,1,'F9.4','Degree','Pol_PA',
     :            COLPA,STAT)
C
C  Create the polarization position angle error column
C
      CALL TBCINI(TABID,D_R4_FORMAT,1,'F9.4','Degree','Pol_PA_Err',
     :            COLPAERR,STAT)
C
C  Write the X values to the 1st column
C
      CALL TABLEX(TABID,COLX,NPIX1(1),XVAL,STAT)
      IF (STAT.NE.0) THEN
        CALL STTPUT(' Error writing 1st column of table',STAT)
        GO TO 990
      ENDIF
C
C  Write the total intensity values to the 2nd column
C
      CALL TABLEV(TABID,COLI,NPIX1(1),MADRID(PNT1),STAT)
      IF (STAT.NE.0) THEN
        CALL STTPUT(' Error writing 2nd column of table',STAT)
        GO TO 990
      ENDIF
C
C  Write the total intensity error values to the 3rd column
C
      CALL TABLEV(TABID,COLIERR,NPIX1(1),MADRID(PNT2),STAT)
      IF (STAT.NE.0) THEN
        CALL STTPUT(' Error writing 3rd column of table',STAT)
        GO TO 990
      ENDIF
C
C  Write the linear polarization values to the 4th column
C
      CALL TABLEV(TABID,COLLP,NPIX1(1),MADRID(PNT3),STAT)
      IF (STAT.NE.0) THEN
        CALL STTPUT(' Error writing 4th column of table',STAT)
        GO TO 990
      ENDIF
C
C  Write the linear polarization error values to the 5th column
C
      CALL TABLEV(TABID,COLLPERR,NPIX1(1),MADRID(PNT4),STAT)
      IF (STAT.NE.0) THEN
        CALL STTPUT(' Error writing 5th column of table',STAT)
        GO TO 990
      ENDIF
C
C  Write the polarization position angle values to the 6th column
C
      CALL TABLEV(TABID,COLPA,NPIX1(1),MADRID(PNT5),STAT)
      IF (STAT.NE.0) THEN
        CALL STTPUT(' Error writing 6th column of table',STAT)
        GO TO 990
      ENDIF
C
C  Write the polarization position angle values to the 7th column
C
      CALL TABLEV(TABID,COLPAERR,NPIX1(1),MADRID(PNT6),STAT)
      IF (STAT.NE.0) THEN
        CALL STTPUT(' Error writing 7th column of table',STAT)
        GO TO 990
      ENDIF
      GO TO 999
C
C  Error exit
C
990   CALL STTPUT(' Error writing to output table',STAT)   
      GO TO 999
C
C  Succesful exit. Release all files, update keywords and exit
C
999   CALL STSEPI

      END

      SUBROUTINE TABLEX(TABID,COL,N1,XVAL,STAT)
C
C  This subroutine puts each element of the double precision 
C  array XVAL into the rows of the column COL
C
      IMPLICIT NONE
      INTEGER TABID,COL,N1,STAT
      DOUBLE PRECISION XVAL(N1)
C
C  Local variables
C
      INTEGER I
C
C  Pick each element from the array and copy it to the output table
C 
      DO I=1,N1,1
        CALL TBEWRD(TABID,I,COL,XVAL(I),STAT)
      ENDDO
 
      END

      SUBROUTINE TABLEV(TABID,COL,N1,VAL,STAT)
C
C  This subroutine puts each element of the real array VAL 
C  into the rows of the column COL
C
      IMPLICIT NONE
      INTEGER TABID,COL,N1,STAT
      REAL VAL(N1)
C
C  Local variables
C
      INTEGER I
C
C  Pick each element from the array and copy it to the output table
C 
      DO I=1,N1,1
        CALL TBEWRR(TABID,I,COL,VAL(I),STAT)
      ENDDO
 
      END
