/* @(#)mosdefine.c      6.1.1.4  (LSW)  07/02/94  13:48:49           */
/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* .COPYRIGHT   (C) 1994 Landessternwarte Heidelberg                 */
/* .IDENT        rectify.c                                           */
/* .AUTHOR       Otmar Stahl - LSW Heidelberg                        */
/* .KEYWORDS                                                         */
/* .PURPOSE      Rectify spectra perpendicular to dispersion         */
/*               Input is a curved spectrogram and a table with      */
/*               polynomial coefficients for each slitlet            */
/*               (the latter is created with follow.c and regr/pol   */
/* .VERSION      1.0  07-FEB-1994  Implementation                    */
/*-------------------------------------------------------------------*/

#include <math.h>
#include <midas_def.h>

#include <proto_fors.h>

#define myabs(A) ((A)  >  0 ? (A) : (-A))
#define MAXSLITS 100
#define LINEAR   0

int Npix[2], Method, debug = 1;
double Start[2], Step[2];

#ifdef __STDC__
void reb_slit (float[], float[], int, int[], int[], int, double[], int);
#else
void reb_slit ();
#endif

main ()

{
  int i, iav, inull, nax, tidmos, tidfol, imno_in, imno_out, selected;
  int col, nslits, acol, arow, nsort, null, slit[MAXSLITS];
  int sl_col, up_col, lo_col, center, nslit, norder, kun, knul;
  int dunit, nulval, upper[MAXSLITS], lower[MAXSLITS];

  char method[21], in_image[60], out_image[60], mos_table[60];
  char descname[20], number[6];
  char follow_table[60], unit[64], inst[72], line[80];

  float *image_in, *image_out, rnull;

  double *coeffd, scan_pos, dnull;

/* get into Midas and get parameters */

  SCSPRO ("rectify");

  SCKGETC ("IN_A", 1, 60, &iav, in_image);
  SCKGETC ("IN_B", 1, 60, &iav, mos_table);
  SCKGETC ("IN_C", 1, 60, &iav, follow_table);
  SCKGETC ("OUT_A", 1, 60, &iav, out_image);
  SCKGETC ("INPUTC", 1, 3, &iav, method);

  SCKRDD ("SCAN", 1, 1, &iav, &scan_pos, &kun, &knul);

  Method = LINEAR;

  SCTPUT ("rectify slitlets ");
  SCTPUT ("------------\n");
  sprintf (line, "Input image:         %s ", in_image);
  SCTPUT (line);
  sprintf (line, "Mos   table:         %s ", mos_table);
  SCTPUT (line);
  sprintf (line, "Coefficients table:  %s\n ", follow_table);
  SCTPUT (line);
  sprintf (line, "Interpolation method: Linear\n");
  SCTPUT (line);

/* map input and output files */

  SCIGET (in_image, D_R4_FORMAT, F_I_MODE, F_IMA_TYPE,
	  2, &nax, Npix, Start, Step, inst, unit, 
	  (char **) &image_in, &imno_in);

  SCIPUT (out_image, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE,
	  2, Npix, Start, Step, inst, unit, 
	  (char **) &image_out, &imno_out);

  center = (int) ((scan_pos - Start[0]) / Step[0]);
  
/* read table MOS to get limits of slitlets  */

  TCMNUL(&inull,&rnull,&dnull);

  TCTOPN (mos_table, F_I_MODE, &tidmos);
  TCIGET (tidmos, &col, &nslits, &nsort, &acol, &arow);
  TCLSER (tidmos, "slit", &sl_col);
  TCLSER (tidmos, "ystart", &lo_col);
  TCLSER (tidmos, "yend", &up_col);

  for (i = 1; i <= nslits; i++)
    {
      TCSGET (tidmos, i, &selected);
      if (selected)
	{
	  TCERDI (tidmos, i, sl_col, &slit[i - 1], &null);
	  if (null)
	    slit[i - 1] = inull;
	  TCERDI (tidmos, i, lo_col, &lower[i - 1], &null);
	  if (null)
	    lower[i - 1] = inull;
	  else
	    lower[i - 1] =  (lower[i-1]-Start[1])/Step[1] + 1;
	  TCERDI (tidmos, i, up_col, &upper[i - 1], &null);
	  if (null)
	    upper[i - 1] = inull;
	  else
	    upper[i - 1] =  (upper[i-1]-Start[1])/Step[1] + 1;
	}
    }

  TCTCLO (tidmos);

  /* open table with coefficients */

  TCTOPN (follow_table, F_I_MODE, &tidfol);

  /* now loop over all slitlets */

  for (nslit = 0; nslit <= nslits - 1; nslit++) 
    {
      if (slit[nslit] > inull)
	{
	  sprintf (line, "Rectifying slitlet: %04i", nslit+1);
	  SCTPUT (line);
	  strcpy(descname,"REGR");
	  sprintf(number,"%04iI",nslit+1);
          strcat(descname,number);
	  SCDRDI(tidfol, descname, 6, 1, &iav, &norder, &dunit, &nulval);

	  coeffd = dvector (1,norder+1);

	  strcpy(descname,"REGR");
	  sprintf(number,"%04iD",nslit+1);
          strcat(descname,number);
	  SCDRDD(tidfol, descname, 1, norder+1, &iav, &coeffd[1], 
		 &dunit, &nulval);

	  reb_slit (image_in, image_out, nslit, upper, lower, center,
		       coeffd, norder);
	  
	  free_dvector(coeffd,1,norder+1);
	}
    }

  TCTCLO (tidfol);  SCSEPI ();

}

/*---------------------------------------------------------------------------*/

#ifdef __STDC__
void 
reb_slit (float image_in[], float image_out[], int nslit, 
	     int upper[], int lower[], int center, double coeffd[], int norder)
#else
void 
reb_slit (image_in, image_out, nslit, upper, lower, center, coeffd, norder)
     double coeffd[];
     float image_in[], image_out[];
     int upper[], lower[], nslit, center, norder;
#endif

{

int jj, j2, j,  i, k, kk;
double xpos, ypos, ypos0;
float frac1, frac2, diff;

/* reference position in x */

xpos = Start[0] + Step[0] * (double) (center);
ypos0 = eval_dpoly(xpos, coeffd, norder+1);
 

/* loop in x */

for (j = 0; j < Npix[0]; j++)
  {
    xpos = Start[0] + Step[0] * (double) j;  	     
    ypos = eval_dpoly(xpos, coeffd, norder+1);

/* how much to shift?, linear interpolation */ 

    diff = (float) (ypos0-ypos);
    k = (int) diff;
    if (diff < 0) k--; /* note that int (-3.2) = -3 */  
    frac1 = (float)( diff - (float) k );
    frac2 = (1.0 - frac1);

    
    if (j >= 450 && i < 500 && debug > 1)
      printf("%7.1f %10.3f %10.3f %10.3f %10.3f %10.3f \n",xpos,ypos,ypos0,diff,frac1,frac2);

/* loop along slitlet, shift all pixel in y by constant amount */

    for ( i = lower[nslit]-k; i <= upper[nslit]-k; i++)
      {
	jj = i*Npix[0];
	j2 = jj+Npix[0];
	kk = (i+k)*Npix[0];
	

	/* simple linear interpolation, to be improved !! */

	if ( i > 0 && i+k > 0 && i < Npix[1] && i+k < Npix[1])
	  {
	    image_out[kk+j] = frac1*image_in[jj+j]+frac2*image_in[j2+j];
	  }
      }
  }
}


#undef MAXSLITS
#undef myabs
#undef LINEAR
