/* spolstokes.c   */
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <midas_def.h>

#include <proto_fors.h>

#include <proto_polarimetry.h>
#include <proto_polarutils.h>


#define LINEAR 0
#define CIRCULAR 1
#define MAXIMS 16
#define myabs(A) ((A)  >  0 ? (A) : (-A))

#ifdef __STDC__
void frame_name(char frame[], char frame_err[], char rebmode[], int *, int *);
#else
void frame_name();
#endif

int main ()
     
{

  double start[2*MAXIMS],step[2*MAXIMS],ostep,ostart,eps;
  int iseq, frmcnt, stat, iav,uni,nulo,onpix,n,i,j,k,imno[6],imno_err[6];
  int polmode, null, tid, frame_col;
  int ncol, nrow, nsort, acol, arow, select;
  int imnol[2*MAXIMS],imno_errl[2*MAXIMS],npix[2*MAXIMS],naxis[2*MAXIMS],
    iselcol[2*MAXIMS+1];
  int offset,erroff,ipos[2];
  int phi_col,debug=1;

  char catfil[64],output[64],frame[64],frame_err[64],cunit[64];
  char outima[64],outerr[64],append[1],mode[3],rebmode[2];

  float *ao,*ae,*bo,*be,*ao_err,*ae_err,*bo_err,*be_err; 
  float *a, *aerr, *b, *berr;
  float phi,zeroang,*phic;

  static char  error1[] = "operands do not match in step size... ";
  static char  error2[] = "operands do not match in start value... ";
  static char  error3[] = "operands do not match in pixel number... ";







/*         get into Midas and get parameters             */

  SCSPRO ("stokes");

  stat = SCKGETC("IN_A",1,60,&iav,catfil);
  stat = SCKGETC("INPUTC",1,3,&iav,mode);
  stat = SCKGETC("REBMODE",1,2,&iav,rebmode);
  stat = SCKRDR("INPUTR",1,1,&iav,&zeroang,&uni,&null);
  stat = SCKRDI("INPUTI",1,2,&iav,ipos,&uni,&null);

  polmode = LINEAR;
  if (!strncmp (mode, "CIR", 3) || !strncmp (mode, "cir", 3))
    polmode = CIRCULAR;

  /* loop through catalogue */

  TCTOPN(catfil, F_I_MODE, &tid);
  TCIGET (tid, &ncol, &nrow, &nsort, &acol, &arow);

  if (toupper(rebmode[0]) == '1')
    TCCSER (tid, ":SC_REB1D", &frame_col);
  else if (toupper(rebmode[0]) == '2')
    TCCSER (tid, ":SC_REB", &frame_col);
  else
    return;
  TCCSER (tid, ":RET_ANG", &phi_col);

  phic = fvector(0, 2*ipos[0] - 1);

  iseq = -1;
  for (frmcnt=0; frmcnt<nrow; frmcnt++)           /* loop through table */
    {
      TCSGET (tid, frmcnt+1, &select);            
      if(select) {                               /* selected entries only */
	iseq++;
	iselcol[iseq] = frmcnt+1;
	stat = TCERDR(tid, frmcnt+1, phi_col, &phic[iseq], &null);
	phic[iseq] =  phic[iseq] * (float) ipos[1] + zeroang / 2.;

	stat = TCERDC(tid, frmcnt+1, frame_col, frame, &null);
	frame_name(frame,frame_err,rebmode,&imnol[iseq],&imno_errl[iseq]);
	
	stat = SCFOPN(frame,D_R4_FORMAT,0,F_IMA_TYPE,&imnol[iseq]); 
	
	/*
	if (debug>0)
	  printf("... %d %d  %s",iseq, frmcnt,frame);
	
	if (toupper(rebmode[1]) == 'E'){
	  i = strlen(frame);
	  strncpy(frame_err,frame,i);
	  frame_err[i]= 'E';
	  frame_err[i+1]='\0'; 
	  
	  stat = SCFOPN(frame_err,D_R4_FORMAT,0,F_IMA_TYPE,&imno_errl[iseq]);
	  
	if (debug>0)
	  printf("  %s   phic = %f\n",frame_err, phic[iseq]);
	}
	else if (toupper(rebmode[0]) == '1'){
	  imno_errl[iseq] = imnol[iseq];
	  if (debug>0)
	    printf("  phic = %f \n",phic[iseq]);
	}
	else
	  return;
	*/
	if (iseq > 0){
	  SCFCLO(imnol[iseq]);
	  if (toupper(rebmode[1]) == 'E'){
	    SCFCLO(imno_errl[iseq]);
	  }
	}

      }
    }

  /* read descriptors of first frame */

  stat = SCDRDI(imnol[0],"NAXIS",1,1,&iav,&naxis[0],&uni,&nulo);
  stat = SCDRDI(imnol[0],"NPIX",1,1,&iav,&npix[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"START",1,1,&iav,&start[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"STEP",1,1,&iav,&step[0],&uni,&nulo);
  stat = SCDGETC(imnol[0],"CUNIT",1,64,&iav,cunit);

  ostart = start[0];
  ostep = step[0];
  eps = 0.0001*step[0];
  onpix = npix[0];

  /* compare catalogue frames with first frame in catalogue */ 

  for (n=1; n < iseq; n++)
    {
      start[n] = 0.0;
      step[n] = 1.0;
      npix[n] = 1;
  
      frame[0] = '\0';
      stat = TCERDC(tid, iselcol[n], frame_col, frame, &null);
      frame_name(frame,frame_err,rebmode,&imnol[n],&imno_errl[n]);

      stat = SCDRDI(imnol[n],"NAXIS",1,1,&iav,&naxis[n],&uni,&nulo);
      stat = SCDRDI(imnol[n],"NPIX",1,1,&iav,&npix[n],&uni,&nulo);
      stat = SCDRDD(imnol[n],"START",1,1,&iav,&start[n],&uni,&nulo);
      stat = SCDRDD(imnol[n],"STEP",1,1,&iav,&step[n],&uni,&nulo);
      
      /*  check step size, start values and pixel numbers ...  */

      if (myabs(step[n]-ostep) > eps) SCETER(5,error1);
      if (myabs(start[n]-ostart) > eps) SCETER(5,error2);
      if (npix[n] != onpix) SCETER(5,error3);
      if (iseq > 0){
	SCFCLO(imnol[n]);
	if (toupper(rebmode[1]) == 'E'){
	  SCFCLO(imno_errl[n]);
	}
      }
    }
  stat = SCDRDI(imnol[0],"NAXIS",1,1,&iav,&naxis[0],&uni,&nulo);
  stat = SCDRDI(imnol[0],"NPIX",1,naxis[0],&iav,&npix[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"START",1,naxis[0],&iav,&start[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"STEP",1,naxis[0],&iav,&step[0],&uni,&nulo);

  

  /* all o.k. now allocate space for arrays */

  ao = fvector(0, npix[0] - 1);
  ae = fvector(0, npix[0] - 1);
  ao_err = fvector(0, npix[0] - 1);
  ae_err = fvector(0, npix[0] - 1);








  /* construct output file names by appending to catalogue name */
  a = fvector (0, 4*npix[0] - 1);
  b = fvector (0, 4*npix[0] - 1);
  aerr = fvector (0, 4*npix[0] - 1);
  berr = fvector (0, 4*npix[0] - 1);

      

  if (naxis[0] == 2 && toupper(rebmode[1]) != 'E') npix[1] = 4*npix[1]/2; 
  /* it means that sigma is appended in the second half of the frame */
  if (naxis[0] == 1) npix[1] = 4 , naxis[0]++; /*1D for extracted frames*/

  append[0] = 'a';
  stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
		    outima,  outerr, &imno[0], &imno_err[0]);
  append[0] = 'b';
  stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
		    outima,  outerr, &imno[1], &imno_err[1]);


  /*for any y row on the CCD*/
  for (j = 0; j < npix[1]/4; j++)
    {
      
      offset = j*npix[0]+1;
      
      
      /*reset sums*/
      for (i = 0; i < npix[0]; i++)
	{
	  a[i] = aerr[i] = b[i] = berr[i]  = 0;
	}
      
      
      /*for any retarder position*/
      for (i = 0; i < ipos[0]; i++)
	{
	  if (debug>1)
	  printf("add to fourier sum i = %d\n",i);

	  frame[0] = '\0';
	  stat = TCERDC(tid, iselcol[i*2], frame_col, frame, &null);
	  frame_name(frame,frame_err,rebmode,&imnol[i*2],&imno_errl[i*2]);
	  printf("... %d %s %s %8.2f\n",iselcol[i*2],frame,frame_err,
		 phic[i*2]);
	  frame[0] = '\0';
	  stat = TCERDC(tid, iselcol[i*2+1], frame_col, frame, &null);
	  frame_name(frame,frame_err,rebmode,&imnol[i*2+1],
		     &imno_errl[i*2+1]);
	  printf("... %d %s %s %8.2f\n",iselcol[i*2+1],frame,frame_err,
		 phic[i*2+1]);

	  SCFGET(imnol[i*2],offset,npix[0],&nulo, (char *) ao);
	  SCFGET(imnol[i*2+1],offset,npix[0],&nulo, (char *) ae);
	  
	  if (toupper(rebmode[1]) == 'E'){
	    SCFGET(imno_errl[i*2],offset,npix[0],&nulo, (char *) ao_err);
	    SCFGET(imno_errl[i*2+1],offset,npix[0],&nulo, (char *) ae_err);
	  }
	  else{
	    erroff = offset + npix[1]*npix[0]/4;
	    
	    SCFGET(imno_errl[i*2],erroff,npix[0],&nulo, (char *) ao_err);
	    SCFGET(imno_errl[i*2+1],erroff,npix[0],&nulo, (char *) ae_err);
	  }

	  SCFCLO(imnol[i*2]);
	  SCFCLO(imnol[i*2+1]);
	  if (toupper(rebmode[1]) == 'E'){
	    SCFCLO(imno_errl[i*2]);
	    SCFCLO(imno_errl[i*2+1]);
	  }


	  
	  sfourier(ao, ao_err, ae, ae_err, phic[i*2]/4.0,
		   &a[0], &aerr[0], 
		   &b[0], &berr[0], 
		   npix[0], ipos[0]);
	  sfourier(ao, ao_err, ae, ae_err, phic[i*2]/2.0,
		   &a[npix[0]], &aerr[npix[0]], 
		   &b[npix[0]], &berr[npix[0]], 
		   npix[0], ipos[0]);
	  sfourier(ao, ao_err, ae, ae_err, phic[i*2],
		   &a[2*npix[0]], &aerr[2*npix[0]],
		   &b[2*npix[0]], &berr[2*npix[0]], 
		   npix[0], ipos[0]);
	  sfourier(ao, ao_err, ae, ae_err, phic[i*2]*2.0,
		   &a[3*npix[0]], &aerr[3*npix[0]], 
		   &b[3*npix[0]], &berr[3*npix[0]], 
		   npix[0], ipos[0]);
	}


      for (i = 0; i < 4*npix[0]; i++) /* sfourier computes sigma**2 */
	{
	  aerr[i] = (aerr[i] > 0.0) ? (float) sqrt(aerr[i]) : 0.0;
	  berr[i] = (berr[i] > 0.0) ? (float) sqrt(berr[i]) : 0.0;
	}
      


      offset = j*npix[0]*4+1;

      stat = SCFPUT(imno[0], offset, npix[0]*4, (char*) a);
      stat = SCFPUT(imno_err[0], offset, npix[0]*4, (char*) aerr);  
      stat = SCFPUT(imno[1], offset, npix[0]*4, (char*) b);
      stat = SCFPUT(imno_err[1], offset, npix[0]*4, (char*) berr);  
      
      
    }

  if (debug>1)
    printf("close output frames\n");

  stat = SCFCLO(imno[0]);
  stat = SCFCLO(imno_err[0]);
  stat = SCFCLO(imno[1]);
  stat = SCFCLO(imno_err[1]);
  
  if (debug>1)
    printf("free output buffers\n");

  free_fvector (a, 0, 4*npix[0] - 1);
  free_fvector (b, 0, 4*npix[0] - 1);
  free_fvector (aerr,0, 4*npix[0] - 1);
  free_fvector (berr,0, 4*npix[0] - 1);

  if (debug>1)
    printf("free input buffers\n");



  for (n = 1 ; n <= iseq ; n++)
    stat = SCFCLO(imnol[n]);

 
  free_fvector(ao, 0, npix[0] - 1);  
  free_fvector(ae, 0, npix[0] - 1);  
  free_fvector(ao_err, 0, npix[0] - 1);  
  free_fvector(ae_err, 0, npix[0] - 1);  
 


  /*          epilogue ...                 */
  if (debug>1)
    printf("epilog\n");

  SCSEPI();
  exit(0);
}

/*---------------------------------------------------------------------------*/


#ifdef __STDC__
void frame_name(char frame[], char frame_err[], char rebmode[], int *iml, 
		int *im_errl)
#else
void frame_name(frame,frame_err,rebmode,iml,im_errl)
  char frame[], frame_err[], rebmode[];
  int *iml,*im_errl;
#endif

{
  int i,stat;

  frame_err[0] = '\0';

  if (strstrs(frame,".bdf") == 0)
    {
      if (toupper(rebmode[1]) == 'E'){
	sprintf(frame_err,"%sE.bdf\0",frame);
	strcat(frame,".bdf\0");
      }
      else
	{
	  strcat(frame,".bdf\0"); 
	}
    }
  else
    {
      if (toupper(rebmode[1]) == 'E'){
	i = strlen(frame) - 4;
	strncpy(frame_err,frame,i);
	strcat(frame_err,"E.bdf\0");
	
	frame[i] = '\0';
	strcat(frame,".bdf\0"); 
      }
    }
  
  stat = SCFOPN(frame,D_R4_FORMAT,0,F_IMA_TYPE,iml); 
  if (toupper(rebmode[1]) == 'E'){
    stat = SCFOPN(frame_err,D_R4_FORMAT,0,F_IMA_TYPE,im_errl);
  }
  else if (toupper(rebmode[0]) == '1'){
    im_errl = iml;
  }
  else{
    printf("unsupported rebin mode\n");
    exit(1);
  }
  
  return;
}
 
 

#undef LINEAR
#undef CIRCULAR
#undef MAXIMS
#undef myabs
