/* @(#)spolframeoff.c     6.1.1.4  (LSW)  07/02/94  13:48:49           */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* .COPYRIGHT   (C) 1994 Landessternwarte Heidelberg                 */
/* .IDENT        spolframeoff.c                                       */
/* .AUTHOR       Sabine Moehler - LSW Heidelberg                     */
/* .KEYWORDS                                                         */
/* .PURPOSE      Search object with median: (val - med(val)) > thres */ 
/*               Center object with Gaussian fitting                 */
/*               (adapted from mosslin.c)                            */
/* .VERSION      1.0  07-FEB-1994  Implementation                    */
/*               1.1  26-JUL-1994  Adaption for SPOLDEFINE           */
/*               1.2  18-OCT-1994  Adaption for SPOLYOFFSET          */
/*               1.3  19-OCT-1994  Adaption for SPOLFRAMEOFF         */
/*-------------------------------------------------------------------*/

#include <math.h> 
#include <tbldef.h>
#include <midas_def.h>
#include <ok.h>

#include <proto_mutil.h>
#include <proto_nrutil.h>

#define max(A,B) ((A)  > (B) ? (A) : (B)) 
#define min(A,B) ((A)  < (B) ? (A) : (B)) 

#define MAXLINES 1000
#define MAXSLITS 100

main()

{
   int   ny, i, k, ref_ang, tidout, tidoff;
   int   nrow, ncol, acol, arow, nsort;
   int   slitlow, slitup;
   int   out_col[4], off_col[6];
   int   comp, iav, kun, knul;

   char  ref_ima[60], obj_ima[60], test_ima[60], outtab[60], slitoff[60];

   double yoff, yref[MAXSLITS+1], yobj[MAXSLITS+1];
   double ylow, yup, yoffmean;

/* get into Midas and get parameters */

   SCSPRO("spolframeoff");
   
   SCKGETC("OUT_A",1,60,&iav,outtab);
   SCKGETC("SLITOFF",1,60,&iav,slitoff);
   SCKRDI("REF_ANG", 1, 1, &iav, &ref_ang, &kun, &knul);

/* open input table and get column numbers */
   TCTOPN (slitoff, F_IO_MODE, &tidoff);
   TCIGET (tidoff, &ncol, &nrow, &nsort, &acol, &arow);
   TCLSER (tidoff, "Frame", &off_col[0]);
   TCLSER (tidoff, "Slit_Low", &off_col[1]);
   TCLSER (tidoff, "Y_Low", &off_col[2]);
   TCLSER (tidoff, "Slit_Up", &off_col[3]);
   TCLSER (tidoff, "Y_Up", &off_col[4]);

/* intialize counters */

   k = 1;
   for (i = 0; i <= MAXSLITS; i++)
     {
      yref[i] = -99999;
      yobj[i] = -99999;
     }

/* get reference frame and data */

   TCERDC (tidoff, 1, off_col[0], ref_ima, &knul);
   for (i = 1; i <= nrow; i++)
     {
      TCERDC (tidoff, i, off_col[0], test_ima, &knul);
      comp = strcmp(ref_ima, test_ima);
      if (comp == 0)
        {
         TCERDI (tidoff, i, off_col[1], &slitlow, &knul);
         TCERDD (tidoff, i, off_col[2], &ylow, &knul);
         TCERDI (tidoff, i, off_col[3], &slitup, &knul);
         TCERDD (tidoff, i, off_col[4], &yup, &knul);
         yref[slitlow] = ylow;
         yref[slitup] = yup;
        }
     }

/* get object frame and data */
/* Since this routine is executed immediately after spolyoffset, the object
   frame must be the last in the table slitoff */

   TCERDC (tidoff, nrow, off_col[0], obj_ima, &knul);
   for (i = nrow; i >= 1; i--)
     {
      TCERDC (tidoff, i, off_col[0], test_ima, &knul);
      comp = strcmp(obj_ima, test_ima);
      if (comp == 0)
        {
         TCERDI (tidoff, i, off_col[1], &slitlow, &knul);
         TCERDD (tidoff, i, off_col[2], &ylow, &knul);
         TCERDI (tidoff, i, off_col[3], &slitup, &knul);
         TCERDD (tidoff, i, off_col[4], &yup, &knul);
         yobj[slitlow] = ylow;
         yobj[slitup] = yup;
        }
     }
   
/* open output table */

   TCTINI(outtab,F_TRANS,F_O_MODE,4,MAXLINES,&tidout);
   TCCINI(tidout,D_R4_FORMAT,1,"F6.0" , "None ","Slit",&out_col[0]);
   TCCINI(tidout,D_R4_FORMAT,1,"F10.2", "Pixel","Y_ref",&out_col[1]);
   TCCINI(tidout,D_R4_FORMAT,1,"F10.2", "Pixel","Y_in",&out_col[2]);
   TCCINI(tidout,D_R4_FORMAT,1,"F10.2", "Pixel","Y_off",&out_col[3]);

   yoffmean = 0;
   ny = 0;

   for (i = 0; i <= MAXSLITS; i++)
     {
      if (yref[i] > -99999 && yobj[i] > -99999)
        {
         yoff = yobj[i]-yref[i];
         TCEWRI (tidout, k, out_col[0], &i);
         TCEWRD (tidout, k, out_col[1], &yref[i]);
         TCEWRD (tidout, k, out_col[2], &yobj[i]);
         TCEWRD (tidout, k++, out_col[3], &yoff);
         yoffmean = yoffmean+yoff;
         ny++;
        }
     }
  yoffmean = yoffmean/ny;
  SCDWRD (tidout, "YOFFMEAN", &yoffmean, 1, 1, &kun);
  SCDWRC (tidout, "REFIMA", 1, ref_ima, 1, 60, &kun);
  SCDWRC (tidout, "OBJIMA", 1, obj_ima, 1, 60, &kun);
  SCDWRI (tidout, "REFANG", &ref_ang, 1, 1, &kun); 

/* epilogue ... */

   TCTCLO(tidoff);    
   TCSINI(tidout);    
   TCTCLO(tidout);    
   SCSEPI();
}
