#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <midas_def.h>
#include <proto_fors.h>


#ifdef __STDC__
int interpol(float x1[], float y1[], int n1, float x2[], float y2[], int n2);
#else
int interpol();
#endif


#ifndef PI
#define PI (acos((double) -1.0))
#endif
#define RAD (((double) 180.0) / PI)
#define HRAD (((double) 12.0) / PI)

#define SLENGTH 160


#define TORAD (PI/((double) 90))

main()
{

  char qout[SLENGTH+1],uout[SLENGTH+1],qin[SLENGTH+1],uin[SLENGTH+1],
    e1in[SLENGTH+1],e2in[SLENGTH+1],e3in[SLENGTH+1],
    cbuff[SLENGTH+1];

  int i,j;
  int imin1,imin2,imout1,imout2,imine1,imine2,imoute1,imoute2,
    tide1,tide2,tide3;
  int imp,imt,impe,imte,impc,istrlen;
  int nrow,kact,kun,knul,stat;
  int naxis,npix[2],fsize;
  int ilog[4],col[3];

  int debug;

  float *depstheta,*ddepol,*dqinspol,*duinspol,*dlambda,*debuff,*dlbuff;
  float *dqinbuff,*duinbuff,*dqoutbuff,*duoutbuff;
  float *dqinebuff,*duinebuff,*dqoutebuff,*duoutebuff;
  float *dpbuff,*dpebuff,*dtbuff,*dtebuff;

  double start[2],step[2],dparam[3];

  SCSPRO ("spolpolcor");

  if (debug > 1)
    printf("read keys\n");

  SCKGETC ("OUT_A", 1, SLENGTH, &kact, qout);
  SCKGETC ("OUT_B", 1, SLENGTH, &kact, uout);
  
  SCKGETC ("IN_A", 1, SLENGTH, &kact, qin);
  SCKGETC ("IN_B", 1, SLENGTH, &kact, uin);

  SCKGETC ("IN_1", 1, SLENGTH, &kact, e1in);
  SCKGETC ("IN_2", 1, SLENGTH, &kact, e2in);
  SCKGETC ("IN_3", 1, SLENGTH, &kact, e3in);

  
  if (!file_exists(qin,".bdf")) return;
  if (!file_exists(uin,".bdf")) return;
  
  ilog[0] = file_exists(e1in,".tbl");
  ilog[1] = file_exists(e2in,".tbl");
  ilog[2] = file_exists(e3in,".tbl");


  stat = SCFOPN(qin, D_R4_FORMAT, 0, F_IMA_TYPE, &imin1);
  stat = SCFOPN(uin, D_R4_FORMAT, 0, F_IMA_TYPE, &imin2);

  SCDRDI( imin1, "NAXIS", 1, 1, &kact, &naxis, &kun, &knul );
  SCDRDI( imin1, "NPIX",  1, naxis, &kact, &npix[0],  &kun, &knul );
  SCDRDD( imin1, "START", 1, naxis, &kact, &start[0], &kun, &knul );
  SCDRDD( imin1, "STEP",  1, naxis, &kact, &step[0],  &kun, &knul );

  SCKRDD ("OUTPUTD", 1, 3, &kact, dparam, &kun, &knul);
  SCKRDI ("INPUTI", 1, 1, &kact, &debug, &kun, &knul);


  if (debug > 1)
    printf("allocate buffers\n");


  depstheta =  fvector(0,npix[0]);
  ddepol    =  fvector(0,npix[0]);
  dqinspol  =  fvector(0,npix[0]);
  duinspol  =  fvector(0,npix[0]);
  dlambda   =  fvector(0,npix[0]);


  for (i = 0 ; i < npix[0] ; i++)
    dlambda[i] = (float) (start[0] + step[0] * (double) i);


  if (ilog[0]){

    if (debug > 1)
      printf("open table eps_theta\n");

    stat = TCTOPN(e1in, F_I_MODE, &tide1);

    SCDRDI( tide1, "TBLCONTR", 4, 1, &kact, &nrow, &kun, &knul );
    TCCSER (tide1, ":LAMBDA", &col[0]);
    TCCSER (tide1, ":EPSTHETA", &col[1]);
  
    debuff = fvector(1,nrow);
    dlbuff = fvector(1,nrow);

    for (i = 0 ; i < nrow ; i++)
      {
	TCERDR (tide1, i+1, col[0], &dlbuff[i+1], &knul);
	TCERDR (tide1, i+1, col[1], &debuff[i+1], &knul);
      }

    interpol(dlbuff,debuff,nrow,dlambda,depstheta,npix[0]);

    if (debug > 1)
      for ( i = 0 ; i <= 100 ; i++)
	printf("%f %f %f %f \n",dlbuff[i+1],debuff[i+1],dlambda[i],
	       depstheta[i]);

    free_fvector(debuff,1,nrow);
    free_fvector(dlbuff,1,nrow);
    
    TCTCLO(tide1);
  }
  else{
    for (i = 0 ; i < npix[0] ; i++)
      depstheta[i] = (float) 0.0;
  }





  if (ilog[1]){

    if (debug > 1)
      printf("open table depol\n");

    stat = TCTOPN(e2in, F_I_MODE, &tide2);

    SCDRDI( tide2, "TBLCONTR", 4, 1, &kact, &nrow, &kun, &knul );
    TCCSER (tide2, ":LAMBDA", &col[0]);
    TCCSER (tide2, ":DEPOL", &col[1]);
  
    debuff = fvector(1,nrow);
    dlbuff = fvector(1,nrow);

    for (i = 0 ; i < nrow ; i++)
      {
	TCERDR (tide2, i+1, col[0], &dlbuff[i+1], &knul);
	TCERDR (tide2, i+1, col[1], &debuff[i+1], &knul);
      }

    interpol(dlbuff,debuff,nrow,dlambda,ddepol,npix[0]);

    if (debug > 1)
      for ( i = 0 ; i <= 10 ; i++)
	printf("%f %f %f %f \n",dlbuff[i+1],debuff[i+1],dlambda[i],ddepol[i]);
    
    free_fvector(debuff,1,nrow);
    free_fvector(dlbuff,1,nrow);
    
    TCTCLO(tide2);
  }
  else{
    for (i = 0 ; i < npix[0] ; i++)
      ddepol[i] = (float) 1.0;
  }





  if (ilog[2]){

    if (debug > 1)
      printf("open tables for inspol Q,U\n");

    stat = TCTOPN(e3in, F_I_MODE, &tide3);

    SCDRDI( tide3, "TBLCONTR", 4, 1, &kact, &nrow, &kun, &knul );
    TCCSER (tide3, ":LAMBDA", &col[0]);
    TCCSER (tide3, ":QINSPOL", &col[1]);
    TCCSER (tide3, ":UINSPOL", &col[2]);
  
    debuff = fvector(1,nrow);
    dlbuff = fvector(1,nrow);

    for (i = 0 ; i < nrow ; i++)
      {
	TCERDR (tide3, i+1, col[0], &dlbuff[i+1], &knul);
	TCERDR (tide3, i+1, col[1], &debuff[i+1], &knul);
      }

    interpol(dlbuff,debuff,nrow,dlambda,dqinspol,npix[0]);

    if (debug > 1){
    printf("\n");
    for ( i = 0 ; i <= 20 ; i++)
      printf("%f %f %f %f \n",dlbuff[i+1],debuff[i+1],dlambda[i],dqinspol[i]);
    }


    for (i = 0 ; i < nrow ; i++)
      {
	TCERDR (tide3, i+1, col[2], &debuff[i+1], &knul);
      }

    interpol(dlbuff,debuff,nrow,dlambda,duinspol,npix[0]);

    if (debug > 1){
    printf("\n");
    for ( i = 0 ; i <= 20 ; i++)
      printf("%f %f %f %f \n",dlbuff[i+1],debuff[i+1],dlambda[i],duinspol[i]);
    }

    free_fvector(debuff,1,nrow);
    free_fvector(dlbuff,1,nrow);
    
    TCTCLO(tide3);


  }
  else{
    for (i = 0 ; i < npix[0] ; i++)
      dqinspol[i] = duinspol[i] = (float) 0.0;
  }






  if (debug > 1)
    printf("create output\n");

  dqinbuff =  fvector(0,npix[0]);  
  duinbuff =  fvector(0,npix[0]);  
  dqoutbuff =  fvector(0,npix[0]);  
  duoutbuff =  fvector(0,npix[0]);  
  dqinebuff =  fvector(0,npix[0]);  
  duinebuff =  fvector(0,npix[0]);  
  dqoutebuff =  fvector(0,npix[0]);  
  duoutebuff =  fvector(0,npix[0]);  

  if (naxis > 1)
    fsize = npix[0] * npix[1];
  else
    fsize = npix[0];
  if (naxis == 1)
    npix[1] = 1;

  printf("open frames %s %s fsize = %d\n",qout,uout,fsize);
  SCFCRE( qout, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &imout1 );
  SCFCRE( uout, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &imout2 );
  SCDCOP( imin1, imout1, 1, "");
  SCDCOP( imin2, imout2, 1, "");


  strcat(qin,"E\0");
  strcat(uin,"E\0");
  strcat(qout,"E\0");
  strcat(uout,"E\0");

  if (debug > 1)
    printf("open frames with/for standard deviation:\n %s %s\n",qin,uin);

  stat = SCFOPN(qin, D_R4_FORMAT, 0, F_IMA_TYPE, &imine1);
  stat = SCFOPN(uin, D_R4_FORMAT, 0, F_IMA_TYPE, &imine2);


  printf("%d\n",fsize);
  SCFCRE( qout, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &imoute1 );
  SCFCRE( uout, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &imoute2 );
  SCDCOP( imine1, imoute1, 1, "");
  SCDCOP( imine2, imoute2, 1, "");


  if (debug > 1)
    printf("open frames for p and theta\n");

  istrlen = strlen(qout) - 2;
  strncpy(cbuff,qout,istrlen);
  strcpy(&cbuff[istrlen],"\0");
  strcat(cbuff,"p\0");
  SCFCRE( cbuff, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &imp );
  strcat(cbuff,"E\0");
  SCFCRE( cbuff, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &impe );

  istrlen = strlen(qout) - 2;
  strncpy(cbuff,qout,istrlen);
  strcpy(&cbuff[istrlen],"\0");
  strcat(cbuff,"c\0");
  SCFCRE( cbuff, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &impc );

  strncpy(cbuff,qout,istrlen);
  strcpy(&cbuff[istrlen],"\0");
  strcat(cbuff,"t\0");
  SCFCRE( cbuff, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &imt );
  strcat(cbuff,"E\0");
  SCFCRE( cbuff, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, fsize, &imte );

  SCDCOP( imin1, imp, 1, "");
  SCDCOP( imin1, impe, 1, "");
  SCDCOP( imin1, imt, 1, "");
  SCDCOP( imin1, imte, 1, "");
  SCDCOP( imin1, impc, 1, "");

  dpbuff =  fvector(0,npix[0]);  
  dpebuff =  fvector(0,npix[0]);  
  dtbuff =  fvector(0,npix[0]);  
  dtebuff =  fvector(0,npix[0]);  


  if (debug > 1)
    printf("loop columns\n");


  for (j = 0 ; j < npix[1] ; j++){


    /*
      calculate corrected Stokes parameters Q,U
      */

    SCFGET(imin1,j*npix[0]+1,npix[0],&kact, (char *) dqinbuff);
    SCFGET(imin2,j*npix[0]+1,npix[0],&kact, (char *) duinbuff);

    for (i = 0; i < npix[0] ; i++)
      {


	dqoutbuff[i] = (float) ((dqinbuff[i] * cos(depstheta[i]*TORAD) + 
			 duinbuff[i] * sin(depstheta[i]*TORAD) ) / ddepol[i]);
	duoutbuff[i] = (float) (( duinbuff[i] * cos(depstheta[i]*TORAD) - 
			 dqinbuff[i] * sin(depstheta[i]*TORAD) ) / ddepol[i]);


	dqoutbuff[i] += (float) (dqinspol[i] * cos(dparam[2]*TORAD) + 
				 duinspol[i] * sin(dparam[2]*TORAD));
	duoutbuff[i] += (float) (duinspol[i] * cos(dparam[2]*TORAD) - 
				 dqinspol[i] * sin(dparam[2]*TORAD));

      }

    
    SCFPUT(imout1, j*npix[0]+1, npix[0], (char *) dqoutbuff);
    SCFPUT(imout2, j*npix[0]+1, npix[0], (char *) duoutbuff);


    /*
      calculate standard deviation of the corrected Stokes parameters Q,U
      */

    SCFGET(imine1,j*npix[0]+1,npix[0],&kact, (char *) dqinebuff);
    SCFGET(imine2,j*npix[0]+1,npix[0],&kact, (char *) duinebuff);

    for (i = 0; i < npix[0] ; i++)
      {
	dqoutebuff[i] = (float) (
	  sqrt( SQR(dqinebuff[i] * cos(depstheta[i]*TORAD)) + 
		SQR(duinebuff[i] * sin(depstheta[i]*TORAD) )) / ddepol[i]);
	duoutebuff[i] = (float) (
	  sqrt( SQR(duinebuff[i] * cos(depstheta[i]*TORAD)) + 
		SQR(dqinebuff[i] * sin(depstheta[i]*TORAD) )) / ddepol[i]);
      }
    
    SCFPUT(imoute1, j*npix[0]+1, npix[0], (char *) dqoutebuff);
    SCFPUT(imoute2, j*npix[0]+1, npix[0], (char *) duoutebuff);



    /*
      calculate p and theta and the standard deviation
      */

    linpol(dqoutbuff, dqoutebuff, duoutbuff, duoutebuff, dpbuff, dpebuff,
	   dtbuff, dtebuff, npix[0], 10, 1);

    SCFPUT(impc, j*npix[0]+1, npix[0], (char *) dpbuff);
    SCFPUT(impe, j*npix[0]+1, npix[0], (char *) dpebuff);
    SCFPUT(imt, j*npix[0]+1, npix[0], (char *) dtbuff);
    SCFPUT(imte, j*npix[0]+1, npix[0], (char *) dtebuff);

    linpol(dqoutbuff, dqoutebuff, duoutbuff, duoutebuff, dpbuff, dpebuff,
	   dtbuff, dtebuff, npix[0], 10, 0);

    SCFPUT(imp, j*npix[0]+1, npix[0], (char *) dpbuff);

  }

  if (debug > 1)
    for ( i = 1000 ; i <= 1100 ; i++)
      printf("%f %f %f %f \n",dqinbuff[i],duinbuff[i],dqoutbuff[i],
	     duoutbuff[i]);


  SCFCLO(imout1);
  SCFCLO(imout2);
  SCFCLO(imin1);
  SCFCLO(imin2);

  
  SCFCLO(imoute1);
  SCFCLO(imoute2);
  SCFCLO(imine1);
  SCFCLO(imine2);

  SCFCLO(imp);
  SCFCLO(impe);
  SCFCLO(imt);
  SCFCLO(imte);
  SCFCLO(impc);




  if (debug > 1)
    printf("dealocate memory\n");


  free_fvector(depstheta,0,npix[0]);
  free_fvector(ddepol ,0,npix[0]);
  free_fvector(dqinspol,0,npix[0]);
  free_fvector(duinspol,0,npix[0]);
  free_fvector(dlambda,0,npix[0]);



  SCSEPI ();
  exit(0);
  return;

}



#ifdef __STDC__
int     
interpol(float x1[], float y1[], int n1, float x2[], float y2[], int n2)
#else
int     
interpol(x1, y1, n1, x2, y2, n2)
int n1,n2;
float x1[],y1[],x2[],y2[];
#endif
{
  int i,istart=0;
  float *buff;

  buff = fvector(1,n1);

  spline(x1, y1, n1, 1.0e30, 1.0e30, buff);

  for (i = 0 ; i < n2 ; i++)
    splint(x1, y1, buff, n1, x2[i], &y2[i]);

  free_fvector(buff,1,n1);
  return 0;
}








