/* @(#)spolrebin1D.c    1.0.0.0 (ESO-La Silla) 10/08/91 12:00:00 */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
.IDENT        spolrebin1d.c
.MODULE       main program -- spolrebin1D.exe
.LANGUAGE     C
.AUTHOR       Cristian Levin - ESO La Silla
              Otmar Stahl, LSW, rebin "image" of extracted spectra !!!
              Sabine Moehler, LSW, rebin 1-dim. extracted spectrum !!!
.PURPOSE      linear, quadratic an spline rebinning.
.KEYWORDS     rebinning.
.VERSION 1.0  1-Apr-1991   Implementation
.        1.1  25-JUL-1994  Adaption for SPOL
.ENVIRONMENT  UNIX
------------------------------------------------------------*/
/* Changes with respect to mosrebin1d.c:

mosrebin1d uses a two-dimensional input frame with 2*nobjects rows,
          where the extracted frames and the corresponding frames
          are stored as image rows
          constructs the names of the output frames

spolrebin1d uses one frame for each object
reads the names of the output frames
The actual rebin algorithm is the same for both!

produce error frames, append E to name of output frames 
(Otmar Stahl, Jan.11, 1995)

*/
     
#include <tbldef.h>
#include <midas_def.h>
#include <proto_fors.h>
#include <proto_mos.h>


#define MAXINTENS   99999.0
#define MININTENS  -99999.0 

#define LINEAR    0
#define QUADRATIC 1
#define SPLINE    2

char    Coerbr[61];
double  Starti, Stepi;
double  Starto, Endo, Stepo;
int     Method, NpixOut, Npix[2], kun, knul;

float   **Coefs, Cutsi[2];
float   Imax = MININTENS,
        Imin = MAXINTENS;


main ()

{
  char inframe[60], outima[60], outerr[60], output[61];
  char number[6];

  int dunit;
  int dnull, imno, imno2, imno3;
  int status, actvals, null;	/* Std Interfaces variables */
  int naxis, ypix, i, id;
  int ypos, slit; 
 
  /* spolrebin1d uses one frame for each object - therefore ypos
     and slit are numbers instead of arrays */

  float *yin, *yerr_in, *yout, *yerr_out, rebpar[3];
  double  pixel;
  double *x, *win, *wout;
  
  SCSPRO ("spolrebin1d");

  status = SCKGETC ("IN_A", 1, 60, &actvals, inframe);
  if (status)
    SCTPUT ("Error while reading IN_A");

  status = SCKGETC ("IN_B", 1, 60, &actvals, Coerbr);
  if (status)
    SCTPUT ("Error while reading IN_B");

  status = SCKGETC ("OUT_A", 1, 60, &actvals, outima);
  if (status)
    SCTPUT ("Error while reading OUT_A");

  strcpy(outerr,outima);
  strcat(outerr,"E.bdf");
  strcat(outima,".bdf");
  printf("%s %s\n",outima, outerr);

  status = SCKRDI ("INPUTI", 1, 1, &actvals, &Method, &kun, &knul);
  
  /* open 1-dim. frame and read descriptors */

  status = SCFOPN(inframe, D_R4_FORMAT, 0, F_IMA_TYPE, &imno);
  if (status != 0) 
    {
      sprintf( output, "Frame %s invalid...", inframe );
      SCTPUT( output );
      return;
    }

  SCDRDI( imno, "NAXIS", 1, 1, &actvals, &naxis, &dunit, &dnull );
  SCDRDI( imno, "NPIX", 1, 2, &actvals, Npix, &dunit, &dnull );
  if ( naxis == 1 )
    Npix[1] = 1;
  SCDRDD( imno, "START", 1, 1, &actvals, &Starti, &dunit, &dnull );
  SCDRDD( imno, "STEP", 1, 1, &actvals, &Stepi, &dunit, &dnull );
  SCDRDR( imno, "LHCUTS", 1, 2, &actvals, Cutsi, &dunit, &dnull );

  /* ypos is the original y-position of the spectra (before extraction) */
  /* spolrebin1d uses one frame for each object - therefore ypos and slit are
     numbers instead of arrays */
  
  SCDRDI (imno, "YPOS", 1, 1, &actvals, &ypos, &dunit, &dnull);
  SCDRDI (imno, "SLIT", 1, 1, &actvals, &slit, &dunit, &dnull);
  
  if (TCTOPN(Coerbr, F_IO_MODE, &id))
    SCTPUT("Error opening table with dispersion coefficients");
  
  SCDRDR( id, "REBPAR", 1, 3, &actvals, rebpar, &dunit, &dnull );
  Starto = (double) rebpar[0];
  Endo   = (double) rebpar[1];
  Stepo  = (double) rebpar[2];
  NpixOut = (int) ((Endo - Starto) / Stepo + 0.5);
  
  TCTCLO( id );
  
  mos_initdisp(Coerbr, "OLD", 0);
  x    = dvector(0, Npix[0] - 1);
  win  = dvector(0, Npix[0] - 1);
  wout = dvector(0, NpixOut - 1);
  yin  = fvector(0, Npix[0] - 1);
  yerr_in  = fvector(0, Npix[0] - 1);
  
  for ( i = 0; i < Npix[0]; i++ )
    x[i]  = (double) (Starti + Stepi * i);
  for ( i = 0; i < NpixOut; i++ )
    wout[i] = (double) (Starto + Stepo * i);

  /* spolrebin1d uses one frame for each object - 
     therefore only one call of rebin
     instead of looping over the objects  */

  SCFGET(imno,         1, Npix[0], &null, (char *)yin); 
  SCFGET(imno, Npix[0]+1, Npix[0], &null, (char *)yerr_in);

  status = mos_readdisp(ypos,slit);
  if (status == 0) 
    {
      /*image*/
      mos_eval_disp(x, win, Npix[0]);      
      SCIPUT(outima, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, 
	     1, &NpixOut, &Starto, &Stepo, 
	     "rebinned image", "uncalibrated    Angstrom",(char **)&yout, 
	     &imno2);
      rebin (win, wout, yin, yout, Npix[0], NpixOut, Stepo, Method);
      SCFCLO(imno2);
      /*image standard deviation*/
      SCIPUT(outerr, D_R4_FORMAT, F_O_MODE, F_IMA_TYPE, 
	     1, &NpixOut, &Starto, &Stepo, 
	     "standard deviation","uncalibrated    Angstrom", 
	     (char **)&yerr_out, &imno3);
      rebin (win, wout, yerr_in, yerr_out, Npix[0], NpixOut, Stepo, Method);
      SCFCLO(imno3);
    }
  else
    {
      printf("error from mos_readdisp\n");
    }
  finishdisp();

  /* Close everything and good bye */

  free_dvector(x, 0, Npix[0] - 1);
  free_dvector(win, 0, Npix[0] - 1);
  free_dvector(wout, 0, NpixOut - 1);
  free_fvector(yin, 0, Npix[0] - 1);
  free_fvector(yerr_in, 0, Npix[0] - 1);

  SCFCLO (imno); 
  SCSEPI ();
}



