/*                                                                         */
/* Sort SPOL frames obtained for one object. As input table an associated  */
/* table produced by the data organzier is used. All science frames that   */
/* have the same values as the reference image for the descriptors listed  */
/* in INPUTC will be copied to OUT_A, including their associated calibra-  */
/* frames.                                                                 */
/* S. Moehler 6.1994                                                       */

#include <tbldef.h>
#include <midas_def.h>
#include <ok.h>

main ()

{
  int i, j, k, l, c, d, r, C, D, R, I, iav, t2;
  int tid1, nrow1, ncol1, arow1, acol1, nsort1, icol1;
  int tid2, nrow2, ncol2, arow2, acol2, nsort2, *icol2;
  int tid3, nrow3, ncol3, arow3, acol3, nsort3, icol3[3];
  int num_des, sum_des, select, kun, knul, *d_num, *des_i, tes_i;
  int store_i, data_type, alen, ref_no, tes_no;
  int c_len, inull, I_num[20], D_num[20], R_num[20];

  char store_c[80], form[9], unit[17], label[17];
  char line[80], ref_image[60], test_ima[60];
  char des_list[60], ass_table[60], spol_table[60];
  char descr[100][9], *d_typ, des_c[20][257], tes_c[257];
  char D_name[20][9], C_name[20][9], R_name[20][9], I_name[20][9];

  float des_r[20], tes_r;
  float rnull;

  double des_d[20], tes_d;
  double dnull;

/*         get into Midas and get parameters             */

  SCSPRO ("spolsort");
  
  TCMNUL(&inull, &rnull, &dnull);

  SCKGETC ("IN_A", 1, 60, &iav, ass_table);
  SCKGETC ("IN_B", 1, 60, &iav, ref_image);
  SCKGETC ("OUT_A", 1, 60, &iav, spol_table);
  SCKGETC ("INPUTC", 1, 60, &iav, des_list);

/* Get list of descriptors to be checked */

  TCTOPN (des_list, F_I_MODE, &tid3);
  TCIGET (tid3, &ncol3, &nrow3, &nsort3, &acol3, &arow3);
  TCCSER (tid3, ":DESCR",&icol3[0]);
  TCCSER (tid3, ":D_TYP",&icol3[1]);
  TCCSER (tid3, ":D_NUM",&icol3[2]);
    
  d_typ = (char *) osmmget((ncol3+1)*sizeof(char));
  d_num = (int *) osmmget((ncol3+1)*sizeof(int));
  des_i = (int *) osmmget((nrow3+1)*sizeof(int));

  for (j = 0; j < nrow3; j++)
    {
      TCSGET (tid3, j+1, &select);
      if (select) 
        {
         TCERDC (tid3, j+1, icol3[0], &descr[j][0], &knul);
         TCERDC (tid3, j+1, icol3[1], &d_typ[j], &knul);
         TCERDI (tid3, j+1, icol3[2], &d_num[j], &knul);
        }
    }

/* initialize counters */

  c = 0;
  i = 0;
  r = 0;
  d = 0;
  t2 = 1; 
  C = 0;
  D = 0;
  I = 0;
  R = 0;

/* Read descriptors of ref_frame */

  SCFOPN (ref_image, D_R8_FORMAT, 0, F_IMA_TYPE, &ref_no); 

  for (j = 0; j <= nrow3-1; j++)
    {
     if (toupper(d_typ[j]) == 'C') 
       {
        SCDRDC (ref_no, &descr[j][0], 1, 1, 72, &iav, &des_c[c][1], &kun, &knul);
        TCERDC (tid3, j+1, icol3[0], &C_name[c][1], &knul);
        c++;
       }

     else if (toupper(d_typ[j]) == 'D') 
       {
        SCDRDD (ref_no, &descr[j][0], d_num[j], 1, &iav, &des_d[d], &kun, &knul);
        TCERDC (tid3, j+1, icol3[0], &D_name[d][1], &knul);
        TCERDI (tid3, j+1, icol3[2], &D_num[d], &knul);
        d++;
       }

     else if (toupper(d_typ[j]) == 'I') 
       {
        SCDRDI (ref_no, &descr[j][0], d_num[j], 1, &iav, &des_i[i], &kun, &knul);
        TCERDC (tid3, j+1, icol3[0], &I_name[i][1], &knul);
        TCERDI (tid3, j+1, icol3[2], &I_num[i], &knul);
        i++;
       }

     else if (toupper(d_typ[j]) == 'R') 
       {
        SCDRDR (ref_no, &descr[j][0], d_num[j], 1, &iav, &des_r[r], &kun, &knul);
        TCERDC (tid3, j+1, icol3[0], &R_name[r][1], &knul);
        TCERDI (tid3, j+1, icol3[2], &R_num[r], &knul);
        r++;
       }
     else 
       {
        sprintf(line,"Wrong type: %s\n", &d_typ[j]);
        SCTPUT(line);
       }
    }

SCFCLO(ref_no); 

sum_des = c+i+r+d;

/* open table with frame names */

  TCTOPN (ass_table, F_I_MODE, &tid1);
  TCIGET (tid1, &ncol1, &nrow1, &nsort1, &acol1, &arow1);
  TCCSER (tid1, ":SC",&icol1);

/* create output table with same columns as ass_table */
  
  icol2 = (int *) osmmget((ncol1+1)*sizeof(int));
  TCTINI(spol_table, F_TRANS, F_O_MODE, acol1, arow1, &tid2); 
  for (j = 1; j <= ncol1; j++)
    {
     TCFGET (tid1, j, form, &alen, &data_type);
     TCUGET (tid1, j, unit);
     TCLGET (tid1, j, label); 
     TCCINI (tid2, data_type, alen, form, unit, label, &icol2[j-1]);
    }
  TCIGET (tid2, &ncol2, &nrow2, &nsort2, &acol2, &arow2);

/* check descriptors of all other science frames */

  for (k = 1; k <= nrow1; k++)
    {
     TCERDC(tid1, k, icol1, test_ima, &knul); 
     SCFOPN(test_ima, D_R8_FORMAT, 0, F_IMA_TYPE, &tes_no); 
     num_des = sum_des;     

     for (j = 0; j <= c-1; j++)
        {
         SCDRDC(tes_no, &C_name[j][1], 1, 1, 60, &iav, &tes_c[1], &kun, &knul);
         if (tes_c[1] == des_c[j][1]) num_des--;
        }
     for (j = 0; j <= d-1; j++)
        {
         SCDRDD(tes_no, &D_name[j][1], D_num[j],  1,&iav, &tes_d, &kun, &knul);
         if (tes_d == des_d[j]) num_des--;
        }
     for (j = 0; j <= i-1; j++)
        {
         SCDRDI(tes_no, &I_name[j][1], I_num[j],  1,&iav, &tes_i, &kun, &knul);
         if (tes_i == des_i[j]) num_des--;
        }
     for (j = 0; j <= r-1; j++)
        {
         SCDRDR(tes_no, &R_name[j][1], R_num[j],  1,&iav, &tes_r, &kun, &knul);
         if (tes_r == des_r[j]) num_des--;
        }
      if (num_des == 0) 
        {
         for (l = 1; l <= ncol1; l++)
           {
            TCFGET (tid1, l, form, &alen, &data_type);
            if (data_type == 30)
              {
               TCERDC (tid1, k, l, store_c, &knul);
               TCEWRC (tid2, t2, l, store_c);
              }
            else if (data_type == 4)
              {
               TCERDI (tid1, k, l, &store_i, &knul);
               if (knul) store_i = inull;
               TCEWRI (tid2, t2, l, &store_i); 
              }
           }
         t2++;
        }
     SCFCLO(tes_no); 
    }
TCTCLO(tid1);
TCTCLO(tid2);
TCTCLO(tid3);
SCSEPI ();
}
