/* sorts extracted SPOL frames and groups them together into one table
   for each object, sorting in the order of the angle of the retarder
   plate. The zero angle of the retarder plate is read from keyword
   ZERO_ANG and it is assumed that the angle increases with steps of 22.5
   degrees */
#include <tbldef.h>
#include <midas_def.h>
#include <ok.h>

main ()
{
  int not_found, found, istart, m, iref, i_ang, i, k, l, iav;
  int kun, knul, tid, tide, tido;
  int ncol, nrow, acol, arow, nsort;
  int nrowe, ncole, acole, arowe, nsorte;
  int zero=0, nrowo, ncolo, acolo, arowo, nsorto;
  int null[3];
  int ocol[5], col[5], ecol[5];

  float zero_ref, ref, ret, zero_ang;
  char line[80], number[2], ex_table[60], tmp[60];
  char zeroc[1], frame_col[17], obj[80], outtab[60];

/*         get into Midas and get parameters             */

  SCSPRO("spolsortex");

  SCKGETC ("IN_A", 1, 60, &iav, ex_table);
  SCKGETC ("IN_B", 1, 60, &iav, frame_col);
  SCKGETC ("OUT_A", 1, 60, &iav, outtab);
  SCKGETC ("TMP", 1, 60, &iav, tmp);
  SCKRDI ("I", 1, 1, &iav, &i, &kun, &knul);
  SCKRDR ("ZERO_ANG", 1, 1, &iav, &zero_ang, &kun, &knul);

/* get name of spol_tableEX, original science frame, column with input frame
   column with extracted frames etc.*/
/* open table spol_tableEX and search for columns */

  TCTOPN (ex_table, F_IO_MODE, &tide);
  TCIGET (tide, &ncole, &nrowe, &nsorte, &acole, &arowe);
  TCCSER (tide, frame_col, &ecol[0]);
  TCCSER (tide, ":RET_ANG", &ecol[2]);
/* open temporary table and search for columns */
  TCTOPN (tmp, F_O_MODE, &tid);
  TCIGET (tid, &ncol, &nrow, &nsort, &acol, &arow);
  TCCSER (tid, frame_col, &col[0]);
  TCCSER (tid, ":RET_ANG", &col[2]);

  k=1;
/* construct name of reference output frame */
  if (i < 10)
    {
     sprintf(zeroc, "%d", zero);
     sprintf(number, "%d", i);
     strcat(zeroc,number);
     strncpy(number,zeroc,2);
    }
  if (i >= 10)
    sprintf(number, "%2d", i);
/* compare frame names */
   for (l = 1; l <= nrowe; l++)
     {
      TCERDC (tide, l, ecol[0], obj, null);
      if (obj[2] == number[1] && obj[1] == number[0])
        {
         TCEWRC(tid, k, col[0], obj);
         TCERDR (tide, l, ecol[2], &ret, null);
         TCEWRR (tid, k++, col[2], &ret);
        }
     }
/* open table spol_tableEX{i} and search for columns */
  TCTOPN (outtab, F_O_MODE, &tido);
  TCIGET (tido, &ncolo, &nrowo, &nsorto, &acolo, &arowo);
  TCCSER (tido, frame_col, &ocol[0]);
  TCCSER (tido, ":RET_ANG", &ocol[2]);
  sprintf (line, "\nInput table:                        %s", ex_table);
  SCTPUT (line);
  sprintf (line, "Output table for object %3d:        %s", i, outtab);
  SCTPUT (line);
  
  not_found = 0;
  found = 0;
  k = 1;
  zero_ref = zero_ang;
  for (iref = 1; iref <= 16; iref++)
    {
     i_ang = iref;
     while (i_ang > 4)
       {
        i_ang = i_ang-4;
        zero_ref = zero_ref+90;
       }
     if (i_ang == 1) ref = zero_ref;
     if (i_ang == 2) ref = zero_ref+45;
     if (i_ang == 3) ref = zero_ref+22.5;
     if (i_ang == 4) ref = zero_ref+67.5;
     for (l = 1; l <= 32; l++)
       {
        TCERDR (tid, l, col[2], &ret, null);
        TCERDC (tid, l, col[0], obj, null);
        if (ret == ref && obj[0] == 'O')
          {
           TCEWRR (tido, k, ocol[2], &ret);
           TCEWRC(tido, k, ocol[0], obj);
           found++;
          }
        if (ret == ref && obj[0] == 'E')
          {
           TCEWRR (tido, k+1, ocol[2], &ret);
           TCEWRC(tido, k+1, ocol[0], obj);
           found++;
          }
       }
    if (found == 2) 
      {
       k = k+2;
       not_found = 0;
      }
    else if (found < 2)
      {
       sprintf(line,"             *** frame(s) missing for angle = %9.3f degrees ***", ref);
       SCTPUT(line);
       sprintf(line,"*** all frames with angles belonging to this set deleted from table %s ***", outtab);
       SCTPUT(line);
/* find starting row for that set of retarder plate angles */
       not_found++;
       istart = k/8;
       iref = (istart+not_found)*4;
       istart = istart*8+1;
       if (istart < k-1)
         for (m = istart; m <= k; m++)
            TCRDEL(tido, m); 
       k = istart;
      } 
    zero_ref = zero_ang;
    found = 0;
   }
  
/*          epilogue ...                 */

  TCSINI(tid);
  TCTCLO(tid);
  TCSINI(tido);
  TCTCLO(tido);
  TCTCLO(tide);
  SCSEPI ();

}
