/* spolstokes.c   */
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <midas_def.h>

#include <proto_fors.h>

#include <proto_polarimetry.h>
#include <proto_polarutils.h>


#define LINEAR 0
#define CIRCULAR 1
#define MAXIMS 16
#define myabs(A) ((A)  >  0 ? (A) : (-A))
#define STRLEN 80
#ifndef PI
#define PI (acos((double) -1))
#endif
#define RAD (((double) 180) / PI)

#ifdef __STDC__
void frame_name(char frame[], char frame_err[], char rebmode[], int *, int *);
#else
void frame_name();
#endif


int main ()
     
{

  double start[2*MAXIMS],step[2*MAXIMS],ostep,ostart,eps;
  int iseq, frmcnt, stat, iav,uni,nulo,onpix,n,i,j,k,imno[6],imno_err[6];
  int polmode, null, tid, frame_col;
  int ncol, nrow, nsort, acol, arow, select;
  int imnol[2*MAXIMS+1],imno_errl[2*MAXIMS+1],npix[2*MAXIMS+1],
    naxis[2*MAXIMS+1],iselcol[2*MAXIMS+1];
  int offset,erroff,ipos[2];
  int phi_col;
  

  char catfil[STRLEN+1],output[STRLEN+1],frame[STRLEN+1], frame_err[STRLEN+1],
    cunit[STRLEN+1],outima[STRLEN+1],outerr[STRLEN+1],
    append[1],mode[3],rebmode[2];

  float *ao,*ae,*bo,*be,*ao_err,*ae_err,*bo_err,*be_err; 
  float *buffima,*bufferr;
  float *Q,*Qerr,*U,*Uerr,*pol, *polerr, *theta, *thetaerr, *phic;
  float phi,zeroang;

  static char  error1[] = "operands do not match in step size... ";
  static char  error2[] = "operands do not match in start value... ";
  static char  error3[] = "operands do not match in pixel number... ";







/*         get into Midas and get parameters             */

  SCSPRO ("stokes");

  stat = SCKGETC("IN_A",1,STRLEN,&iav,catfil);
  stat = SCKGETC("INPUTC",1,3,&iav,mode);
  stat = SCKGETC("REBMODE",1,2,&iav,rebmode);
  stat = SCKRDR("INPUTR",1,1,&iav,&zeroang,&uni,&null);
  stat = SCKRDI("INPUTI",1,2,&iav,ipos,&uni,&null);

  polmode = LINEAR;
  if (!strncmp (mode, "CIR", 3) || !strncmp (mode, "cir", 3))
    polmode = CIRCULAR;

  /* loop through catalogue */

  TCTOPN(catfil, F_I_MODE, &tid);
  TCIGET (tid, &ncol, &nrow, &nsort, &acol, &arow);

  if (toupper(rebmode[0]) == '1')
    TCCSER (tid, ":SC_REB1D", &frame_col);
  else if (toupper(rebmode[0]) == '2')
    TCCSER (tid, ":SC_REB", &frame_col);
  else
    return;
  TCCSER (tid, ":RET_ANG", &phi_col);

  phic = fvector(0, 2*ipos[0] - 1);

  iseq = -1;
  for (frmcnt=0; frmcnt<nrow; frmcnt++)           /* loop through table */
    {
      TCSGET (tid, frmcnt+1, &select);            
      if(select) {                              /* selected entries only */
	iseq++;
	iselcol[iseq] = frmcnt+1;
	stat = TCERDR(tid, iselcol[iseq], phi_col, &phic[iseq], &null);
	if (polmode == LINEAR)
	  phic[iseq] =  phic[iseq] * (float) ipos[1] + zeroang / 2.;
	else
	  phic[iseq] =  phic[iseq] * (float) ipos[1];

	frame[0] = '\0';
	stat = TCERDC(tid, iselcol[iseq], frame_col, frame, &null);

	frame_name(frame,frame_err,rebmode,&imnol[iseq],&imno_errl[iseq]);

	if (iseq > 0){
	  SCFCLO(imnol[iseq]);
	  if (toupper(rebmode[1]) == 'E')
	    SCFCLO(imno_errl[iseq]);
	}

      }
    }

  /* read descriptors of first frame */

  stat = SCDRDI(imnol[0],"NAXIS",1,1,&iav,&naxis[0],&uni,&nulo);
  stat = SCDRDI(imnol[0],"NPIX",1,1,&iav,&npix[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"START",1,1,&iav,&start[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"STEP",1,1,&iav,&step[0],&uni,&nulo);
  stat = SCDGETC(imnol[0],"CUNIT",1,64,&iav,cunit);

  ostart = start[0];
  ostep = step[0];
  eps = 0.0001*step[0];
  onpix = npix[0];

  /* compare catalogue frames with first frame in catalogue */ 

  for (n=1; n < iseq; n++)
    {
      start[n] = 0.0;
      step[n] = 1.0;
      npix[n] = 1;
  
      frame[0] = '\0';
      stat = TCERDC(tid, iselcol[n], frame_col, frame, &null);
      frame_name(frame,frame_err,rebmode,&imnol[n],&imno_errl[n]);

      stat = SCDRDI(imnol[n],"NAXIS",1,1,&iav,&naxis[n],&uni,&nulo);
      stat = SCDRDI(imnol[n],"NPIX",1,1,&iav,&npix[n],&uni,&nulo);
      stat = SCDRDD(imnol[n],"START",1,1,&iav,&start[n],&uni,&nulo);
      stat = SCDRDD(imnol[n],"STEP",1,1,&iav,&step[n],&uni,&nulo);
      
      /*  check step size, start values and pixel numbers ...  */

      if (myabs(step[n]-ostep) > eps) SCETER(5,error1);
      if (myabs(start[n]-ostart) > eps) SCETER(5,error2);
      if (npix[n] != onpix) SCETER(5,error3);

      if (iseq > 0){
	SCFCLO(imnol[n]);
	if (toupper(rebmode[1]) == 'E')
	  SCFCLO(imno_errl[n]);
      }

    }

  stat = SCDRDI(imnol[0],"NAXIS",1,1,&iav,&naxis[0],&uni,&nulo);
  stat = SCDRDI(imnol[0],"NPIX",1,naxis[0],&iav,&npix[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"START",1,naxis[0],&iav,&start[0],&uni,&nulo);
  stat = SCDRDD(imnol[0],"STEP",1,naxis[0],&iav,&step[0],&uni,&nulo);

  
  /* all o.k. now allocate space for arrays */

  ao = fvector(0, npix[0] - 1);
  ae = fvector(0, npix[0] - 1);
  ao_err = fvector(0, npix[0] - 1);
  ae_err = fvector(0, npix[0] - 1);

  buffima = fvector(0, npix[0] - 1);
  bufferr = fvector(0, npix[0] - 1);







  if (polmode == CIRCULAR)
    {
      append[0] = 'V';      
      if (naxis[0] == 1) npix[1] = 1; /*1D for extracted frames*/

      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[0], &imno_err[0]);

      sprintf(output,"\nStokes parameter: %c\n",append[0]);
      SCTPUT(output);
      sprintf(output,"Output frames: %s %s\n",outima,outerr); 
      SCTPUT(output);


      bo = fvector(0, npix[0] - 1);
      be = fvector(0, npix[0] - 1);
      bo_err = fvector(0, npix[0] - 1);
      be_err = fvector(0, npix[0] - 1);

      for (j = 0; j < npix[1]; j++)
	{
	  offset = j*npix[0]+1;

	  for (n = 0; n < 4; n++){
	    frame[0] = '\0';
	    stat = TCERDC(tid, iselcol[n], frame_col, frame, &null);
	    frame_name(frame,frame_err,rebmode,&imnol[n],&imno_errl[n]);
	    printf("... %d %s %s %8.2f\n",iselcol[n],frame,frame_err,
		   phic[n]);
	  }

	  


	  SCFGET(imnol[0],offset,npix[0],&nulo, (char *)ao);
	  SCFGET(imnol[1],offset,npix[0],&nulo, (char *)ae);
	  SCFGET(imnol[2],offset,npix[0],&nulo, (char *)bo);
	  SCFGET(imnol[3],offset,npix[0],&nulo, (char *)be);

	  SCFGET(imno_errl[0],offset,npix[0],&nulo, (char *)ao_err);
	  SCFGET(imno_errl[1],offset,npix[0],&nulo, (char *)ae_err);
	  SCFGET(imno_errl[2],offset,npix[0],&nulo, (char *)bo_err);
	  SCFGET(imno_errl[3],offset,npix[0],&nulo, (char *)be_err);

	  if (sin(2.0 * phic[0] / RAD) > 0.0)
	    {
	      stat = stokesv(ao,ae,bo,be,ao_err,ae_err,bo_err,be_err,
			     buffima,bufferr,npix[0]);
	      printf("first frame phic > 0\n");
	    }
	  else
	    {
	    stat = stokesv(bo,be,ao,ae,bo_err,be_err,ao_err,ae_err,
			   buffima,bufferr,npix[0]);
	    printf("first frame phic < 0\n");
	    }
	  stat = SCFPUT(imno[0], offset, npix[0], (char*)buffima);
	  stat = SCFPUT(imno_err[0], offset, npix[0], (char*)bufferr);  

	}
      SCFCLO(imno[0]);
      SCFCLO(imno_err[0]);

      free_fvector(bo, 0, npix[0] - 1);  
      free_fvector(be, 0, npix[0] - 1);  
      free_fvector(bo_err, 0, npix[0] - 1);  
      free_fvector(be_err, 0, npix[0] - 1);  

    }
  else
    { /*linear polarisation*/

      /* construct output file names by appending to catalogue name */
      Q = fvector (0, npix[0] - 1);
      U = fvector (0, npix[0] - 1);
      Qerr = fvector (0, npix[0] - 1);
      Uerr = fvector (0, npix[0] - 1);
      pol = fvector(0, npix[0] - 1);
      polerr = fvector(0, npix[0] - 1);
      theta = fvector(0, npix[0] - 1);
      thetaerr = fvector(0, npix[0] - 1);
      

      if (naxis[0] == 1) npix[1] = 1; /*1D for extracted frames*/
      if (naxis[0] == 2 && toupper(rebmode[1]) != 'E') npix[1] = npix[1]/2; 
      /* it means that sigma is appended in the second half of the frame */

      append[0] = 'Q';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[0], &imno_err[0]);
      append[0] = 'U';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[1], &imno_err[1]);
      append[0] = 'I';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[2], &imno_err[2]);
      append[0] = 'p';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[3], &imno_err[3]);
      append[0] = 't';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[4], &imno_err[4]);
      append[0] = 'c';
      stat = outframes (append, catfil, start, step, npix, naxis, imnol, 
			outima,  outerr, &imno[5], &imno_err[5]);


      /*for any y row on the CCD*/
      for (j = 0; j < npix[1]; j++)
        {

          offset = j*npix[0]+1;


	  /*reset sums*/
	  for (i = 0; i < npix[0]; i++)
	    {
	      Q[i] = Qerr[i] = U[i] = Uerr[i]  = 0;
	    }
	  stat = stokesi(buffima,bufferr,ae,ae_err,npix[0],ipos[0],-1);


	  /*for any retarder position*/
          for (i = 0; i < ipos[0]; i++)
            {
	      frame[0] = '\0';
	      stat = TCERDC(tid, iselcol[i*2], frame_col, frame, &null);
	      frame_name(frame,frame_err,rebmode,&imnol[i*2],&imno_errl[i*2]);
	      printf("... %d %s %s %8.2f\n",iselcol[i*2],frame,frame_err,
		     phic[i*2]);
	      frame[0] = '\0';
	      stat = TCERDC(tid, iselcol[i*2+1], frame_col, frame, &null);
	      frame_name(frame,frame_err,rebmode,&imnol[i*2+1],
			 &imno_errl[i*2+1]);
	      printf("... %d %s %s %8.2f\n",iselcol[i*2+1],frame,frame_err,
		     phic[i*2+1]);


	      SCFGET(imnol[i*2],offset,npix[0],&nulo, (char *)ao);
	      SCFGET(imnol[i*2+1],offset,npix[0],&nulo, (char *)ae);

	      if (toupper(rebmode[1]) == 'E'){
		SCFGET(imno_errl[i*2],offset,npix[0],&nulo, (char *)ao_err);
		SCFGET(imno_errl[i*2+1],offset,npix[0],&nulo, (char *)ae_err);
	      }
	      else{
		erroff = offset + npix[1]*npix[0];

		SCFGET(imno_errl[i*2],erroff,npix[0],&nulo, (char *)ao_err);
		SCFGET(imno_errl[i*2+1],erroff,npix[0],&nulo, (char *)ae_err);
	      }


	      SCFCLO(imnol[i*2]);
	      SCFCLO(imnol[i*2+1]);
	      if (toupper(rebmode[1]) == 'E'){
		SCFCLO(imno_errl[i*2]);
		SCFCLO(imno_errl[i*2+1]);
	      }




              sfourier(ao,ao_err,ae,ae_err,phic[i*2],Q,Qerr,U,Uerr,npix[0],
		       ipos[0]);


              stat = stokesi(buffima,bufferr,ao,ao_err,npix[0],ipos[0],0);
              stat = stokesi(buffima,bufferr,ae,ae_err,npix[0],ipos[0],0);

	    }
	  for (i = 0; i < npix[0]; i++) /* sfourier computes sigma**2 */
	    {
	      Qerr[i] = (float) sqrt(Qerr[i]);
	      Uerr[i] = (float) sqrt(Uerr[i]);
	    }


	  stat = SCFPUT(imno[0], offset, npix[0], (char*)Q);
	  stat = SCFPUT(imno_err[0], offset, npix[0], (char*)Qerr);  
	  stat = SCFPUT(imno[1], offset, npix[0], (char*)U);
	  stat = SCFPUT(imno_err[1], offset, npix[0], (char*)Uerr);  
	  
	  stat = stokesi(buffima,bufferr,ae,ae_err,npix[0],ipos[0],1);
	  
	  stat = SCFPUT(imno[2], offset, npix[0], (char*)buffima);
	  stat = SCFPUT(imno_err[2], offset, npix[0], (char*)bufferr);  

	  /* compute p and theta WITH statistical corections */
	  stat = linpol(Q,Qerr,U,Uerr,pol,polerr,theta,thetaerr,npix[0],
			ipos[0],1);

	  stat = SCFPUT(imno[5], offset, npix[0], (char*)pol);

	  stat = SCFPUT(imno_err[3], offset, npix[0], (char*)polerr);  
	  stat = SCFPUT(imno_err[4], offset, npix[0], (char*)thetaerr);  

	  /* compute p and theta WITHOUT statistical corections */
	  stat = linpol(Q,Qerr,U,Uerr,pol,polerr,theta,thetaerr,npix[0],
			ipos[0],0);

	  stat = SCFPUT(imno[3], offset, npix[0], (char*)pol);
	  stat = SCFPUT(imno[4], offset, npix[0], (char*)theta);


	}
      
      stat = SCFCLO(imno[0]);
      stat = SCFCLO(imno_err[0]);
      stat = SCFCLO(imno[1]);
      stat = SCFCLO(imno_err[1]);
      stat = SCFCLO(imno[2]);
      stat = SCFCLO(imno_err[2]);
      stat = SCFCLO(imno[3]);
      stat = SCFCLO(imno_err[3]);
      stat = SCFCLO(imno[4]);
      stat = SCFCLO(imno_err[4]);
      stat = SCFCLO(imno[5]);
      stat = SCFCLO(imno_err[5]);

      free_fvector(Q, 0, npix[0] - 1);     
      free_fvector(U, 0, npix[0] - 1);     
      free_fvector(Qerr, 0, npix[0] - 1);    
      free_fvector(Uerr, 0, npix[0] - 1);    
      free_fvector(pol, 0, npix[0] - 1);  
      free_fvector(theta, 0, npix[0] - 1);  
      free_fvector(polerr, 0, npix[0] - 1);  
      free_fvector(thetaerr, 0, npix[0] - 1);  

      
    }      
      


  /*          epilogue ...                 */

  free_fvector(ao, 0, npix[0] - 1);  
  free_fvector(ae, 0, npix[0] - 1);  
  free_fvector(ao_err, 0, npix[0] - 1);  
  free_fvector(ae_err, 0, npix[0] - 1);  
  free_fvector(buffima, 0, npix[0] - 1);  
  free_fvector(bufferr, 0, npix[0] - 1);  
  
  SCSEPI();
  exit(0);
}

/*---------------------------------------------------------------------------*/



#ifdef __STDC__
void frame_name(char frame[], char frame_err[], char rebmode[], int *iml, 
		int *im_errl)
#else
void frame_name(frame,frame_err,rebmode,iml,im_errl)
     char frame[], frame_err[], rebmode[];
     int *iml,*im_errl;
#endif

{
  int i,stat;

  frame_err[0] = '\0';

  if (strstrs(frame,".bdf") == 0)
    {
      if (toupper(rebmode[1]) == 'E'){
	sprintf(frame_err,"%sE.bdf\0",frame);
	strcat(frame,".bdf\0");
      }
      else
	{
	  strcat(frame,".bdf\0"); 
	}
    }
  else
    {
      if (toupper(rebmode[1]) == 'E'){
	i = strlen(frame) - 4;
	strncpy(frame_err,frame,i);
	strcat(frame_err,"E.bdf\0");
	
	frame[i] = '\0';
	strcat(frame,".bdf\0"); 
      }
    }

  stat = SCFOPN(frame,D_R4_FORMAT,0,F_IMA_TYPE,iml); 
  if (toupper(rebmode[1]) == 'E'){
    stat = SCFOPN(frame_err,D_R4_FORMAT,0,F_IMA_TYPE,im_errl);
  }
  else if (toupper(rebmode[0]) == '1'){
    im_errl = iml;
  }
  else{
    printf("unsupported rebin mode\n");
    exit(1);
  }

  return;
}

#undef LINEAR
#undef CIRCULAR
#undef MAXIMS
#undef myabs






