/* @(#)spolyoffset.c     6.1.1.4  (LSW)  07/02/94  13:48:49           */
/*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* .COPYRIGHT   (C) 1994 Landessternwarte Heidelberg                 */
/* .IDENT        spolyoffset.c                                       */
/* .AUTHOR       Sabine Moehler - LSW Heidelberg                     */
/* .KEYWORDS                                                         */
/* .PURPOSE      Search object with median: (val - med(val)) > thres */ 
/*               Center object with Gaussian fitting                 */
/*               (adapted from mosslin.c)                            */
/* .VERSION      1.0  07-FEB-1994  Implementation                    */
/*               1.1  26-JUL-1994  Adaption for SPOLDEFINE           */
/*               1.2  20-OCT-1994  Adaption for SPOLYOFFSET          */
/*-------------------------------------------------------------------*/

#include <math.h> 
#include <tbldef.h>
#include <midas_def.h>
#include <ok.h>

#include <proto_mutil.h>
#include <proto_nrutil.h>
#include <proto_fitnol.h>
#include <dlfit.h>

#define max(A,B) ((A)  > (B) ? (A) : (B)) 
#define min(A,B) ((A)  < (B) ? (A) : (B)) 

#define MAXLINES 1000
#define MAXSLITS 100

#define GAUSSIAN 1
#define MARK -9999.9

int    inull, off_col[6];
int    Ntot,Npix[2],Method,lower[40],upper[40];
int    iav, kun, knul, Xbin,Iwin,Width;
double dnull, Start[2],Step[2], *Xgaus, *Ygaus, *A;
float  rnull, thresh;

#ifdef __STDC__
double  center_source( float [], float [], int ); 
void  search_obj( float [], int [], int, int *);
double  fit_source( float [], int [], int, int);
float own_median(int, float [], float *);
#else
double  center_source(); 
void  search_obj();
double  fit_source();
float own_median();
#endif

main()

{
   int   nax, tidoff, imno;
   int   slit[40],nslit,acol,arow,nsort,param[2];
   int   yn, nrow, ncol, posrow;

   char  method[21],obj_image[60],slitoff[60];
   char  unit[64],inst[72],line[80],linem[80];

   float *buff,*image, rpar[1];
   double yoff, obj[3];

/* get into Midas and get parameters */

   SCSPRO("spolyoffset");
   
   SCKGETC("IN_A",1,60,&iav,obj_image);
   SCKGETC("SLITOFF",1,60,&iav,slitoff);
   SCKGETC("INPUTC",1,3,&iav,method);
   SCKRDI("INPUTI",1,2,&iav,param,&kun,&knul);
   SCKRDI("SLITSEL",1,40,&iav,slit,&kun,&knul);
   SCKRDI("LIM_LOW",1,40,&iav,lower,&kun,&knul);
   SCKRDI("LIM_UP",1,40,&iav,upper,&kun,&knul);
   SCKRDR("INPUTR",1,1,&iav,rpar,&kun,&knul);

/* map input file */

   SCIGET(obj_image,D_R4_FORMAT,F_I_MODE,F_IMA_TYPE,
		2,&nax,Npix,Start,Step,inst,unit,(char **)&image,&imno);

   Method = GAUSSIAN;
   sprintf(linem,"centering method:    Gaussian");

/* open output table and get column numbers */
   TCTOPN (slitoff, F_IO_MODE, &tidoff);
   TCIGET (tidoff, &ncol, &nrow, &nsort, &acol, &arow);
   TCLSER (tidoff, "Frame", &off_col[0]);
   TCLSER (tidoff, "Slit_Low", &off_col[1]);
   TCLSER (tidoff, "Y_Low", &off_col[2]);
   TCLSER (tidoff, "Slit_Up", &off_col[3]);
   TCLSER (tidoff, "Y_Up", &off_col[4]);
   TCLSER (tidoff, "Y_Off", &off_col[5]);
   posrow = nrow+1;

/* set parameters */

   thresh  =  rpar[0];
   Width  =   param[0];
   Iwin  =   (param[0]-1)/2;
   Xbin  =   param[1];

   SCTPUT("\nsearch object ");
   SCTPUT("-------------");
   sprintf(line,"Input image:         %s ",obj_image);
   SCTPUT(line);
   sprintf(line,"Output table:        %s ",slitoff);
   SCTPUT(line);
   sprintf(line,"detection threshold: %6.2f DN",thresh);
   SCTPUT(line);
   sprintf(line,"search window:       %i pixels",Width); 
   SCTPUT(line);
   sprintf(line,"average on:          %i scan lines",Xbin);
   SCTPUT(line);
   SCTPUT(linem);

   Ntot = Start[1]+Step[1]*Npix[1];
   buff = (float *) osmmget(Ntot * sizeof (float));
/*   buff = (float *) osmmget(Npix[1] * sizeof (float)); */
   Xgaus = dvector(1,Width);
   Ygaus = dvector(1,Width);	
   A = dvector(1,3);

   yoff = 0;
   yn = 0;
   nslit = 0;

/* loop over all selected slitlets */
/*   for (nslit=0; nslit < 40; nslit = nslit+2) */
   while (nslit < 40)
     {
      if (slit[nslit] > inull)
        {
/* center object in first slitlet  */
         obj[1] = center_source(image, buff, nslit);
         TCEWRC (tidoff, posrow, off_col[0], obj_image);
         TCEWRI (tidoff, posrow, off_col[1], &slit[nslit]);
         TCEWRD (tidoff, posrow, off_col[2], &obj[1]);
/* center object in second slitlet  */
         nslit++;
         obj[2] = center_source(image, buff, nslit);
         obj[3] = obj[2]-obj[1];
         TCEWRI (tidoff, posrow, off_col[3], &slit[nslit]);
         TCEWRD (tidoff, posrow, off_col[4], &obj[2]);
         TCEWRD (tidoff, posrow, off_col[5], &obj[3]);
         nslit++;
         yoff = yoff+obj[3];
         yn++;
         posrow++;
        }
     else nslit = nslit+2; 
     }
  yoff = yoff/yn;
  SCDWRD (tidoff, "YOFFMEAN", &yoff, 1, 1, &kun);
/* epilogue ... */

   TCTCLO(tidoff);    
   SCSEPI();
}
/*--------------------------------------------------------------------------*/
#ifdef __STDC__
double center_source(float image[], float rval[], int nslit)
#else
double center_source( image, rval, nslit )
  float image[],rval[];
  int   nslit;
#endif

{
   int    nact, j, k, jj, objlist[MAXLINES];
   int    center;
   float  sum;
   double ypos, scan_pos;

   SCKRDD("SCAN_POS", 1, 1, &iav, &scan_pos, &kun, &knul); 
   center = (int) ((scan_pos-Start[0])/Step[0]); /* transform world coordinates
                                                 to pixels */
/* loop over all slitlets 

   for (nslit=0; nslit < 40; nslit+2) 
     {
      if (slit[nslit] > inull)
        {
         obj[1] = slit[nslit]; */

/* average XBIN columns at chosen position */
         for (j=0;j<Ntot;j++) rval[j] = 0.0;
/*         for (j=0;j<Npix[1];j++) rval[j] = 0.0;    */
         for (k=lower[nslit]-1; k<upper[nslit]-1; k++)
            {
             sum = 0;
             for (j=center-Xbin/2-1; j<=center+Xbin/2-0.5; j++) 
               {                      /* MIDAS starts arrays with 1, C with 0 */
                jj = j+Npix[0]*(k-(int)(Start[1])+1);
/*                jj = Npix[0]*k+j;   */
                rval[k] = rval[k]+image[jj]; /* image contains intensities of
                                                frame in pixel coordinates! */
                sum = sum+1;
	       }
             rval[k] = rval[k]/sum;
            }
/* now search for object ...  */
          search_obj(rval,objlist,nslit,&nact);
/* ... fit object ...  */
          ypos = fit_source(rval,objlist,nact,nslit);	
          return ypos;
/*
         }
     }  */
}
/*---------------------------------------------------------------------------*/
#ifdef __STDC__
void search_obj( float rval[], int objlist[], int nslit, int *nact)
#else
void search_obj( rval, objlist, nslit, nact )
  float rval[];
  int   objlist[],nslit,*nact;
#endif

{
   int number = 0,j,i,imax;
   float med,diff,max_intens,Thresh;

   for (j=lower[nslit]-1+Iwin;j<upper[nslit]-1-Iwin;j++) 
     {
      
      diff = rval[j]-own_median(Width,&rval[j-Iwin],&med);                  
      if (med<0.) med = -med;
      Thresh = (thresh>0.) ? thresh : -thresh*med;
      if (diff > Thresh) 
        {
         imax = j;
         max_intens = rval[j];
         for (i=j-Iwin;i<=j+Iwin;i++) 
           if (rval[i] > max_intens )  
             {
              max_intens = rval[i];
              imax = i;
             }
           objlist[number] =  imax;
           number++;
        }
     }
   *nact = number;
}
/*---------------------------------------------------------------------------*/
#ifdef __STDC__
double fit_source( float rval[], int objlist[], int nact, int nslit)
#else
double fit_source( rval, objlist, nact, nslit)
  float rval[];
  int   objlist[],nact,nslit;
#endif
{
   int    j,k,jj,ii,index;
   double Aold[3];
   float  back, med; 

   for (jj=0; jj<=2; jj++) Aold[jj] = 0.0;
   for (jj=0;jj<nact;jj++) 
     {
      index = objlist[jj];

/* Determine local background (but ensure to take it within the slitlet) */

      if (index-Iwin-10 > lower[nslit] && index+Iwin+10 < upper[nslit])
        { 
         back = own_median(Width,&rval[index-Iwin-10],&med);
         back = (back + own_median(Width,&rval[index+Iwin+10],&med))/2;
        }
      else if (index-Iwin-10 <= lower[nslit]) 
          back = own_median(Width,&rval[index+Iwin+10],&med);   
      else if (index+Iwin+10 >= upper[nslit]-1) 
          back = own_median(Width,&rval[index-Iwin-10],&med);
      med = own_median(Width,&rval[index],&med);
      if (back > med) back = med;

/* GAUSSIAN fit of object profile */

      A[1] = rval[index];
/*      A[2] = Start[1] + Step[1]*objlist[jj];  */
      A[2] = objlist[jj]+1;
      A[3] = Step[1];
      for (k = 1 , j = index-Iwin ; j<= index+Iwin; j++, k++) 
        {
/*         Xgaus[k] = Start[1] + Step[1]*objlist[jj]+k-Iwin-1;    */
         Xgaus[k] = objlist[jj]+k-Iwin;
         Ygaus[k] = rval[j]-back;
        }	
      fit_gauss(Xgaus,Ygaus,Width,A);  

/* Avoid finding the same object several times  */
      if (Aold[0] != A[0] || Aold[1] != A[1] || Aold[2] != A[2])
         for (ii=0; ii<=2; ii++) Aold[ii] = A[ii];
     }                 
   return A[2];
}
/*---------------------------------------------------------------------------*/
#ifdef __STDC__
float own_median(int n, float arr[], float *med)
#else
float own_median(n, arr, med)
int n;
float arr[],*med;
#endif
/*              find median of distribution            */
{
   int   i,j,index;
   float a,b[21];

   for (j=0;j<n;j++) b[j] = arr[j];

   for (j=1;j<n;j++) 
     {
      a=b[j];
      i=j-1;
      while (i >= 0 && b[i] > a) 
        {
         b[i+1]=b[i];
         i--;
        }
      b[i+1]=a;
     }
   index = (n-1)/2;
   *med = b[index];
   return *med;
}
